# -------------------------------------
# Script: econometrics_functions.R
# Author: Michal Kvasnička
# Purpose: This script contains functions used for
# Notes: For usage see script
#
# Copyright(c) Masaryk University
# -------------------------------------


# load necessary packages
library(dplyr)
library(tidyr)
library(stringr)
library(purrr)


# other necessary functions
source(file.path(CODEDIR, "utilities.R"))
source(file.path(CODEDIR, "preparatory_functions_general.R"))





# add brand names --------------------------------------------------------------

# add_brandnames(station_data, brands) adds a column brand_name to station_data;
# it contains names of selected brands---the rest of them have brand_name equal
# to "other"
#
# inputs:
# - station_data ... (tibble) table created by prepare_average_fuel_data()
# - brands ... (character vector) vector of station names that should be kept by
#   name; other stations will have its value "other"
#
# output:
# - tibble with the same structure as station_data and added column brand_name;
#   brand_name is a factor; its reference level is "other"; the rest of names
# are sorted alphabetically
#
# usage:
#   named_brands <- c("shell", "benzina", "eurooil", "omv")
#   nsel_brands <- nsel %>% add_brandnames(named_brands)
add_brandnames <- function(station_data, brands) {
    lvls <- c("other", sort(unique(brands)))
    station_data %>%
        dplyr::mutate(brand_name = dplyr::if_else(name %in% brands, name, "other"),
                      brand_name = factor(brand_name, levels = lvls))
}





# takeovers --------------------------------------------------------------------

# function get_taken_over_ids(data, takeovers) tries to guess from the data the
# ids of stations that were taken over
#
# inputs:
# - data ... (tibble) a data table created by prepare_average_fuel_data() or its
#   subset
# - takeovers ... (tibble) a table with four columns:
#   - takeover_date, terminal_date ... (Date) of the takeover and its last
#     period (from_date)
#   - source, target ... (character scalar) the brand name taken over and the
#     brand name that took over
#
# warnings:
# - you should always check the result of this function manually! for reasons,
#   see notes
#
# notes:
# - the function works like this:
#   - first, if finds all stations that at the takeover_date or before belonged
#     to the taken-over stations
#   - second, it removes all stations that between takeover_date and
#     terminal_date belonged to any other brand than source and target
#   - third, it removes all stattions that do not belong to target at the
#     terminal_date
# - what can go wrong:
#   - if a station belonged to source in the past, then to other brand, and at
#     the end to target, it can be wrongly classified as a part of the merger;
#     (the special case is when the other brand is in fact target, i.e. if
#     target first takes over part of source and later the rest of it)
#   - if a station is rebranded to target at takeover_date, it is wrongly
#     classified as not-part of the merger
#
# output:
# - tibble takeovers which has added a list column ids which includes ids of the
#   stations that were taken over from the source to target within the given
#   time as described by the input tibble takeovers
get_taken_over_ids <- function(data, takeovers) {
    get_ids <- function(data, takeover) {
        # TODO: přidat kontrolu o datum!!!
        ids <- data %>%
            dplyr::filter(name == takeover$source,
                          from_date <= takeover$takeover_date) %>%
            dplyr::distinct(id) %>%
            dplyr::pull()
        not_taken_over <- data %>%
            dplyr::filter(id %in% ids,
                          from_date >= takeover$takeover_date,
                          from_date <= takeover$terminal_date) %>%
            dplyr::group_by(id) %>%
            dplyr::arrange(desc(from_date), .by_group = TRUE) %>%
            dplyr::filter(!base::all(name %in% c(takeover$source,
                                                 takeover$target)) |
                              !dplyr::first(name) == takeover$target) %>%
            dplyr::distinct(id) %>%
            dplyr::ungroup() %>%
            pull(id)
        if (length(not_taken_over) > 0)
            warning("Takeover from ",
                    takeover$source, " to ", takeover$target, ": ",
                    "Stations with the following ids are not taken over: ",
                    stringr::str_c(not_taken_over, collapse = ", "), ".")
        base::setdiff(ids, not_taken_over)
    }

    takeovers$ids <- purrr::map(base::seq_len(base::nrow(takeovers)),
                                ~get_ids(data, takeovers[., ]))
    takeovers
}


# add_brand_changes_in_takeover(data, takeovers) adds to data table a new column
# on brand changes after big takeovers
#
# inputs:
# - data ... (tibble) a data table created by repare_average_fuel_data() or its
#   subset
# - takeovers ... (tibble) a table with four columns:
#   - takeover_date, terminal_date ... (Date) of the takeover and its last
#     period (from_date)
#   - source, target ... (character scalar) the brand name taken over and the
#     brand name that took over
# - keep_levels ... (logical) implicitely FALSE; if TRUE, it keeps all former
#   levels of the brand_change variable
# - keep_values ... (logical) implicitely FALSE; if TRUE, it keeps all values
#   of the brand_change variable for stations that have not been taken over,
#   otherwise it sets them to the implicit value ("none")
#
# output:
# - tibble of the same number of rows with the same variables as input data plus
#   one more: brand_change (factor) in
#   format "brand_taken_over -> brand_taking_over" or "none"
#
# details:
# - if takeover has a column ids, it is used; otherwise it is calculated first
# - if keep_level or keep_values is TRUE, the brand_changes column must already
#   be present in the data
# - there are two uses of this function: 1) to add brand_changes column from the
#   scratch -- then leave keep_level and keep_values to its implicit (FALSE)
#   values; 2) in merger simulation -- there both these parameters must be set
#   to TRUE
#
# usage:
#   takeovers <- tribble(
#       ~takeover_date, ~terminal_date, ~source, ~target,
#       "2014-08-01", "2018-12-31", "lukoil", "mol",
#       "2015-08-01", "2018-12-31", "agip", "mol",
#       "2016-02-01", "2018-12-31", "omv", "benzina"
#   ) %>%
#       mutate_at(c("takeover_date", "terminal_date"), as.Date)
#   natural_monthly <- add_brand_changes_in_takeover(natural_monthly, takeovers)
add_brand_changes_in_takeover <- function(data, takeovers,
                                          keep_levels = FALSE,
                                          keep_values = FALSE) {
    if (keep_levels || keep_values) {
        if (!("brand_change" %in% base::colnames(data)))
            stop("Levels or values cannot be kept---",
                 "there is no brand_change column in the data!")
        ori_lvls <- base::levels(data$brand_change)
    }
    implicit_level <- "none"
    if (!("ids" %in% base::colnames(takeovers)))
        takeovers <- get_taken_over_ids(data, takeovers)
    if (!keep_values)
        data <- data %>% dplyr::mutate(brand_change = implicit_level)
    for (k in base::seq_len(base::nrow(takeovers)))
        data$brand_change[data$id %in% takeovers$ids[[k]] &
                              data$from_date >= takeovers$takeover_date[[k]] &
                              data$from_date <= takeovers$terminal_date[[k]]
        ] <- stringr::str_c(takeovers$source[k], " -> ", takeovers$target[k])
    lvls <- base::as.character(base::sort(base::unique(data$brand_change)))
    lvls <- base::unique(c(implicit_level, lvls))
    if (keep_levels) {
        if (base::length(base::setdiff(lvls, ori_lvls)) > 0)
            stop("Levels cannot be kept---",
                 "there are levels in takeover which are not in data!")
        lvls <- ori_lvls
    }
    data %>%
        dplyr::mutate(brand_change = base::factor(brand_change, levels = lvls))
}





# econometrics -----------------------------------------------------------------

# remove_nonexistent_stations(station_data, existence = NULL) removes stations
# that do not exist from station_data; the result is a balanced panel
#
# inputs:
# - station_data ... (tibble) table created by prepare_average_fuel_data()
# - existence ... (character) name of a logical columns in station_data which
#   says which stations do exist in given periods
#
# output:
# - tibble with same colums as station_data but only selected rows
#
# usage:
#   existence_criterium <- "station_exists_30"
#   natural_monthly <- natural_monthly %>% remove_nonexistent_stations(existence_criterium)
remove_nonexistent_stations <- function(station_data, existence = NULL) {
    # pokud není zadán žádný sloupec existence, vrať původní data nezměněná
    if (is.null(existence))
        return(station_data)
    # v opačném případě vyber sloupec, který indikuje existenci a vrať jen ty řádky ze
    # station data, které mají existence TRUE
    existence <- station_data[[existence]]
    station_data[existence, ]
}


# balance_panel(data, existence) takes tibble created by
# prepare_average_fuel_data() and augmented with add_spatial_clustering(),
# add_competitors_in_distances(), add_distance_to_closest_competitor() and the
# like, and it removed all stations that did not existed in all periods, i.e.
# it balances the panel
#
# input:
# - station_data ... (tibble) table created by prepare_average_fuel_data();
#   it included data on many stations in time many periods but only one fuel;
#   technically, it is a panel but potentially unbalanced
# - existence ... (character scalar) name of column in station_data which
#   indicates whether the station existed in a particular period of time
#
# output:
# - the same tibble as station_data but the panel is balanced by removing
#   observations
#
# warnings:
# - strictly speaking, this is a poor man's solution; the correct solution is to
#   have different weighting matrix for each period; however, this estimation
#   procedure is not available in R now
#
# usage:
#   natural_monthly_balanced <- balance_panel(natural_monthly,
#       "station_exists_30")
balance_panel <- function(station_data, existence) {
    # learn the number of periods in the data
    number_of_periods <- station_data %>%
        dplyr::distinct(from_date) %>%
        base::nrow()
    # find ids of stations that existed in all periods
    keep_id <- station_data %>%
        dplyr::group_by(id) %>%
        dplyr::summarise(n = sum(!!rlang::sym(existence)),
                         .groups = "drop") %>%
        dplyr::filter(n == number_of_periods) %>%
        dplyr::pull(id)
    # return the stations that existed in every period of time
    dplyr::filter(station_data, id %in% keep_id)
}


# prepare_for_spatial_panel_econometrics() prepares data for spatial panel
# estimation
#
# inputs:
# - station_data ... (tibble) data on a selected fuel created by
#   prepare_average_fuel_data()
# - existence ... (character scalar) name of logical column in station_data
#   which says whether the given stations exists in the particular time
# - formula ... (formula) formula that includes all variables that is necessary
#   to keep in station_data
# - ... ... names of other variables that must be kept in the table; usually,
#   it is not necessary to use them as they they are read from the formula
#
# output:
# - tibble with the same structure as station_data but
#   - only the selected variables (columns)
#   - all rows containing NA are dropped (estimation function spml() cannot
#     handle them)
#   - a balanced panel is created
#   - observations are sorted by id and then time (as necessary for spml())
prepare_for_spatial_panel_econometrics <- function(station_data,
                                                   existence = NULL,
                                                   formula = NULL,
                                                   ...) {
    vars <- rlang::enquos(...)
    station_data %>%
        dplyr::select(id, from_date, to_date, lon, lat, name, price,
                      if (!is.null(existence)) tidyselect::all_of(existence),
                      if (!is.null(formula))
                          tidyselect::all_of(base::all.vars(formula)),
                      !!! vars) %>%
        tidyr::drop_na() %>%
        balance_panel(existence) %>%
        dplyr::arrange(id, from_date)
}


# prepare_for_crosssectional_econometrics() prepares data for spatial
# cross-sectional estimation
#
# inputs:
# - station_data ... (tibble) data on a selected fuel created by
#   prepare_average_fuel_data() filtered for just one period (one unique
#   from_date)
# - existence ... (character scalar) name of logical column in station_data
#   which says whether the given stations exists in the particular time
# - formula ... (formula) formula that includes all variables that is necessary
#   to keep in station_data
# - ... ... names of other variables that must be kept in the table; usually,
#   it is not necessary to use them as they they are read from the formula
#
# output:
# - data.frame with the same structure as station_data but
#   - only the selected variables (columns)
#   - all rows containing NA are dropped (estimation function lagsarlm() cannot
#     handle them)
#   - rownames equal to ids are added -- it is necessary for prediction in
#     merger simulations (that's why this is data.frame, not tibble)
#
# warnings:
# - the output is data.frame, not tibble, and it has rownames -- you must not
#   alter these properties
prepare_for_crosssectional_econometrics <- function(station_data,
                                                    existence = NULL,
                                                    formula = NULL,
                                                    ...) {
    station_data <- station_data %>%
        prepare_for_spatial_panel_econometrics(existence, formula, ...) %>%
        as.data.frame()
    rownames(station_data) <- station_data$id
    station_data
}


# function create_weights() creates a list of weights needed for spatial econometrics
#
# inputs:
# - data ... (tibble) dataset created by function prepare_average_fuel_data() which
#   includes data on many stations in many periods but for only one fuel (i.e. a panel) or
#   a selection from such an dataset for just one period (for a cross-sectional analysis)
# - index ... (optional, character vector of length 2) names of columns in data which
#   include indices of cross-sectional and time units (in this order); implicitely
#   c("id", "from_date")
# - neighbors ... (character scalar) a type of neighbors; must be one of "beeline" (the
#   implicite value), "driving", or "voronoi"
# - driving_distances ... (tibble) a table of driving ditances created by function
#   get_driving_distances(); is used only when neighbor == "driving"
# - max_dist ... (double scalar) the maximal distance in kilometers to which a station is
#   considered as a neighbor (the minimal max_dist is zero); it is used only when neighbor
#   is equal to "beeline" or "driving"
# - transform ... a function of one variable whicht transforms a distance to a weight
#   (before the row standardization); implcitely 1/x; NULL, if no transformation is
#   needed; only useful for neighbors equal to "beeline" or "driving"
# - select_function ... function which takes two variables .x and .y (typically min, max,
#   or mean) which selects which of the distances from A to B or from B to A is used,
#   whether the shorter one, the longer one, or their mean; implicitly the mean (it is not
#   sufficient to say mean, so if you want to use mean, leave the implicite values); you
#   can use a formula which includes .x and .y that are used in map2(); only used when
#   neighbor = "driving"
# - arccr_folder ... (character scalar) a path to the folder where ARCCR administrative
#   data are stored; only used when neighbor = "voronoi"
# - layer ... (character scalar) a layer used in ARCCR data, implicitely "StatPolygon";
#   only used when neighbor = "voronoi"
#
# output:
#   the structure returned by spdep::nb2listw()
#
# warnings:
# - if non-existing stations must be eliminated from the data before the spatial weight
#   are calculated; when the spatial weights are computed, the number and composition of
#   the stations must not change!
# - voronoi neighbors calculated by this function are then those that were not eliminated,
#   i.e. they consists of the stations that existed in all periods considered; thus they
#   might differ from the voronoi neighbors used to calculate the spacial clustering (SC)
#   in any period!
create_weights <- function(data,
                           index = c("id", "from_date"),
                           neighbors = "beeline",
                           max_dist = 20, transform = function(x) 1 / x,
                           driving_distances = NULL,
                           select_function = ~mean(c(.x, .y)),
                           arccr_folder = NULL,   # "map_shapes/AdministrativniCleneni_v13.gdb"
                           layer = "StatPolygon") {
    # type of neighbors must be one of the following key words
    stopifnot(neighbors %in% c("beeline", "driving", "voronoi"))
    # reorder data by cross-sections, then time
    ord <- order(data[[index[2]]], data[[index[1]]])
    data <- data[ord, ]
    # construct nb and dlist for beeline neighbors
    if (neighbors == "beeline") {
        data <- data %>%
            distinct(id, lon, lat) %>%
            dplyr::arrange(id)
        stopifnot((data %>%
                       dplyr::group_by(id) %>%
                       summarise(n = n(), .groups = "drop") %>%
                       pull(n)) == 1)
        # get GPS location of each station
        coords <- as.matrix(data[, c("lon", "lat")])
        nb <- spdep::dnearneigh(coords, d1 = 0, d2 = max_dist,
                                row.names = data$id, longlat = TRUE)
        dlist <- spdep::nbdists(nb, coords, longlat = TRUE)
    }
    # construct nb and dlist for driving neighbors
    if (neighbors == "driving") {
        if (is.null(driving_distances))
            stop("Driving distances must be given!")
        covered_ids <- data %>% distinct(id) %>% dplyr::arrange(id) %>% pull(id)
        driving_distances <- driving_distances %>%
            dplyr::filter(from_id %in% covered_ids, to_id %in% covered_ids)
        driving_distances <- dplyr::left_join(driving_distances,
                                              dplyr::rename(driving_distances,
                                                            other_distance = driving_distance),
                                              by = c("from_id" = "to_id", "to_id" = "from_id")) %>%
            dplyr::mutate(distance = purrr::map2_dbl(driving_distance, other_distance,
                                                     select_function)) %>%
            dplyr::arrange(from_id, to_id)
        nb <- purrr::map(covered_ids,
                         ~match(driving_distances$to_id[(driving_distances$from_id == .) &
                                                            (driving_distances$distance <= (max_dist * 1e3))],
                                covered_ids))
        dlist <- purrr::map(covered_ids,
                            ~driving_distances$distance[(driving_distances$from_id == .) &
                                                            (driving_distances$distance <= (max_dist * 1e3))])
    }
    # construct nb and dlist for voronoi neighbors
    if (neighbors == "voronoi") {
        if (is.null(arccr_folder))
            stop("The arccr_folder must be given!")
        max_dist <- 1
        data <- data %>%
            distinct(id, lon, lat) %>%
            dplyr::arrange(id)
        stopifnot((data %>%
                       dplyr::group_by(id) %>%
                       summarise(n = n(), .groups = "drop") %>%
                       pull(n)) == 1)
        covered_ids <- data %>% distinct(id) %>% dplyr::arrange(id) %>% pull(id)
        data <- data %>% dplyr::mutate(.row_number = row_number())
        res <- get_voronoi_polygons(data, arccr_folder = arccr_folder, layer = layer)
        # nb <- map(res$voronoi_neighbors_id, ~dplyr::filter(data, id %in% .) %>% pull(.row_number))
        nb <- map(res$voronoi_neighbors_id, ~data$.row_number[data$id %in% .])
        dlist <- map(nb, ~rep(1, times = length(.)))
    }
    # set classes and attributes for nb and dlist for driving and voronoi neighbors
    if (neighbors %in% c("driving", "voronoi")) {
        class(nb) <- "nb"
        attr(nb, "call") <- "my own"
        attr(nb, "nbtype") <- "distance"
        attr(nb, "region.id") <- covered_ids
        attr(nb, "dnn") <- c(0, max_dist)
        attr(nb, "distances") <- c(0, max_dist)
        attr(nb, "bounds") <- c("GT", "LT")
        attr(nb, "sym") <- TRUE
        class(dlist) <- "nbdist"
        attr(dlist, "call") <- "my own"
    }
    # transform dlist distances (unless transform is NULL)
    if (!is.null(transform))
        dlist <- purrr::map(dlist, transform)
    # create and return wlist
    spdep::nb2listw(nb, glist = dlist, style = "W")
}


# estimate_spml() estimates spatial panel model; it prepares its data using
# prepare_for_spatial_panel_econometrics() which drops some observations;
# it also creates a weight matrix
#
# inputs:
# - data ... (tibble) data created by prepare_average_fuel_data()
# - existence_criterium ... (character scalar) a name of one column in data
#   which says what stations do exist in the given periods
# - formula ... (formula) ... estimated model; only variables listed in the
#   formula will be kept, the other variables will be dropped
# - driving_distances ... (optional, tibble) table of driving distances;
#   if it is given, then the weighting matrix is computed using driving
#   distances; if it is NULL (default value) then the weighting matrix is
#   based on beeline distances
# - max_dist ... (numeric scalar) maximal distance between stations that are
#   considered as neighbors (in km; default is 20 km)
# - transform ... (function) how distances among stations should be transformed;
#   implicitly function(x) 1 / x
# - lag ... (logical) whether the function should include autocorrelation
#   in dependent variable (TRUE) or not (FALSE); default is TRUE
# - spatial.error ... (character scalar) ... whether autocorrelation of
#   residuals should be covered:
#   - "none" ... none
#   - "b" / "kpp" ... yes -- for details see documentation for spml()
#
# details:
# - a model with fixed effects for both time and stations is estimated by
#   default
estimate_spml <- function(data, existence_criterium, formula,
                          driving_distances = NULL, max_dist = 20,
                          transform = function(x) 1 / x,
                          lag = TRUE, spatial.error = "none") {
    data <- data %>%
        prepare_for_spatial_panel_econometrics(existence_criterium, formula)
    W <- create_weights(data,
                        driving_distances = driving_distances,
                        max_dist = max_dist,
                        transform = transform)
    spml(formula = formula,
         data = data,
         index = c("id", "from_date"),
         listw = W,
         model = "within",
         effect = "twoways",
         lag = lag,
         spatial.error = spatial.error)
}


# function spml2() is a wrapper for splm::spml(); it has the same inputs and
# outputs as the original function; it adds the the output object data necessary
# to make predictions
#
# inputs are the same as for spml() + two more
# - index_id ... (character scalar) ... the name of the column in data which
#   contains id index -- the index must be present in data even if there is no
#   fixed effect for it in the model; must be supplied only when effect is set
#   to "time"
# - index_time ... (character scalar) ... the name of the column in data which
#   contains time index -- the index must be present in data even if there is no
#   fixed effect for it in the model; must be supplied only when effect is set
#   to "individual"
#
# output
# - see help for splm::spml()
# - some information needed for predictions is added to spml2 slot
#
#
# notes:
# - index must be supplied; it is not guessed from data
# - data must include both column that contains individual id and column that
#   contains time id -- even when you want to add fixed effect of one of these
#   only
# - the output has to return the same class as spml() because its methods do
#   not test for inheritance but for the class exactly; therefore, if you need
#   to find out which function has created the result, search for spml2 slot in
#   the output
spml2 <- function(formula, data, index = NULL, listw, listw2=listw, na.action,
                  model = c("within","random","pooling"),
                  effect = c("individual","time","twoways"),
                  lag = FALSE, spatial.error = c("b","kkp","none"),
                  index_id = NULL, index_time = NULL,
                  ...) {
    require(splm)
    if (is.null(index))
        stop("You must supply index; I do not guess it from data.")
    if (effect == "time" && is.null(index_id))
        stop("You must supply index_id when you estimate time effect only.")
    if (effect == "individual" && is.null(index_time))
        stop("You must supply index_time when you estimate time effect only.")
    if (effect == "time" && !(index_id %in% names(data)))
        stop("index_id must contain a name of valid column in data.")
    if (effect == "individual" && !(index_time %in% names(data)))
        stop("index_time must contain a name of valid column in data.")
    estimate <- splm::spml(formula = formula, data = data, index = index,
                           listw = listw, listw2 = listw2, na.action = na.action,
                           model = model, effect = effect,
                           lag = lag, spatial.error = spatial.error,
                           ...)
    if (estimate$effects == "spfe")
        index <- c(index, index_time)
    if (estimate$effects == "tpfe")
        index <- c(index_id, index)
    estimate$spml2$index <- index
    estimate$spml2$formula <- formula
    estimate$spml2$data <- data
    estimate$spml2$listw <- listw
    estimate
}


# predict.splm() is a prediction methods for selected splm models; the model
# must be estimated by spml2() wrapper
#
# inputs:
# - object ... (splm) a model estimated by spml2()
# - newdata ... (optional; inherits from data.frame) new data used for
#   prediction; if the model includes time or id fixed effect, only
#   the time and id may have only the levels that were present at the time of
#   the model estimation; if not supplied, data used at the time of estimation
#   are used
# - listw ... (optional; listw) a weighting matrix given as list; if not given,
#   the listw used at the time of estimation is used; (honestly, I can't see any
#   sensible reason to provide a new listw for the prediction)
#
# output:
# - (numeric vector) of predictions sorted as newdata
#
# limitations:
# - when the estimated formula includes a factor or string and newdata is
#   supplied, then the corresponding variable in newdata must include all levels
#   and only the levels that were there when the model was estimated;
#   the levels may be there physically or as factors levels (i.e. levels must
#   not be dropped).
#
# warning:
# - the estimate must be produced by spml2() wrapper
predict.splm <- function(object, newdata = NULL, listw = NULL) {
    # check the inputs
    stopifnot(inherits(object, "splm"))
    stopifnot(object$type %in% c("fixed effects lag"))
    if (!("spml2" %in% names(object)))
        stop("You must use spml2() to estimate the model.")
    # get type of effects (spfe / tpfe / sptpfe)
    effects <- object$effects
    # get cross-sectional and time ids used in estimation
    id_index <- object$spml2$index[1]
    ids <- sort(unique(object$spml2$data[[id_index]]))
    time_index <- object$spml2$index[2]
    periods <- sort(unique(object$spml2$data[[time_index]]))
    # get the listw
    if (is.null(listw))
        listw <- object$spml2$listw
    # get the input data
    if (is.null(newdata))
        newdata <- object$spml2$data
    id <- newdata[[id_index]]
    stopifnot(all(id %in% ids))
    stopifnot(all(ids %in% id))
    time <- newdata[[time_index]]
    stopifnot(all(time %in% periods))
    fml <- object$spml2$formula
    mt <- delete.response(terms(fml, data = newdata))
    # reorder data by id and time
    ord <- order(id, time)
    time <- time[ord]
    newdata <- newdata[ord, ]
    mm <- model.matrix(mt, newdata)[, -1]
    # get parameters
    # - estimate of spatial autocorrelation parameter
    if (object$type %in% c("fixed effects lag"))
        lambda <- as.numeric(coef(object)["lambda"])
    # - betas (estimates of ordinary parameters)
    beta <- matrix(coef(object)[-1], ncol = 1)
    # - fixed effects
    fixed_effects <- effects(object)
    #   - intercept
    intercept <- fixed_effects$INTTable
    #   - time effects
    time_effects <- fixed_effects$TETable
    #   - id effects
    id_effects <- fixed_effects$SETable
    # compute V(W)
    Wmat <- spdep::listw2mat(listw)
    N <- nrow(Wmat)  # number of gas stations
    VW <- 1
    if (object$type %in% c("fixed effects lag"))
        VW <- base::solve(base::diag(N) - lambda * Wmat)
    # predict for each time period separately (should give same results yet save memory)
    predictions <- purrr::map(
        unique(time),
        function(period) {
            X <- mm[time == period, ]
            prediction <- X %*% beta + intercept[1, 1]
            if (effects %in% c("tpfe", "sptpfe"))
                prediction <- prediction + time_effects[periods == period, 1]
            if (effects %in% c("spfe", "sptpfe"))
                prediction <- prediction + id_effects[, 1, drop = FALSE]
            prediction <- VW %*% prediction
            data.frame(id = 1:N, period = period, prediction = prediction)
        }
    )
    # join, restore the order, and return results
    predictions <- do.call(rbind, predictions)
    predictions <- predictions[order(predictions$id, predictions$period), ]
    predictions[match(seq_along(ord), ord), "prediction"]
}


# logLik.splm() is a method of logLik() function for splm objects; it's needed
# e.g. for AIC()
#
# inputs:
# - object ... (splm) a estimate by spml() function from splm package
#
# output:
# - object of logLik class
#
# note:
# - Presently it works only for SAR models. When any form of spatial error is
#   present, the resulting object has logLik slot but it is equal to NULL. I do
#   not know what it should mean.
logLik.splm <- function(object, ...) {
    if (is.null(object$logLik))
        stop("The object's logLik is NULL!")
    LL <- c(object$logLik)
    class(LL) <- "logLik"
    N <- length(residuals(object))
    attr(LL, "nall") <- N
    attr(LL, "nobs") <- N
    attr(LL, "df") <- length(object$coefficients)
    LL
}


# function formula.splm() is a formula methods for splm object; it works only
# for the objects etimated by spml2() wrapper
formula.splm <- function(object) {
    stopifnot(inherits(object, "splm"))
    if (!("spml2" %in% names(object)))
        stop("You must estimate the model with spml2().")
    object$spml2$formula
}


# An implementaion of predict for felm
# updated from predict.felm() by Sebastian Kranz's regtools package only
# available on github: https://github.com/skranz/regtools
# BEWARE!!! The original version from regtools does not work -- it throws
# a warning and gives wrong predictions
predict.felm <- function(object, newdata, use.fe = TRUE, ...) {
    # coefficient estimates
    co <- coef(object)
    # newdata made compatible with tibbles
    newdata = as.data.frame(newdata)
    # add row names
    rownames(newdata) <- seq_along(newdata[,1])
    # simplify the formula -- take only the stuff in front of first |
    form <- object %>%
        formula() %>%
        deparse() %>%
        str_c(collapse = " ") %>%
        str_extract("^[^\\|]*") %>%
        as.formula()
    # model matrix
    mf <- model.frame(form, newdata)
    mm <- model.matrix(form, data = mf)
    # check the data
    if (NROW(mm) < NROW(newdata))
        warning("Observations dropped from newdata due to NA.")
    # remove intercept if not included in coefficients
    if (NCOL(mm) == length(co) + 1)
        mm <- mm[ , -1, drop = FALSE]
    # prediction of demeaned values
    y.pred <- mm %*% co
    # add back estimates of fixed effects
    fe.vars <- names(object$fe)
    if (use.fe & length(fe.vars) > 0) {
        rows <- as.integer(rownames(mm))
        nd <- newdata[rows, ]
        all.fe <- getfe(object)
        for (fe.var in fe.vars) {
            # df = all.fe[all.fe$fe == fe.var,]
            # frows = match(nd[[fe.var]],df$idx)
            # myfe = df$effect[frows]
            # myfe[is.na(myfe)] = 0
            zz <- right_join(all.fe %>%
                                 dplyr::filter(fe == fe.var) %>%
                                 dplyr::select(idx, effect) %>%
                                 dplyr::mutate(idx = as.character(idx)),
                             nd %>%
                                 mutate_at(fe.var, as.character),
                             by = c("idx" = fe.var))
            y.pred <- y.pred + zz$effect
        }
    }
    # return
    as.vector(y.pred)
}


# lagsarlm2() is a wrapper for spatialreg::lagsarlm(); its inputs are subset of
# inputs of the original function; its output is similar to the output of the
# original function but it adds to it data necessary to make predictions
#
# inputs ... see help for formula, data, and listw in spatialreg::lagsarlm()
#
# outputs ... see help for spatialreg::lagsarlm()
#
# notes:
# - the output has to return the same class as lagsarlm()
lagsarlm2 <- function(formula, data, listw) {
    formula_name <- deparse(substitute(formula))
    data_name <- deparse(substitute(data))
    listw_name <- deparse(substitute(listw))
    pars <- list(formula = formula,
                 data = data,
                 listw = listw,
                 na.action = stats::na.omit)
    estimate <- do.call(spatialreg::lagsarlm, pars)
    estimate$call <- str2lang(str_c("lagsarlm(formula = ", formula_name,
                                    ", data = ", data_name,
                                    ", listw = ", listw_name,
                                    ", na.action = na.omit",
                                    ")"))
    estimate$sarlm2 <- list()
    estimate$sarlm2$listw <- listw
    estimate
}


# predict.sarlm() is a prediction methods for selected sarlm models; the model
# must be estimated by lagsarlm2() wrapper
#
# inputs:
# - object ... (sarlm) a model estimated by lagsarlm2()
# - newdata ... (data.frame with rownames, see help for
#   prepare_for_crosssectional_econometrics()) new data used for
#   prediction
#
# output:
# - (numeric vector) of predictions sorted as newdata
#
# warning:
# - the estimate must be produced by lagsarlm2() wrapper
predict.sarlm <- function(estimate, newdata) {
    if (!"sarlm2" %in% names(estimate))
        stop("You must estimate the cross-sectional SAR with lagsarlm2().")
    listw <- estimate$sarlm2$listw
    spatialreg::predict.sarlm(estimate, newdata, listw = listw) %>%
        as.numeric()
}


# function present_impacts() returns a nicely formated summapry for direct,
# indirect and total effects for impact estimate for a SAR mode
#
# inputs:
# - impacts ... a result of impacts() function
# - digits ... (optional) a number of decimal places to put in the table
#
# output:
# a tibble with the summary statistics formated
#
# usage:
# imp <- impacts(estimate, listw = W, time = 72, R = 1000)
# present_impacts(imp)
present_impacts <- function(impacts, digits = 3) {
    stopifnot(class(impacts) == "lagImpact")
    # get statistics for one kind of impact
    stats <- function(m, digits) {
        stat <- as.data.frame(m$statistics) %>%
            mutate(var = rownames(m$statistics),
                   val = format(round(Mean, digits), digits = digits),
                   sd = str_c("(",
                              format(round(SD, digits), digits = digits),
                              ")"),
                   sd = str_remove_all(sd, "\\s"))
        stat
    }
    # get stars form one kind of impact
    stars <- function(m) {
        quant <- m$quantiles
        stars <- case_when(quant[ , 1] * quant[ , 6] > 0 ~ "***",
                           quant[ , 2] * quant[ , 5] > 0 ~ "**",
                           quant[ , 3] * quant[ , 4] > 0 ~ "*",
                           TRUE ~ "    ")
        tibble(var = rownames(quant), stars = stars)
    }
    # get a column of statistcs and stars for one kind of statistcs
    one_impact <- function(m, name, digits) {
        stats(m, digits = digits) %>%
            left_join(stars(m), by = "var") %>%
            mutate(val = str_c(val, stars, " ", sd)) %>%
            select(var, val) %>%
            rename_at(2, ~name)
    }
    # significance levels
    lvls <- c(1, 5, 10)
    # get statistics with the usual significance levels
    impacts <- summary(impacts,
                       quant = sort(c(lvls / 2 / 100, 1 - lvls / 2 / 100)))
    # get and join statistics for all impacts and return
    one_impact(impacts$direct_sum, "direct", digits = digits) %>%
        left_join(one_impact(impacts$indirect_sum, "indirect", digits = digits),
                  by = "var") %>%
        left_join(one_impact(impacts$total_sum, "total", digits = digits),
                  by = "var") %>%
        rename(parameter = var)
}



# merger simulation ------------------------------------------------------------

# function update_competition_statistics() recalculates selected competition
# statistics within a given data; it is used in merger simulation: first we
# update the data on stations' ownership, and then we recalculte the competition
# statistics
#
# inputs:
# necessary inpus:
# - data ... (tibble) data prepared by prepare_average_fuel_data() where
#   competition statistics have been added by functions
#   add_spatial_clustering(), add_competitors_in_distances(), or
#   add_distance_to_closest_competitor(); it can be only a selection of this
#   data; however, if no stations are present (i.e. some were filtered out), the
#   full data must be supplied in full_data parameter; on the other hand,
#   dropping some time periods causes no harm
# - full_data ... (tibble) full data created by prepare_average_fuel_data()
#   where competition statistics have been added by functions
#   add_spatial_clustering(), add_competitors_in_distances(), or
#   add_distance_to_closest_competitor(); if not provided, it is assumed that
#   parameter data includes data on all stations, i.e. that no station has been
#   filtered out; if this is not the case, the competition statistics would be
#   incorrect
# - change_only_vars ... umožňuje omezit, jaká data se mají aktualizovat
# - existence ... (character scalar) the name of a column within both data and
#   full_data (when provided) which has TRUE for stations that exist in the
#   period and FALSE for stations that do not
# - same_group ... (tibble) a table stating which stations belong to a group; it
#   must have the following columns:
#   - from_date, to_date (character including a date in form "YYYY-MM-DD")
#   describing a period when a station name was owned by a group_name
#   - group_name ... (character) a name of the group
# - name ... (character) a name of a station owned by a group
#
# inputs that must be included only when used for the calculations:
# - arccr_folder ... (character scalar) a path to a folder where
#   administrative data from ARC ČR data (e.g. AdministrativniCleneni_v13.gdb)
#   resides
# - layer ... (character scalar) a name of layer used (the implicit value is
#   "StatPolygon")
# - driving_distances ... (tibble) a table including driving distances
#
# notes:
# - when the data consists of more than one period, you can improve the
#   performance of the function if you first run plan(multisession) to make it
#   run in parallel
#
# beware:
# - if data have only a selection of stations (i.e. not all station ids are
#   included), full_data must be provided and must include all station ids,
#   otherwise the competition statistics would be incorrect
# - several statistics are usually computed at the same time; their group is
#   always recalculated as a group; however, statistics not present in data are
#   dropped when the function returns
# - this function can only update the statistics created by the following
#   functions:
#   - add_spatial_clustering(),
#   - add_competitors_in_distances(), and
#   - add_distance_to_closest_competitor()
#   if any other competition statistics is used or it is renamed, you must
#   either update the code of this function, or update the statistics manually;
#   in the second case, you can't use simulate_merger() function but must
#   simulate a merger manually
#
# usage:
# - make it run in parallel
#   plan(multisession)
# - vzdálenost k nejbližšímu konkurentovi na výběru
#   test <- update_competition_statistics(
#       sel,
#       natural_monthly,
#       change_only_vars = "distance_to_closest_competitor",
#       existence = existence_criterium,
#       same_group = same_owners_correction_table,
#       driving_distances = station_driving_distances
#   )
# - sc na výběru
#   test <- update_competition_statistics(
#       sel,
#       natural_monthly,
#       change_only_vars = "sc",
#       existence = existence_criterium,
#       same_group = same_owners_correction_table,
#       arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb"
#   )
# - počet pump ve vzdálenosti na výběru
#   test <- update_competition_statistics(
#       sel,
#       natural_monthly,
#       change_only_vars = c("other_stations_in_driving_distance_0_3",
#                            "other_stations_in_driving_distance_3_6"),
#       existence = existence_criterium,
#       same_group = same_owners_correction_table,
#       driving_distances = station_driving_distances
#   )
# - všechny vlastnosti na výběru
# test <- update_competition_statistics(
#       sel,
#       natural_monthly,
#       existence = existence_criterium,
#       same_group = same_owners_correction_table,
#       driving_distances = station_driving_distances,
#       arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb"
#   )
update_competition_statistics <- function(data,
                                          full_data = NULL,
                                          change_only_vars = NULL,
                                          existence = NULL,
                                          same_group = NULL,
                                          arccr_folder = NULL,
                                          layer = "StatPolygon",
                                          driving_distances = NULL) {
    # check inputs
    if (!(is.list(change_only_vars) || inherits(change_only_vars, "formula") ||
          is.character(change_only_vars) || is.null(change_only_vars)))
        stop("change_only_vars must be an estimated model, formula or a character vector")
    # remember variables in data
    data_vars <- names(data)
    # full_data
    if (is.null(full_data))
        full_data <- data
    full_data <- dplyr::semi_join(full_data, data, by = c("from_date", "to_date"))
    # guess what competition statistics must be re-calculated
    if (is.null(change_only_vars))
        change_only_vars <- names(data)
    if (is.list(change_only_vars))
        change_only_vars <- formula(change_only_vars)
    if (inherits(change_only_vars, "formula"))
        change_only_vars <- as.character(change_only_vars)[3] %>%
        stringr::str_split("[^[:alnum:]_\\.]") %>%
        purrr::pluck(1) %>%
        stringr::str_subset("[[:alnum:]]")
    update_sc <- any(change_only_vars %in% c("sc", "cluster_area", "cluster_members_ids",
                                             "voronoi_neighbors_id"))
    update_stations_in_distance <-
        any(stringr::str_detect(change_only_vars,
                                "^(other|own)_stations_in_driving_distance_\\d+_\\d+"))
    update_distance_to_closest <-
        any(change_only_vars %in% c("distance_to_closest_competitor",
                                    "distance_to_closest_same_brand"))
    # re-calculate sc if needed
    if (update_sc) {
        message("Updating sc")
        full_data <- add_spatial_clustering(full_data,
                                            arccr_folder = arccr_folder,
                                            layer = layer,
                                            existence = existence,
                                            same_group = same_group)
    }
    # recalculate the number of stations within donuts if needed
    if (update_stations_in_distance) {
        message("Updating number of stations within donuts")
        dists <- change_only_vars %>%
            stringr::str_subset("^(other|own)_stations_in_driving_distance_\\d+_\\d+") %>%
            stringr::str_match("^(other|own)_stations_in_driving_distance_(\\d+)_(\\d+)")
        dists <- tibble::tibble(from = dists[, 3], to = dists[, 4]) %>%
            dplyr::distinct() %>%
            dplyr::mutate_all(as.numeric)
        for (k in seq_len(nrow(dists)))
            full_data <- add_competitors_in_distances(full_data,
                                                      driving_distances = driving_distances,
                                                      distances = c(dists$from[k], dists$to[k]),
                                                      existence = existence,
                                                      same_group = same_group,
                                                      quiet = TRUE)
    }
    # recalculate distances to the closest competitor and same-brand station
    if (update_distance_to_closest) {
        message("Updating distance to closest station")
        full_data <- add_distance_to_closest_competitor(full_data,
                                                        driving_distances = driving_distances,
                                                        existence = existence,
                                                        same_group = same_group)
    }
    # return a selection from full_data that corresponds to the original data
    full_data %>%
        dplyr::semi_join(data, by = c("id", "from_date", "to_date")) %>%
        dplyr::select(one_of(data_vars)) %>%
        dplyr::arrange(id, from_date)
}


# functions simulate_merger()
#
# inputs:
# - estimate ... (lm, sarlm, fixest, or splm) the model estimate
# - data ... (tibble) the data used to estimate the estimate; it must be an
#   output of prepare_average_fuel_data() with competition statistics added or
#   its subset; if some station ids have been filtered out then the full data
#   including all stations must be provided in full_data, or the competiton
#   statistics would be incorrect
# - full_data ... (tibble) full data from which data might be a subset; if data
#   are complete, full_data might be missing or NULL
# - periods ... (optional, Date) from_dates for which the estimation should be
#   done; you should set it for estimates on panel data, not on cross-sectional
#   ones because they should be done on one period anyway
# - takeovers ... table describing the takeovers that should be applied; it must
#   have one of the following two structures (see also the notes):
#     1. at least two columns:
#       - target ... (character scalars) the name of a brand that took over the
#         ids
#       - ids ... (list column of integer vectors) ... ids of stations taken
#         over by the brand
#     2. at least four columns:
#       - takeover_date ... (Date) the date when the merger took the place
#       - terminal_date ... (Date) the date when all brand names are correctly
#         inputed in the data and full_data
#       - source ... (character) the brand name of stations being taken over
#       - target ... (character) the brand name of stations that took source
#         over
# - existence ... (character scalar) the name of a column within both data and
#   full_data (when provided) which has TRUE for stations that exist in the
#   period and FALSE for stations that do not
# - same_group ... (tibble) a table stating which stations belong to a group;
#   it must have the following columns:
#   - from_date, to_date (character including a date in form "YYYY-MM-DD")
#     describing a period when a station name was owned by a group_name
#   - group_name ... (character) a name of the group
#   - name ... (character) a name of a station owned by a group
# - arccr_folder ... (character scalar) a path to a folder where administrative
#   data from ARC ČR data (e.g. AdministrativniCleneni_v13.gdb) resides
# - layer ... (character scalar) a name of layer used (the implicit value is
#   "StatPolygon")
# - driving_distances ... (tibble) a table including driving distances
# - full_impact ... (logical) if TRUE (the implicit value), both medium and full
#   predicted prices and price changes are calculated; if set to FALSE, only the
#   medium predicted price and price change are calculated
#
# output:
#
# notes:
# - takeover: if the tibble includes a column ids, only two columns are used:
#   the target and ids -- the name target is imputed to all stations with the
#   given ids; if ids columns is not present, four columns are expected:
#   from, to, source, and target; the ids is guessed from the data in such a
#   case, and then only target and ids is used
#
# beware:
# - if some ids have been filtered out from data, full_data must be supplied,
#   see above; however, time periods missing from data cause no harm
# - you must not change names of any variables produced by the functions
#   provided in this script, or this function may fail
# - the function recalculates competition statistics when simulating mergers;
#   however, it works only for the statistics computed by one of the following
#   functions (see documentationfor update_competition_statistics())
#     - add_spatial_clustering(),
#     - add_competitors_in_distances(), and
#     - add_distance_to_closest_competitor()
#   if you use any other statistic, you must adjust that function or simulate
#   the merger manually; to be sure, read the documentation for
#   update_competition_statistics()
# - the function also updates brand_name and brand_change if they are present in
#   data
#
# usage:
# takeovers <- tribble(
#   ~takeover_date, ~terminal_date, ~source, ~target,
#   "2014-08-01", "2018-12-31", "lukoil", "mol",
#   "2015-08-01", "2018-12-31", "agip", "mol",
#   "2016-02-01", "2018-12-31", "omv", "benzina",
#   "2016-02-01", "2018-12-31", "avanti", "benzina"
# ) %>%
#   mutate_at(c("takeover_date", "terminal_date"), as.Date)
simulate_merger <- function(estimate,
                            data,
                            full_data = NULL,
                            periods = NULL,
                            takeovers,
                            existence = NULL,
                            same_group = NULL,
                            arccr_folder = NULL,
                            layer = "StatPolygon",
                            driving_distances = NULL,
                            full_impact = TRUE) {
    # check inputs
    stopifnot(class(estimate) %in% c("fixest", "splm", "lm", "sarlm"))
    if (class(estimate) == "splm" && !("spml2" %in% names(estimate)))
        stop("You must estimate the model with spml2() wrapper.")
    if (class(estimate) == "sarlm" && !("sarlm2" %in% names(estimate)))
        stop("You must estimate the model with lagsarlm2() wrapper.")
    if ((class(estimate) %in% c("fixest", "splm")) && is.null(periods))
        stop("You must provide periods for fixest and splm estimates.")
    if (class(estimate) %in% c("lm", "sarlm")) {
        if (!is.null(periods))
            stop("You are not supposed to provide periods for lm and sarlm estimates.")
        periods <- unique(data$from_date)
        if (length(periods) > 1)
            stop("Data should be for one period for lm and sarlm estimates.")
    }
    # get full_data
    if (is.null(full_data))
        full_data <- data
    # if takeovers do not have ids, get them
    if (!("ids" %in% base::names(takeovers)))
        takeovers <- get_taken_over_ids(data, takeovers)  # full_data
    # store rownames of data -- it is necessary for sarlm prediction
    data_rownames <- tibble::tibble(id = data$id,
                                    from_date = data$from_date,
                                    ..rowname = base::rownames(data))
    # function to repair rownames in data -- necessary for sarlm prediction
    repair_rownames <- function(data, data_rownames) {
        data <- as.data.frame(left_join(data, data_rownames,
                                        by = c("id", "from_date")))
        base::rownames(data) <- data$..rowname
        data$..rowname <- NULL
        data
    }
    # filter prediction data
    data <- dplyr::filter(data, from_date %in% periods)
    full_data <- dplyr::filter(full_data, from_date %in% periods)
    stopifnot(nrow(data) > 0)
    stopifnot(base::isTRUE(
        dplyr::all_equal(data,
                         dplyr::semi_join(data, full_data,
                                          by = c("id", "from_date", "to_date")))))
    # 1. calculate in-sample prediction for the true data
    data <- repair_rownames(data, data_rownames)
    data$price_prediction_basic <- predict(estimate, newdata = data)
    # 2. change brand names, update competition statistics, and calculate the
    # second prediction
    full_data$.name_original <- full_data$name
    full_data$.merged_station <- FALSE
    for (k in seq_len(nrow(takeovers))) {
        full_data$name[full_data$id %in% takeovers$ids[[k]]] <- takeovers$target[k]
        full_data$.merged_station[full_data$id %in% takeovers$ids[[k]]] <- TRUE
    }
    full_data$.name_merger <- full_data$name
    full_data <- update_competition_statistics(data = full_data,
                                               change_only_vars = estimate,
                                               existence = existence,
                                               same_group = same_group,
                                               arccr_folder = arccr_folder,
                                               layer = "StatPolygon",
                                               driving_distances = driving_distances)
    full_data$name <- full_data$.name_original
    new_data <- dplyr::semi_join(full_data, data,
                                 by = c("id", "from_date", "to_date")) %>%
        dplyr::select(dplyr::any_of(names(data)))
    new_data <- repair_rownames(new_data, data_rownames)
    new_data$price_prediction_medium <- predict(estimate, newdata = new_data)
    data <- dplyr::left_join(data,
                             new_data %>%
                                 dplyr::select(id, from_date, to_date,
                                               price_prediction_medium),
                             by = c("id", "from_date", "to_date"))
    # 3. change brand_changes too and calculate the third prediction
    if (full_impact) {
        takeovers <- takeovers %>%
            dplyr::mutate(takeover_date = as.Date("0000-01-01"),
                          terminal_date = as.Date("5000-01-01"))
        full_data$name <- full_data$.name_merger
        if ("brand_change" %in% base::names(full_data)) {
            full_data <- add_brand_changes_in_takeover(full_data, takeovers,
                                                       keep_levels = TRUE,
                                                       keep_values = TRUE)
            if (any(is.na(full_data$brand_change)))
                warning("Some mergers are not covered in brand_change.")
            full_data$brand_change[is.na(full_data$brand_change)] <- "none"
        }
        if ("brand_name" %in% base::names(full_data)) {
            if (
                any(!(unique(c(takeovers$source, takeovers$target)) %in%
                      levels(data$brand_name)))
            )
                warning("Some mergers are not covered in brand_name.")
            full_data <- add_big_brand_name_effects(full_data,
                                                    levels(data$brand_name))
        }
        new_data <- dplyr::semi_join(full_data, data,
                                     by = c("id", "from_date", "to_date")) %>%
            dplyr::select(dplyr::any_of(names(data)))
        new_data <- repair_rownames(new_data, data_rownames)
        new_data$price_prediction_full <- predict(estimate, newdata = new_data)
        data <- dplyr::left_join(data,
                                 new_data %>%
                                     dplyr::select(id, from_date, to_date,
                                                   price_prediction_full),
                                 by = c("id", "from_date", "to_date"))
    } else {
        data$price_prediction_full <- NA
    }
    # calculate the price changes due to mergers and return
    data %>%
        tibble::as_tibble() %>%
        dplyr::left_join(full_data %>%
                             dplyr::select(id, from_date, to_date,
                                           .name_merger, .merged_station),
                         by = c("id", "from_date", "to_date")) %>%
        dplyr::mutate(merged_station = .merged_station,
                      name_after_merger = .name_merger,
                      price_change_medium =
                          price_prediction_medium - price_prediction_basic,
                      price_change_full =
                          price_prediction_full - price_prediction_basic) %>%
        dplyr::select(id, from_date, to_date, name, name_after_merger,
                      merged_station,
                      price_prediction_basic, price_prediction_medium,
                      price_prediction_full,
                      price_change_medium, price_change_full)
}


# boot_test_zero_mean(pc, n) runs bootstrap test that mean(pc) is significantly
# different from zero
#
# inputs:
# - pc ... (numeric vector)
# - n ... number of bootstraps
#
# output:
# - significance coding: * .. 0.1, ** .. 0.05, *** .. 0.01
boot_test_zero_mean <- function(pc, n = 1e4) {
    signif <- function(res, p) {
        q <- stats::quantile(res, c(p / 2, 1 - p / 2))
        q <- as.numeric(q)
        q[1] * q[2] > 0
    }
    res <- purrr::map_dbl(base::seq_len(n),
                          function(x) {
                              s <- base::sample(pc, base::length(pc),
                                                replace = TRUE)
                              base::mean(s)
                          })
    stringr::str_dup("*",
                     base::sum(purrr::map_lgl(c(0.1, 0.05, 0.01),
                                              ~signif(res, .))))
}


# summarize_merger() summarizes basic merger statistics
#
# inputs:
# - merger_data ... (tibble) table created by simulate_merger()
# - former_names ... (character vector) vector of brand names of stations that
#   belonged to the former group, i.e. the group that takes over other stations
# - var ... (character vector) name of variable (column) in merger_data for
#   which statistics should be computed
# - stats ... (named list of statistical functions that should be computed for
#   variables in var); implicitely:
#   - n ... number of observations in each group
#   - mean ... mean of observations in each group
#   - min ... min of observations in each group
#   - max ... max of observations in each group
#   - sd ... sd of observations in each group
#   - signif ... significance test that mean(var) is significantly different
#     from zero; test is done via bootstrap; confidence level coding is
#     * .. 0.1, ** .. 0.05, *** .. 0.01
# - simplify ... (logical scalar) if TRUE and length(var) == 1, it simplifies
#   column names of the results to names(stats)
#
# output:
# - tibble where rows correspond to groups of observations and columns to
#   computed statistics, see stats
#
# usage:
#   summarize_merger(mol_merger_fe, "mol", "price_change_full", simplify = TRUE)
#   summarize_merger(mol_merger_fe, "mol", c("price_change_full",
#                    "price_change_medium"))
summarize_merger <- function(merger_data, former_names, var,
                             stats = list(n = length,
                                          mean = mean,
                                          min = min,
                                          max = max,
                                          sd = sd,
                                          signif = boot_test_zero_mean),
                             simplify = FALSE) {
    g1 <- merger_data %>%
        dplyr::mutate(group = dplyr::case_when(merged_station ~ "purchased",
                                               name %in% former_names ~ "former",
                                               TRUE ~ "other"))
    g2 <- merger_data %>%
        dplyr::mutate(group = "all")
    data <- dplyr::bind_rows(g1, g2) %>%
        dplyr::mutate(group = base::factor(group,
                                           levels = c("former", "purchased",
                                                      "other", "all"))) %>%
        dplyr::group_by(group)
    res <- dplyr::summarize(data,
                            dplyr::across(tidyselect::all_of(var), stats),
                            .groups = "drop")
    if (simplify) {
        if (!(base::length(var) == 1))
            stop("I can simplify names only when calculated for just one variable.")
        base::names(res) <- c(base::names(res)[1], base::names(stats))
    }
    res
}
