# -------------------------------------
# Script: example_crosssectional_estimation_and_merger_simulations.R
# Author: Michal Kvasnička
# Purpose: This script shows how to
#          - estimate simple and spatial models on cross-sectional data
#          - use the model to simulate mergers
#          - visualize the merger results
#
# Copyright(c) Masaryk University
# -------------------------------------

# cesty
CODEDIR <- "R"
DATADIR <- "data"


# potřebné balíky a definované funkce
# library(spdep)
library(spatialreg)
library(ggplot2)
source(file.path(CODEDIR, "utilities.R"))
source(file.path(CODEDIR, "econometric_functions.R"))
source(file.path(CODEDIR, "preparatory_functions_general.R"))
source(file.path(CODEDIR, "graphing_functions.R"))





# příprava dat -----------------------------------------------------------------

# načti data
load(file.path(DATADIR, "pumpdroid_data.RData"))
station_driving_distances <- load_from(
    file.path(DATADIR, "station_driving_distances_fastest.RData"),
    "station_driving_distances"
)
station_beeline_distances <- load_from(
    file.path(DATADIR, "station_beeline_distances.RData"),
    "station_beeline_distances"
)
same_owners_correction_table <- load_from(
    file.path(DATADIR, "handmade_corrections_general.RData"),
    "same_owners_correction_table"
)


# kritérium existence nastav na 30 dnů (musí být v datech)
existence_criterium <- "station_exists_30"


# data velkých fúzí a převzetí
takeovers <- tribble(
    ~takeover_date, ~terminal_date, ~source, ~target,
    "2014-08-01", "2018-12-31", "lukoil", "mol",
    "2015-08-01", "2018-12-31", "agip", "mol",
    "2016-02-01", "2018-12-31", "omv", "benzina",
    "2016-02-01", "2018-12-31", "avanti", "benzina"
) %>%
    mutate_at(c("takeover_date", "terminal_date"), as.Date)


# základní odhadovaný model
basic_model <- price ~
    sc + distance_to_closest_competitor +
    other_stations_in_driving_distance_0_3 +
    other_stations_in_driving_distance_3_6 +
    # other_stations_in_driving_distance_6_9 +
    osm_highway +
    praha + brno + ostrava + mesto100 + mesto50 + mesto20 +
    brand_name +
    nonstop


# vyber období, kdy se odhadne model (a zároveň se zde bude simulovat merger); z
# dat je potřeba vyhodit NA na používaných proměnných a je také potřeba data
# převést na data.frame s pojmenovanými řádky -- jména řádků musejí nést
# informace o id stanic
# do plných dat je třeba přidat faktor pro zvolené velké značky
obdobi <- "2014-01-01"
big_brands <- natural_monthly %>%
    filter(from_date == obdobi) %>%
    find_big_brands_in_prepared_data() %>%
    pull(name)
natural_monthly <- natural_monthly %>%
    mutate(price = price * 100) %>%
    add_big_brand_name_effects(big_brands) %>%
    add_city_effects()  # TODO: vyhodit, až přegeneruje natural_monthly
cs_data <- natural_monthly %>%
    filter(from_date == obdobi) %>%
    prepare_for_crosssectional_econometrics(existence_criterium, basic_model)






# odhady průřezových modelů ----------------------------------------------------

# odhad OLS
estimate_ols <- lm(basic_model, cs_data)
summary(estimate_ols)

# odhad SAR
W <- create_weights(cs_data, neighbors = "beeline", max_dist = 20)
estimate_sar <- lagsarlm2(basic_model,
                          data = cs_data,
                          listw = W)
summary(estimate_sar)





# grafy reziduí ----------------------------------------------------------------

# víc možností viz example_panel_estimation_and_merger_simulations.R

# pro interaktivní graf zrušte komentář následujícího řádku
# tmap_mode("view")

# průměrná rezidua v jednotlivých okresech pro jedno období
cs_data %>%
    plot_average_residuals(estimate_ols,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")
cs_data %>%
    plot_average_residuals(estimate_sar,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")


# rezidua pro jedno období po jednotlivých stanicích -- větší puntíky
cs_data %>%
    plot_station_residuals(estimate_ols,
                           natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy",
                           size = 0.08)
cs_data %>%
    plot_station_residuals(estimate_sar,
                           natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy",
                           size = 0.08)





# simulace fúze ----------------------------------------------------------------

# simulace mergeru MOL: Mol převzal Lukoil v srpnu 2014 a Agip v srpnu 2015
# simulace pro OLS a SAR na průřezových datech
mol_merger_ols <- simulate_merger(
    estimate_ols,
    cs_data,
    natural_monthly,
    # periods = unique(cs_data$from_date),
    takeovers = takeovers[1:2, ],
    existence = existence_criterium,
    same_group = same_owners_correction_table,
    arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
    driving_distances = station_driving_distances
)
mol_merger_sar <- simulate_merger(
    estimate_sar,
    cs_data,
    natural_monthly,
    # periods = unique(cs_data$from_date),
    takeovers = takeovers[1:2, ],
    existence = existence_criterium,
    same_group = same_owners_correction_table,
    arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
    driving_distances = station_driving_distances
)





# vizualizace fúze -------------------------------------------------------------

# histogramy očekávaných změn cen v důsledku fůze
ggplot(mol_merger_ols, aes(price_change_full)) + geom_histogram()
ggplot(mol_merger_sar, aes(price_change_full)) + geom_histogram()


# očekávané změny cen v grafu po jednotlivých stanicích pro simulaci SAR
# - pro interaktivní graf odkomentujte následující řádek
# tmap_mode("view")
# - skutečná změna ceny, tj. včetně změn cenových úrovní převzatých stanic
plot_station_merger_prices(mol_merger_sar, natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy")
# - změna ceny jen v důsledku změn konkurence, tj. beze změny cenových úrovní
#   převzatých stanic
plot_station_merger_prices(mol_merger_sar, natural_monthly,
                           "price_change_medium",
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy")


# průměrné očekávané změny cen v jednotlivých okresech
plot_average_merger_prices(mol_merger_sar, natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")

# průměrné očekávané změny cen v jednotlivých okresech
plot_average_merger_prices(mol_merger_sar, natural_monthly,
                           var = "price_change_medium",
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")


# základní statistiky mergeru podle typů stanic
# - plná změna ceny pro simulaci pomocí modelu SAR
summarize_merger(mol_merger_sar, c("mol", "papoil", "slovnaft"),
                 "price_change_full", simplify = TRUE)
# - plná změna i změna ceny způsobená jen změnami konkurence pro simulace OLS
summarize_merger(mol_merger_ols,  c("mol", "papoil", "slovnaft"),
                 c("price_change_full", "price_change_medium"))

# POZNÁMKA: Proč vychází simulace na průřezových datech výrazně jinak než
# simulace pro stejné období na panelových datech?
#
# Rozdíl se týká cenových hladin jednotlivých čerpacích stanic. Změna cen jen v
# důsledku změny konkurence vychází v obou případech podobně -- ceny všude
# rostou. V panelovém odhadu je cena Mol vyšší než cena kupovaných stanic, a v
# důsledku toho ceny na kupovaných stanicích rostou (a v SARu se tento růst dál
# rozlije). Naproti tomu v odhadu průřezových dat je dummy proměnná pro Mol sice
# o něco vyšší než pro Lukoil, ale výrazně nižší než pro Agip. Protože Agipů Mol
# koupil víc než Lukoilů, cena na mnoha kupovaných stanicích klesla (a tento
# pokles se v SARu dál rozlil).





# uložení hlavních spočítaných hodnot ------------------------------------------

save(cs_data, big_brands, estimate_ols, estimate_sar, mol_merger_ols,
     mol_merger_sar,
     file = "data/cross_sectional_econometrics_and_merger_simulation.RData")
