# -------------------------------------
# Script: example_panel_estimation_and_merger_simulations.R
# Author: Michal Kvasnička
# Purpose: This script shows how to
#          - estimate simple and spatial models on panel data
#          - use the model to simulate mergers
#          - visualize the merger results
#
# Copyright(c) Masaryk University
# -------------------------------------

# cesty
CODEDIR <- "R"
DATADIR <- "data"


# potřebné balíky a definované funkce
library(tibble)
library(dplyr)
library(fixest)
library(splm)
library(lmtest)
library(Formula)
library(ggplot2)
source(file.path(CODEDIR, "utilities.R"))
source(file.path(CODEDIR, "preparatory_functions_general.R"))
source(file.path(CODEDIR, "econometric_functions.R"))
source(file.path(CODEDIR, "graphing_functions.R"))





# příprava dat -----------------------------------------------------------------

# načti data
load(file.path(DATADIR, "pumpdroid_data.RData"))
station_driving_distances <- load_from(
    file.path(DATADIR, "station_driving_distances_fastest.RData"),
    "station_driving_distances"
)
station_beeline_distances <- load_from(
    file.path(DATADIR, "station_beeline_distances.RData"),
    "station_beeline_distances"
)
same_owners_correction_table <- load_from(
    file.path(DATADIR, "handmade_corrections_general.RData"),
    "same_owners_correction_table"
)


# převeď ceny na haléře
natural_monthly <- natural_monthly %>%
    mutate(price = price * 100)


# kritérium existence nastav na 30 dnů (musí být v datech)
existence_criterium <- "station_exists_30"


# data velkých fúzí a převzetí
takeovers <- tribble(
    ~takeover_date, ~terminal_date, ~source, ~target,
    "2014-08-01", "2018-12-31", "lukoil", "mol",
    "2015-08-01", "2018-12-31", "agip", "mol",
    "2016-02-01", "2018-12-31", "omv", "benzina",
    "2016-02-01", "2018-12-31", "avanti", "benzina"
) %>%
    mutate_at(c("takeover_date", "terminal_date"), as.Date)


# přidej sloupec brand_change, který indikuje vybranou fúzi
natural_monthly_original <- natural_monthly  # TODO: je potřeba?
natural_monthly <- add_brand_changes_in_takeover(natural_monthly, takeovers)


# základní odhadovaný model
basic_model <- price ~ brand_change +
    sc +
    other_stations_in_driving_distance_0_3 +
    other_stations_in_driving_distance_3_6 +
    other_stations_in_driving_distance_6_9 +
    distance_to_closest_competitor +
    nonstop


# vyber data za roky, kde jsou data dostupná a použitelná (zde 2013--2018) a
# vytvoř z nich balancovaný panel (to je potřeba jen pro spatial model, ale
# stejná data zde použijeme pro všechny odhady); vybírají se jen proměnné v
# selekční rovnici
panel_data <- natural_monthly %>%
    filter(from_date >= "2013-01-01", from_date <= "2018-12-01") %>%
    prepare_for_spatial_panel_econometrics(existence_criterium, basic_model)




# odhady panelových modelů -----------------------------------------------------

# odhad modelu s fixními efekty na id a from_date (cena v halířích)
# POZOR: v současné době je chyba ve "stabilní" verzi fixest 0.7.x na CRANu
# je potřeba nainstalovat dev verzi 0.8 z GitHubu:
# devtools::install_github("lrberge/fixest")
update_fe <- function(m) update(as.Formula(m), . ~ . | id + from_date)
basic_model_fe <- update_fe(basic_model)
estimate_fe <- feols(basic_model_fe, data = panel_data)
summary(estimate_fe)


# odhad spatial modelu
W <- create_weights(panel_data, neighbors = "beeline", max_dist = 20)
estimate_sar <- spml2(basic_model,
                      data = panel_data,
                      index = c("id", "from_date"),
                      listw = W,
                      model = "within",
                      effect = "twoways",
                      lag = TRUE,
                      spatial.error = "none")
summary(estimate_sar)





# odhad přímých a nepřímých vlivů ----------------------------------------------
# POZOR: vyžaduje hodně paměti (aspoň 24 GB)
imp <- impacts(estimate_sar, listw = W, time = 72, R = 1000)
present_impacts(imp)





# grafy reziduí ----------------------------------------------------------------

# pro interaktivní graf zrušte komentář následujícího řádku
# tmap_mode("view")

# průměrná rezidua v jednotlivých okresech pro jedno období
panel_data %>%
    filter(from_date == "2018-01-01") %>%
    plot_average_residuals(estimate_sar,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")

# medián reziduí v jednotlivých okresech pro jedno období
panel_data %>%
    filter(from_date == "2018-01-01") %>%
    plot_average_residuals(estimate_sar,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           stat = median)

# průměrná rezidua pro jednotlivé okresy pro dvě období -- facetovaný graf
panel_data %>%
    filter(from_date == c("2018-01-01", "2018-06-01")) %>%
    plot_average_residuals(estimate_sar,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           facet = TRUE)

# rezidua pro jedno období po jednotlivých stanicích -- větší puntíky
panel_data %>%
    filter(from_date == "2018-01-01") %>%
    plot_station_residuals(estimate_sar,
                           natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy",
                           size = 0.08)

# uložení grafu -- dokumentace funkce tmap::tmap_save()
p <- panel_data %>%
    filter(from_date == "2018-01-01") %>%
    plot_average_residuals(estimate_sar,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")
tmap_save(p, filename = "fig01.png")





# simulace fúze ----------------------------------------------------------------

# simulace mergeru MOL: Mol převzal Lukoil v srpnu 2014 a Agip v srpnu 2015
# simulace pro FE a SAR
period <- as.Date("2014-01-01")  # měsíc, ke kterému se simulace provádí
mol_merger_fe <- simulate_merger(
    estimate_fe,
    panel_data,
    natural_monthly,
    periods = period,
    takeovers = takeovers[1:2, ],
    existence = existence_criterium,
    same_group = same_owners_correction_table,
    arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb",
    driving_distances = station_driving_distances
)
mol_merger_sar <- simulate_merger(
    estimate_sar,
    panel_data,
    natural_monthly,
    periods = period,
    takeovers = takeovers[1:2, ],
    existence = existence_criterium,
    same_group = same_owners_correction_table,
    arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb",
    driving_distances = station_driving_distances
)




# vizualizace fúze -------------------------------------------------------------

# histogramy očekávaných změn cen v důsledku fůze
ggplot(mol_merger_fe, aes(price_change_full)) + geom_histogram()
ggplot(mol_merger_sar, aes(price_change_full)) + geom_histogram()


# očekávané změny cen v grafu po jednotlivých stanicích pro simulaci SAR
# - pro interaktivní graf odkomentujte následující řádek
# tmap_mode("view")
# - skutečná změna ceny, tj. včetně změn cenových úrovní převzatých stanic
plot_station_merger_prices(mol_merger_sar, natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy")
# - změna ceny jen v důsledku změn konkurence, tj. beze změny cenových úrovní
#   převzatých stanic
plot_station_merger_prices(mol_merger_sar, natural_monthly,
                           "price_change_medium",
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy")


# průměrné očekávané změny cen v jednotlivých okresech
plot_average_merger_prices(mol_merger_sar, natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")


# základní statistiky mergeru podle typů stanic
# - plná změna ceny pro simulaci pomocí modelu SAR
summarize_merger(mol_merger_sar, c("mol", "papoil", "slovnaft"),
                 "price_change_full", simplify = TRUE)
# - plná změna i změna ceny způsobená jen změnami konkurence pro simulace FE
summarize_merger(mol_merger_fe,  c("mol", "papoil", "slovnaft"),
                 c("price_change_full", "price_change_medium"))




# uložení hlavních spočítaných hodnot ------------------------------------------

save(panel_data, estimate_fe, estimate_sar, imp, mol_merger_fe, mol_merger_sar,
     file = "data/panel_econometrics_and_merger_simulation.RData")
