# -------------------------------------
# Script: R/example_panel_estimation_and_merger_simulations_time_fixed_effects_only.R
# Author: Michal Kvasnička
# Purpose: This script shows how to
#          - estimate simple and spatial models on panel data with fixed effects
#            on time only
#          - use the model to simulate mergers
#          - use the model to simulate divestitures in mergers
#
# Copyright(c) Masaryk University
# -------------------------------------

# cesty
CODEDIR <- "R"
DATADIR <- "data"


# potřebné balíky a definované funkce
library(tibble)
library(dplyr)
library(fixest)
library(splm)
library(lmtest)
library(Formula)
library(ggplot2)
source(file.path(CODEDIR, "utilities.R"))
source(file.path(CODEDIR, "preparatory_functions_general.R"))
source(file.path(CODEDIR, "econometric_functions.R"))
source(file.path(CODEDIR, "graphing_functions.R"))





# příprava dat -----------------------------------------------------------------

# načti data
load(file.path(DATADIR, "pumpdroid_data.RData"))
station_driving_distances <- load_from(
    file.path(DATADIR, "station_driving_distances_fastest.RData"),
    "station_driving_distances"
)
station_beeline_distances <- load_from(
    file.path(DATADIR, "station_beeline_distances.RData"),
    "station_beeline_distances"
)
same_owners_correction_table <- load_from(
    file.path(DATADIR, "handmade_corrections_general.RData"),
    "same_owners_correction_table"
)


# převeď ceny na haléře
natural_monthly <- natural_monthly %>%
    mutate(price = price * 100)


# kritérium existence nastav na 30 dnů (musí být v datech)
existence_criterium <- "station_exists_30"


# data velkých fúzí a převzetí
takeovers <- tribble(
    ~takeover_date, ~terminal_date, ~source, ~target,
    "2014-08-01", "2018-12-31", "lukoil", "mol",
    "2015-08-01", "2018-12-31", "agip", "mol",
    "2016-02-01", "2018-12-31", "omv", "benzina",
    "2016-02-01", "2018-12-31", "avanti", "benzina"
) %>%
    mutate_at(c("takeover_date", "terminal_date"), as.Date)


# přidej sloupec brand_change, který indikuje vybranou fúzi, brand_names, který
# indikuje velké značky
big_brands <- map(unique(natural_monthly$from_date),
    ~find_big_brands_in_prepared_data(filter(natural_monthly, from_date == .))) %>%
    bind_rows() %>%
    pull(name) %>%
    unique()
natural_monthly <- natural_monthly %>%
    add_brand_changes_in_takeover(takeovers) %>%
    add_brandnames(big_brands) %>%
    add_city_effects()



# základní odhadovaný model
basic_model <- price ~ brand_change + brand_name +
    sc +
    other_stations_in_driving_distance_0_3 +
    other_stations_in_driving_distance_3_6 +
    other_stations_in_driving_distance_6_9 +
    distance_to_closest_competitor +
    osm_highway +
    praha + brno + ostrava + mesto100 + mesto50 + mesto20 +
    nonstop


# vyber data za roky, kde jsou data dostupná a použitelná (zde 2013--2018) a
# vytvoř z nich balancovaný panel (to je potřeba jen pro spatial model, ale
# stejná data zde použijeme pro všechny odhady); vybírají se jen proměnné v
# selekční rovnici
panel_data <- natural_monthly %>%
    filter(from_date >= "2013-01-01", from_date <= "2018-12-01") %>%
    prepare_for_spatial_panel_econometrics(existence_criterium, basic_model)




# odhady panelových modelů -----------------------------------------------------

# odhad modelu s fixními efekty na id a from_date (cena v halířích)
# POZOR: v současné době je chyba ve "stabilní" verzi fixest 0.7.x na CRANu
# je potřeba nainstalovat dev verzi 0.8 z GitHubu:
# devtools::install_github("lrberge/fixest")
update_fe <- function(m) update(as.Formula(m), . ~ . | from_date)
basic_model_fe <- update_fe(basic_model)
estimate_fe2 <- feols(basic_model_fe, data = panel_data)
summary(estimate_fe2)


# odhad spatial modelu
W <- create_weights(panel_data, neighbors = "beeline", max_dist = 20)
estimate_sar2 <- spml2(basic_model,
                      data = panel_data,
                      index = c("from_date"),
                      listw = W,
                      model = "within",
                      effect = "time",
                      lag = TRUE,
                      spatial.error = "none",
                      index_id = "id")
summary(estimate_sar2)





# odhad přímých a nepřímých vlivů ----------------------------------------------
# POZOR: vyžaduje hodně paměti (aspoň 24 GB)
# imp <- impacts(estimate_sar2, listw = W, time = 72, R = 1000)
# present_impacts(imp)






# simulace fúze ----------------------------------------------------------------

# simulace hypotetické fúze, ve které by Benzina koupila všechny stanice
# Eurooilu a bude je provozovat pod svou značkou
# - období, ve kterém se fúze simuluje
fake_period <- as.Date("2018-01-01")
# - zjištění id čerpacích stanic Eurooil a tvorba tabulky fúzí
eurooil <- natural_monthly %>%
    filter(from_date == fake_period, name == "eurooil") %>%
    pull(id)
fakeovers <- tribble(
    ~takeover_date, ~terminal_date, ~source, ~target, ~ids,
    NA, NA, "eurooil", "benzina", ids = eurooil
)
# - simulace fúze
benzina_merger_sar <- simulate_merger(
    estimate_sar2,
    panel_data,
    natural_monthly,
    periods = fake_period,
    takeovers = fakeovers,
    existence = existence_criterium,
    same_group = same_owners_correction_table,
    arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb",
    driving_distances = station_driving_distances
)






# vizualizace fúze -------------------------------------------------------------

# histogramy očekávaných změn cen v důsledku fůze
ggplot(benzina_merger_sar, aes(price_change_full)) + geom_histogram()


# očekávané změny cen v grafu po jednotlivých stanicích pro simulaci SAR
# - pro interaktivní graf odkomentujte následující řádek
# tmap_mode("view")
# - skutečná změna ceny, tj. včetně změn cenových úrovní převzatých stanic
plot_station_merger_prices(benzina_merger_sar, natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy")
# - změna ceny jen v důsledku změn konkurence, tj. beze změny cenových úrovní
#   převzatých stanic
plot_station_merger_prices(benzina_merger_sar, natural_monthly,
                           "price_change_medium",
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
                           regions = "okresy")


# průměrné očekávané změny cen v jednotlivých okresech
plot_average_merger_prices(benzina_merger_sar, natural_monthly,
                           arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")


# základní statistiky mergeru podle typů stanic
# - plná změna ceny pro simulaci pomocí modelu SAR
summarize_merger(benzina_merger_sar, c("mol", "papoil", "slovnaft"),
                 "price_change_full", simplify = TRUE)
# - plná změna i změna ceny způsobená jen změnami konkurence pro simulace SAR
summarize_merger(benzina_merger_sar,  c("mol", "papoil", "slovnaft"),
                 c("price_change_full", "price_change_medium"))






# simulace závazku -------------------------------------------------------------

# simulace vlivu závazku, když 4 stanice zůstanou Eurooil
# - pro jednoduchost vezmeme 4 stanice, kde cena roste nejvíc (všechny patří
#   Eurooilu)
benzina_merger_sar %>%
    filter(price_change_full > 35) %>%
    arrange(desc(price_change_full)) %>%
    select(id, name, merged_station, price_change_full, price_change_medium) %>%
    slice(1:4)
eurooil_exception <- benzina_merger_sar %>%
    arrange(desc(price_change_full)) %>%
    slice(1:4) %>%
    pull(id)
# - řekněme, že zůstanou pod značkou Eurooil
eurooil_taken_over <- setdiff(eurooil, eurooil_exception)
fakeover_exception <- tribble(
    ~takeover_date, ~terminal_date, ~source, ~target, ~ids,
    NA, NA, "eurooil", "benzina", ids = eurooil_taken_over
)
# - simulace fúze
benzina_merger_sar_exceptions <- simulate_merger(
    estimate_sar2,
    panel_data,
    natural_monthly,
    periods = fake_period,
    takeovers = fakeover_exception,
    existence = existence_criterium,
    same_group = same_owners_correction_table,
    arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb",
    driving_distances = station_driving_distances
)
# - vizualizace rozdílů
dat <- tibble(change = benzina_merger_sar_exceptions$price_change_full -
                  benzina_merger_sar$price_change_full)
ggplot(dat, aes(change)) + geom_histogram()
ggplot(dat %>% filter(change < -1), aes(change)) + geom_histogram()

comp <- left_join(benzina_merger_sar %>%
                      select(id, name, price_chage_merger = price_change_full),
                  benzina_merger_sar_exceptions %>%
                      select(id, price_change_exceptions = price_change_full),
                  by = "id") %>%
    mutate(difference = price_change_exceptions - price_chage_merger) %>%
    arrange(difference)


# simulace závazku, kdy tyto 4 stanice převezme OMV
fakeover_exception2 <- tribble(
    ~takeover_date, ~terminal_date, ~source, ~target, ~ids,
    NA, NA, "eurooil", "benzina", eurooil_taken_over,
    NA, NA, "eurooil", "omv", eurooil_exception
)
# - simulace fúze
benzina_merger_sar_exceptions2 <- simulate_merger(
    estimate_sar2,
    panel_data,
    natural_monthly,
    periods = fake_period,
    takeovers = fakeover_exception2,
    existence = existence_criterium,
    same_group = same_owners_correction_table,
    arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb",
    driving_distances = station_driving_distances
)
# - vizualizace rozdílů
comp2 <- left_join(benzina_merger_sar %>%
                      select(id, name, price_chage_merger = price_change_full),
                  benzina_merger_sar_exceptions2 %>%
                      select(id, price_change_exceptions = price_change_full),
                  by = "id") %>%
    mutate(difference = price_change_exceptions - price_chage_merger) %>%
    arrange(difference)
ggplot(comp2, aes(difference)) + geom_histogram()
ggplot(comp2 %>% filter(abs(difference) > 1), aes(difference)) + geom_histogram()
comp2






# uložení hlavních spočítaných hodnot ------------------------------------------

save(panel_data, estimate_fe2, estimate_sar2, benzina_merger_sar,
     benzina_merger_sar_exceptions, benzina_merger_sar_exceptions2,
     file = "data/panel_econometrics_and_merger_simulation_time_effects_only.RData")
