# -------------------------------------
# Script: graphing_functions.R
# Author: Michal Kvasnička
# Purpose: This script contains functions used for drawing various charts.
# Notes: For usage see script example_plotting.
#
# Copyright(c) Masaryk University
# -------------------------------------

library(tmap)
library(dplyr)
library(stringr)
library(sf)


# check_plotting_parameters() checks plotting parameters, see its usage in
# plot_average_prices() and plot_station_prices()
check_plotting_parameters <- function(data, regions,
                                      extra_region_option = character(0)) {
    if (!base::all(c("id", "from_date") %in% base::names(data)))
        stop("Data are not valid.")
    if (!(regions %in% c(extra_region_option, "obce", "okresy", "kraje", "cr")))
        stop("Regions must be one of",
             dplyr::if_else(length(extra_region_option) == 0,
                            "",
                            stringr::str_c(" ", extra_region_option, ",")),
             " 'obce', 'okresy', 'kraje', or 'cr'.")
}




# plot prices ------------------------------------------------------------------

# plot_average_prices() plots average prices for given regions and periods of
# time
#
# inputs:
# - data ... (tibble) aggregated data created by prepare_average_fuel_data()
# - arccr_folder ... (character scalar) path to "AdministrativniCleneni_v13.gdb"
#   folder
# - layer ... (optional character scalar) layer from arccr_folder used; if not
#   given, the function selects the appropriate one
# - regions ... (one of "obce", "okresy", "kraje", and "cr"; default is
#   "okresy") for which kind of regions the average prices should be plotted
# - existence ... (optional character scalar) name of logical column in data
#   which indicates whether the station existed in the given period; if not
#   given, it is assumed that all stations exist
# - stat ... (function, default is mean) statistics computed for each region
# - facet ... (logical, default is FALSE) if TRUE, the plots are facetted by
#   from_date; if FALSE (default), the average prices are computed over all
#   periods
# - facet_col, facet_row ... (integer scalar) how many columns / rows should be
#   in the facetted plot; if not given, tmap chooses some reasonable defaults
# - legend_title ... (character scalar) if given, it is used as a title of the
#   legend; if not given, the title is composed of the name of the stat function
#   and "prices"
# - .info_type ... (character scalar) means for wrappers to interact with this
#   functions; please, do no play with it
# - plot_labels ... (optional logical scalar, default is FALSE) if TRUE, prices
#   are plotted in the regions
# - ... ... other parameters sent to tm_polygons()
#
# output:
# - object of class "tmap"; if it is not stored in a variable, it is plotted;
#   it can be joined with other tmap objects
#
# details:
# - user should pre-filter the data to include only the periods which (s)he
#   wants to see in the chart
# - when several periods are selected, the prices are averaged over the periods
#   and regions; the average is computed only from existing stations with price
#   different from NA -- unless facet = TRUE
# - if facet = TRUE, the plot is facetted
# - in interactive mode, additional information on regions are available on
#   clicking
#
# usage (see example_plotting.R):
#   natural_monthly %>%
#       filter(from_date == c("2013-01-01", "2014-01-01")) %>%
#       plot_average_prices("map_shapes/AdministrativniCleneni_v13.gdb/",
#           existence = "station_exists_30")
#   natural_monthly %>%
#       filter(from_date == c("2013-01-01", "2014-01-01")) %>%
#       plot_average_prices("map_shapes/AdministrativniCleneni_v13.gdb/",
#           existence = "station_exists_30",
#           stat = median,
#           facet = TRUE)
plot_average_prices <- function(data,
                                arccr_folder, layer = NULL,
                                regions = "okresy",
                                existence = NULL,
                                stat = base::mean,
                                facet = FALSE,
                                facet_col = NA,
                                facet_row = NA,
                                legend_title = NULL,
                                plot_labels = FALSE,
                                .info_type = "price",
                                ...) {
    # check and fill in inputs
    check_plotting_parameters(data, regions)
    if (is.null(existence)) {
        existence <- TRUE
    } else {
        existence <- data[[existence]]
    }

    # get regional codes and names
    if (!is.null(layer)) {
        layer <- layer
    } else {
        layer <- dplyr::case_when(regions == "obce" ~ "ObcePolygony",
                                  regions == "okresy" ~ "OkresyPolygony",
                                  regions == "kraje" ~ "KrajePolygony",
                                  regions == "cr" ~ "StatPolygon")
    }
    region <- dplyr::case_when(regions == "obce" ~ "KOD_OBEC",
                               regions == "okresy" ~ "KOD_OKRES",
                               regions == "kraje" ~ "KOD_KRAJ",
                               regions == "cr" ~ "NAZ_STAT")
    region_name <- dplyr::case_when(regions == "obce" ~ "NAZ_OBEC",
                                    regions == "okresy" ~ "NAZ_LAU1",
                                    regions == "kraje" ~ "NAZ_CZNUTS3",
                                    regions == "cr" ~ "NAZ_STAT")

    # filter only existing stations
    data <- data[existence, ]

    # add shapes
    # I have to remove geometry from reg_data and then set it back because
    # summarize is extremly slow on sf while it is prettey fast on tibbles
    data <- data %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326, remove = FALSE) %>%
        sf::st_transform(5514)
    regions <- sf::st_read(arccr_folder, layer,
                           stringsAsFactors = FALSE,
                           quiet = TRUE) %>%
        sf::st_transform(5514)
    reg_data <- sf::st_join(regions, data) %>%
        sf::st_drop_geometry()

    # get average price for each region
    remove_nas <- function(x) x[!is.na(x)]
    reg_data <- reg_data %>%
        dplyr::group_by(!!rlang::sym(region))
    if (facet)
        reg_data <- reg_data %>%
        dplyr::group_by(from_date, .add = TRUE)
    reg_data <- reg_data %>%
        dplyr::summarize(name = unique(!!rlang::sym(region_name)),
                         price = stat(remove_nas(price)),
                         stations = base::length(base::unique(id)),
                         .groups = "drop") %>%
        dplyr::mutate(text = stringr::str_c(round(price, 1)))
    reg_data <- left_join(regions %>% select(all_of(region)), reg_data,
                          by = region)

    # return the plot
    if (is.null(legend_title)) {
        fname <- base::as.character(base::substitute(stat))
        fname <- fname[base::length(fname)]
        legend_title <- stringr::str_c(fname, " prices")
    }
    popup_vars <- c("price", "stations")
    if (.info_type == "residuals") {
        reg_data$residual <- reg_data$price
        popup_vars <- c("residual", "stations")
    }
    if (.info_type == "merger") {
        reg_data$merger_stat <- reg_data$price
        popup_vars <- c("merger_stat", "stations")
    }
    p <- tmap::tm_shape(reg_data) +
        tmap::tm_polygons("price",
                          id = "name",
                          popup.vars = popup_vars,
                          title = legend_title,
                          ...)
    if (facet)
        p <- p + tmap::tm_facets(by = "from_date",
                                 ncol = facet_col, nrow = facet_row)
    if (plot_labels)
        p <- p + tmap::tm_text("text", style = "pretty")
    p
}


# plot_station_prices() plots average prices for individual stations and periods
# of time
#
# inputs:
# - data ... (tibble) aggregated data created by prepare_average_fuel_data()
# - arccr_folder ... (character scalar) path to "AdministrativniCleneni_v13.gdb"
#   folder; it is necessary only when regions is not "nic"
# - layer ... (optional character scalar) layer from arccr_folder used; if not
#   given, the function selects the appropriate one; it is necessary only when
#   regions is not "nic"
# - regions ... (one of "nic", "obce", "okresy", "kraje", and "cr"; default is
#   "nic") if not "nic" (which is default), the selected regions' border are
#   plotted
# - existence ... (optional character scalar) name of logical column in data
#   which indicates whether the station existed in the given period; if not
#   given, it is assumed that all stations exist
# - stat ... (function, default is mean) statistics computed for each station
# - facet ... (logical, default is FALSE) if TRUE, the plots are facetted by
#   from_date; if FALSE (default), the average prices are computed over all
#   periods
# - facet_col, facet_row ... (integer scalar) how many columns / rows should be
#   in the facetted plot; if not given, tmap chooses some reasonable defaults
# - plot_regional_averages ... (logical scalar, default is FALSE) if TRUE,
#   regions' polygons are filled with the regions' average prices
# - regional_averages_alpha ... (double between 0 and 1) if
#   plot_regional_averages = TRUE, regions' polygons alpha (opacity)
# - legend_title ... (character scalar) if given, it is used as a title of the
#   legend; if not given, the title is composed of the name of the stat function
#   and "prices"
# - ... ... other parameters sent to tm_dots()
#
# output:
# - object of class "tmap"; if it is not stored in a variable, it is plotted;
#   it can be joined with other tmap objects
#
# details:
# - when several periods are selected, the prices are averaged over the periods
#   and regions; the average is computed only from existing stations with price
#   different from NA
# - in interactive mode, additional information on regions are available on
#   clicking
#
# usage (see example_plotting.R):
#   tmap_mode("view")  # for interactive mode
#   natural_monthly %>%
#       filter(from_date == "2018-01-01") %>%
#       plot_station_prices(existence = "station_exists_30")
plot_station_prices <- function(data,
                                arccr_folder = NULL, layer = NULL,
                                regions = "nic",
                                existence = NULL,
                                stat = base::mean,
                                facet = FALSE,
                                facet_col = NA,
                                facet_row = NA,
                                plot_regional_averages = FALSE,
                                regional_averages_alpha = NA,
                                legend_title = NULL,
                                .info_type = "price",
                                ...) {

    # regions
    p <- NULL
    if (regions != "nic") {
        if (is.null(arccr_folder))
            stop("If you want to plot regions, you must supply path to ARC ČR folder.")
        if (!plot_regional_averages)
            regional_averages_alpha <- 0
        p <- plot_average_prices(data, arccr_folder, layer, regions, existence,
                                 facet = facet,
                                 facet_col = facet_col, facet_row = facet_row,
                                 alpha = regional_averages_alpha,
                                 legend.show = FALSE)
    }

    # check and fill in inputs
    check_plotting_parameters(data, regions, extra_region_option = "nic")
    if (is.null(existence)) {
        existence <- TRUE
    } else {
        existence <- data[[existence]]
    }

    # filter only existing stations
    data <- data[existence, ]

    # average the data
    remove_nas <- function(x) x[!is.na(x)]
    data <- data %>%
        dplyr::group_by(id)
    if (facet)
        data <- data %>% dplyr::group_by(from_date, .add = TRUE)
    data <- data %>%
        dplyr::summarise(name = name[1],
                         price = stat(remove_nas(price)),
                         address = address[1],
                         lon = lon[1],
                         lat = lat[1],
                         .groups = "drop") %>%
        dplyr::mutate(station = stringr::str_c(id, ": ", name))

    # add shapes
    data <- data %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326, remove = FALSE) %>%
        sf::st_transform(5514)

    # return the plot
    if (is.null(legend_title))
        legend_title <- "station prices"
    popup_vars <- c("id", "name", "price", "address", "lat", "lon")
    if (.info_type == "residuals") {
        data$residual <- data$price
        popup_vars <- c("id", "name", "residual", "address", "lat", "lon")
    }
    p <- p +
        tmap::tm_shape(data) +
        tmap::tm_dots("price",
                      id = "station",
                      popup.vars = popup_vars,
                      title = legend_title,
                      ...)
    if (facet)
        p <- p + tmap::tm_facets(by = "from_date",
                                 ncol = facet_col, nrow = facet_row)
    p
}




# plot residuals ---------------------------------------------------------------

# plot_average_residuals() plots residuals of panel econometrics model for given
# regions and periods of time
#
# inputs:
# - data ... (tibble) aggregated data created by prepare_average_fuel_data() for
#   the periods covered in estimate
# - estimate ... (estimate of fixest or sar) the estimated model
# - arccr_folder ... (character scalar) path to "AdministrativniCleneni_v13.gdb"
#   folder
# - layer ... (optional character scalar) layer from arccr_folder used; if not
#   given, the function selects the appropriate one
# - regions ... (one of "obce", "okresy", "kraje", and "cr"; default is
#   "okresy") for which kind of regions the average residuals should be plotted
# - stat ... (function, default is mean) statistics computed for each region
# - facet ... (logical, default is FALSE) if TRUE, the plots are facetted by
#   from_date; if FALSE (default), the average residuals are computed over all
#   periods
# - facet_col, facet_row ... (integer scalar) how many columns / rows should be
#   in the facetted plot; if not given, tmap chooses some reasonable defaults
# - legend_title ... (character scalar) if given, it is used as a title of the
#   legend; if not given, the title is composed of the name of the stat function
#   and "residuals"
# - plot_labels ... (optional logical scalar, default is FALSE) if TRUE, prices
#   are plotted in the regions
# - ... ... other parameters sent to tm_polygons()
#
# output:
# - object of class "tmap"; if it is not stored in a variable, it is plotted;
#   it can be joined with other tmap objects
#
# details:
# - data must be (subbset of) the panel_data used to estimate the estimate
# - when several periods are selected, the residuals are averaged over the
#   periods and regions -- unless facet = TRUE
# - if facet = TRUE, the plot is facetted
# - in interactive mode, additional information on regions are available on
#   clicking
#
# usage (see example_panel_estimation_and_merger_simulations.R):
#   panel_data %>%
#       filter(from_date == "2018-01-01") %>%
#       plot_average_residuals(estimate_sar,
#                               arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")
#   panel_data %>%
#       filter(from_date == "2018-01-01") %>%
#       plot_average_residuals(estimate_sar,
#                               arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
#                               stat = median)
#   panel_data %>%
#       filter(from_date == c("2018-01-01", "2018-06-01")) %>%
#       plot_average_residuals(estimate_sar,
#                               arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
#                               facet = TRUE)
plot_average_residuals <- function(data, estimate,
                                   arccr_folder = NULL, layer = NULL,
                                   regions = "okresy",
                                   stat = base::mean,
                                   facet = FALSE,
                                   facet_col = NA,
                                   facet_row = NA,
                                   legend_title = NULL,
                                   plot_labels = FALSE,
                                   ...) {
    stopifnot(inherits(estimate, "fixest") || inherits(estimate, "splm") ||
                  inherits(estimate, "lm") || inherits(estimate, "sarlm"))
    if (inherits(estimate, "sarlm") && !("sarlm2" %in% base::names(estimate)))
        stop("You must estimate cross-sectional SAR with lagsarlm2().")
        pred <- predict(estimate, newdata = data)
    data$price <- data$price - pred
    fname <- base::as.character(base::substitute(stat))
    fname <- fname[length(fname)]
    if (is.null(legend_title))
        legend_title <- stringr::str_c(fname, " residuals")
    param_list <- list(data = data,
                       arccr_folder = arccr_folder,
                       layer = layer,
                       regions = regions,
                       existence = NULL,
                       stat = stat,
                       facet = facet,
                       facet_col = facet_col, facet_row = facet_row,
                       legend_title = legend_title,
                       plot_labels = plot_labels,
                       .info_type = "residuals",
                       palette = "-RdYlGn", midpoint = 0,
                       ...)
    if (!("palette" %in% base::names(list(...))))
        param_list$palette  <- "-RdYlGn"
    if (!("midpoint" %in% base::names(list(...))))
        param_list$midpoint <- 0
    base::do.call(plot_average_prices, param_list)
}


# plot_station_residuals() plots residuals of an estimated model for individual
# stations and periods of time
#
# inputs:
# - data ... (tibble) aggregated data created by prepare_average_fuel_data()
# - estimate ... (estimate of fixest or sar) the estimated model
# - full_data ... (tibble) the full table of data from which data is
#   a column-subset; it is used to supplement necessary columns if they are
#   missing from data
# - arccr_folder ... (character scalar) path to "AdministrativniCleneni_v13.gdb"
#   folder; it is necessary only when regions is not "nic"
# - layer ... (optional character scalar) layer from arccr_folder used; if not
#   given, the function selects the appropriate one; it is necessary only when
#   regions is not "nic"
# - regions ... (one of "nic", "obce", "okresy", "kraje", and "cr"; default is
#   "nic") if not "nic" (which is default), the selected regions' border are
#   plotted
# - stat ... (function, default is mean) statistics computed for each station
# - facet ... (logical, default is FALSE) if TRUE, the plots are facetted by
#   from_date; if FALSE (default), the average prices are computed over all
#   periods
# - facet_col, facet_row ... (integer scalar) how many columns / rows should be
#   in the facetted plot; if not given, tmap chooses some reasonable defaults
# - legend_title ... (character scalar) if given, it is used as a title of the
#   legend; if not given, the title is composed of the name of the stat function
#   and "residuals"
# - ... ... other parameters sent to tm_dots()
#
# output:
# - object of class "tmap"; if it is not stored in a variable, it is plotted;
#   it can be joined with other tmap objects
#
# details:
# - when several periods are selected, the prices are averaged over the periods
#   and regions; the average is computed only from existing stations with price
#   different from NA
# - in interactive mode, additional information on regions are available on
#   clicking
#
# usage (see example_panel_estimation_and_merger_simulations.R):
#   panel_data %>%
#       filter(from_date == "2018-01-01") %>%
#       plot_station_residuals(estimate_sar,
#                              natural_monthly,
#                              arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
#                              regions = "okresy",
#                              size = 0.08)
plot_station_residuals <- function(data, estimate,
                                   full_data = NULL,
                                   arccr_folder = NULL, layer = NULL,
                                   regions = "nic",
                                   stat = base::mean,
                                   facet = FALSE,
                                   facet_col = NA,
                                   facet_row = NA,
                                   legend_title = NULL,
                                   ...) {
    stopifnot(inherits(estimate, "fixest") || inherits(estimate, "splm") ||
                  inherits(estimate, "lm") || inherits(estimate, "sarlm"))
    if (inherits(estimate, "sarlm") && !("sarlm2" %in% base::names(estimate)))
        stop("You must estimate cross-sectional SAR with lagsarlm2().")
    if (!all(c("address", "name", "lat", "lon") %in% base::names(data)) &&
        is.na(full_data))
        stop("Some of the necessary columns is missing from data. ",
             "Supply full_data.")
        pred <- predict(estimate, newdata = data)
    data$price <- data$price - pred
    add_names <- base::setdiff(c("address", "name", "lon", "lat"),
                         base::names(data))
    data <- dplyr::left_join(data,
                             full_data %>%
                                 select("id", "from_date", all_of(add_names)),
                             by = c("id", "from_date"))
    fname <- base::as.character(base::substitute(stat))
    fname <- fname[base::length(fname)]
    if (is.null(legend_title))
        legend_title <- stringr::str_c(fname, " residuals")
    param_list <- list(data = data,
                       arccr_folder = arccr_folder, layer = layer,
                       regions = regions,
                       existence = NULL,
                       stat = stat,
                       facet = facet,
                       facet_col = facet_col, facet_row = facet_row,
                       plot_regional_averages = FALSE,
                       legend_title = legend_title,
                       .info_type = "residuals",
                       ...)
    if (!("palette" %in% base::names(list(...))))
        param_list$palette  <- "-RdYlGn"
    if (!("midpoint" %in% base::names(list(...))))
        param_list$midpoint <- 0
    base::do.call(plot_station_prices, param_list)
}





# plot predicted prices / price changes ----------------------------------------

# plot_average_merger_prices() plots average simulated merger price changes (or
# prices) for each region on the map
#
# inputs:
# - merger_data ... (tibble) table created by simulate_merger()
# - full_data ... (tibble) full data from which a (balanced) panel is used to
#   estimate the model used for the merger simulation
# - var ... (character scalar) the name of variable (column) in merger_data used
#   to set colors of dots; default is "price_change_full"
# - arccr_folder ... (character scalar) path to "AdministrativniCleneni_v13.gdb"
#   folder
# - layer ... (optional character scalar) layer from arccr_folder used; if not
#   given, the function selects the appropriate one
# - regions ... (one of "obce", "okresy", "kraje", and "cr"; default is
#   "okresy")
# - legend_title ... (character scalar) if given, it is used as a title of the
#   legend; if not given, the title is compbination of the stat name and var
# - ... ... other parameters sent to tm_dots()
#
# output:
# - object of class "tmap"; if it is not stored in a variable, it is plotted;
#   it can be joined with other tmap objects
#
# usage (see example_panel_estimation_and_merger_simulations.R):
#   plot_average_merger_prices(mol_merger_sar, natural_monthly,
#       arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/")
plot_average_merger_prices <- function(merger_data, full_data,
                                       var = "price_change_full",
                                       arccr_folder, layer = NULL,
                                       regions = "okresy",
                                       stat = base::mean,
                                       legend_title = NULL,
                                       ...) {
    # merge missing data
    merger_data <- dplyr::left_join(merger_data %>% dplyr::select(-to_date),
                                    full_data %>%
                                        dplyr::select(id, from_date,
                                                      address, lon, lat),
                                    by = c("id", "from_date"))
    merger_data$price <- merger_data[[var]]

    # plot
    if (is.null(legend_title)) {
        fname <- base::as.character(base::substitute(stat))
        fname <- fname[base::length(fname)]
        legend_title <- stringr::str_c(fname, " ", var)
    }
    param_list <- list(data = merger_data,
                       arccr_folder = arccr_folder, layer = layer,
                       regions = regions,
                       existence = NULL,
                       stat = stat,
                       facet = FALSE,
                       facet_col = NA,
                       facet_row = NA,
                       legend_title = legend_title,
                       plot_labels = FALSE,
                       .info_type = "merger",
                       ...)
    if (!("palette" %in% base::names(param_list)))
        param_list$palette <- "-RdYlGn"
    if (!("midpoint") %in% base::names(list(...)))
        param_list$midpoint <- 0
    base::do.call(plot_average_prices, param_list)
}


# plot_station_merger_prices() plots simulated merger price changes (or prices)
# of individual stations on the map
#
# inputs:
# - merger_data ... (tibble) table created by simulate_merger()
# - full_data ... (tibble) full data from which a (balanced) panel is used to
#   estimate the model used for the merger simulation
# - var ... (character scalar) the name of variable (column) in merger_data used
#   to set colors of dots; default is "price_change_full"
# - arccr_folder ... (character scalar) path to "AdministrativniCleneni_v13.gdb"
#   folder; it is necessary only when regions should be plotted
# - layer ... (optional character scalar) layer from arccr_folder used; if not
#   given, the function selects the appropriate one
# - regions ... (one of "nic", "obce", "okresy", "kraje", and "cr"; default is
#   "nic") no regions are plotted
# - legend_title ... (character scalar) if given, it is used as a title of the
#   legend; if not given, the title is "station price changes"
# - ... ... other parameters sent to tm_dots()
#
# output:
# - object of class "tmap"; if it is not stored in a variable, it is plotted;
#   it can be joined with other tmap objects
#
# usage (see example_panel_estimation_and_merger_simulations.R):
#   plot_station_merger_prices(mol_merger_sar, natural_monthly)
#   plot_station_merger_prices(mol_merger_sar, natural_monthly,
#       arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb/",
#       regions = "okresy")
plot_station_merger_prices <- function(merger_data, full_data,
                                       var = "price_change_full",
                                       arccr_folder, layer = NULL,
                                       regions = "nic",
                                       legend_title = NULL,
                                       ...) {
    # regions
    p <- NULL
    if (regions != "nic") {
        if (is.null(arccr_folder))
            stop("If you want to plot regions, you must supply path to ARC ČR folder.")
        # if (!plot_regional_averages)
        #     regional_averages_alpha <- 0
        data <- dplyr::semi_join(full_data, merger_data, by = "from_date")
        p <- plot_average_prices(data, arccr_folder, layer, regions,
                                 existence = NULL,
                                 facet = FALSE,
                                 alpha = 0,
                                 legend.show = FALSE)
    }

    # merge missing data
    merger_data <- dplyr::left_join(merger_data %>% dplyr::select(-to_date),
                                    full_data %>%
                                        dplyr::select(id, from_date,
                                                      address, lon, lat),
                                    by = c("id", "from_date"))

    # add shapes
    merger_data <- merger_data %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326, remove = FALSE) %>%
        sf::st_transform(5514)

    # return the plot
    if (is.null(legend_title))
        legend_title <- "station price changes"
    popup_vars <- base::setdiff(base::names(merger_data), "geometry")
    param_list <- list(col = var,
                       id = "id",
                       popup.vars = popup_vars,
                       title = legend_title,
                       ...)
    if (!("palette") %in% base::names(param_list))
        param_list$palette <- "-RdYlGn"
    if (!("midpoint") %in% base::names(param_list))
        param_list$midpoint <- 0
    p + tmap::tm_shape(merger_data) +
        base::do.call(tmap::tm_dots, param_list)
}
