# -------------------------------------
# Script: preparatory_functions_general.R
# Author: Michal Kvasnička
# Purpose: This script contains functions used for general data preparation.
# Notes: For usage see script prepare_pumpdroid_data.R.
#
# Copyright(c) Masaryk University
# -------------------------------------


# libraries
library(tibble)
library(dplyr)
library(purrr)
library(furrr)
library(tidyr)
library(stringr)
library(lubridate)
library(readr)
library(sf)


# other necessary functions
source(file.path(CODEDIR, "utilities.R"))





# adding location data --------------------------------------------------------------

# add_arccr_municipality(station_locations, arccr_folder, layer) takes
# the station location table and adds four new columns: obec, okres, kraj, and
# country; it grabs these data from ARC ČR data
# (https://www.arcdata.cz/produkty/geograficka-data/arccr-500)
#
# inputs:
# - station_locations ... tibble including station locations, such as provided
#   by read_pumpdroid_station_locations()
# - arccr_path ... a path (character vector of length 1) to
#       "AdministrativniCleneni_v13.gdb" folder
# - layer ... (optional) the layer in ARC ČR database that should be used; until
#   there is some ugly change in the database, don't change the default value
#
# output:
# - a tibble of stations_locations where four new columns are added; the country
#   column is either "Česká republika" or "cizina"
#
# usage:
#   station_locations <- read_pumpdroid_station_locations(station_data) %>%
#       add_arccr_municipality("map_shapes/AdministrativniCleneni_v13.gdb/")
add_arccr_municipalities <- function(station_locations, arccr_folder,
                                     layer = "ObcePolygony") {
    # remember all column names
    ori_names <- base::names(station_locations)

    # trasform to a spatial dataset and set CRS to WGS84 ~ GPS coordinates
    station_locations <- station_locations %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326, remove = FALSE)

    # load map of municipalities (ArcCR maps)
    municipalities <- sf::st_read(arccr_folder, layer,
                                  stringsAsFactors = FALSE,
                                  quiet = TRUE)

    # transform CRS of station_locations and roads to JTSK
    # 5514 is a magic number for JTSK
    station_locations <- sf::st_transform(station_locations, 5514)
    municipalities <- sf::st_transform(municipalities, 5514)

    # assign gas stations to municipalities
    station_locations <- sf::st_join(station_locations, municipalities)

    # if multiple match, choose the municipality with larger population
    station_locations <- station_locations %>%
        sf::st_drop_geometry() %>%
        dplyr::group_by(id) %>%
        dplyr::arrange(desc(POCET_OBYV)) %>%
        dplyr::slice(1) %>%
        dplyr::ungroup() %>%
        dplyr::select(tidyselect::all_of(ori_names),
                      obec = NAZ_OBEC,
                      okres = NAZ_LAU1,
                      kraj = NAZ_CZNUTS3) %>%
        dplyr::mutate(country = dplyr::if_else(is.na(obec), "cizina", "Česká republika"))

    # return
    station_locations
}


# add_arccr_roads(station_locations, arccr_folder, layer, max_dist, highway_classes)
# takes the station location table and adds four new columns: acr_road_class,
# acr_road_int_class, acr_road_number, and acr_road_lanes; it grabs these data
# from ARC ČR data (https://www.arcdata.cz/produkty/geograficka-data/arccr-500)
#
# inputs:
# - station_locations ... tibble including station locations, such as provided
#   by read_pumpdroid_station_locations()
# - arccr_path ... a path (character vector of length 1) to "ArcCR500_v33.gdb"
#   folder
# - layer ... (optional) the layer in ARC ČR database that should be used; until
#   there is some ugly change in the database, don't change the default value
# - max_dist ... (optional) the maximal distance in meters in which the roads
#   are searched
# - highway_classes ... (optional) integer vector of acr_road_number (TRIDA) of
#   road classes that are considered as highways; defautl = 1:2 (highways and
#   motorways)
#
# output:
# - a tibble of stations_locations where four new columns are added;
#   - acr_road_class (integer) ... czech road classification (třída silnice):
#       1 - dálnice
#       2 - rychlostní silnice
#       3 - silnice I. třídy
#       4 - silnice II. třídy
#       5 - silnice III. třídy
#       6 - neevidovaná silnice)
#   - acr_road_int_class (character) ... international road classification
#           (such as E55)
#   - acr_road_number (character) ... czech road number (such as 55 or 38H)
#   - arc_road_distance (double) ... distance to the selected road (Inf if
#       not found)
#   - acr_road_lanes (integer) ... number of lanes
#   - arc_highway (logical) ... TRUE if the road is highway
#
# usage:
#   station_locations <- read_pumpdroid_station_locations(station_data) %>%
#       add_arccr_roads("map_shapes/ArcCR500_v33.gdb/")
add_arccr_roads <- function(station_locations, arccr_folder,
                            layer = "Silnice_2016",
                            max_dist = 150,
                            highway_classes = 1:2) {
    ori_names <- base::names(station_locations)

    # convert to spatial data.frame; the CRS is WGS84 ~ GPS coordinates
    station_locations <- station_locations %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326, remove = FALSE)

    # load map of roads (ArcCR maps)
    roads <- sf::st_read(arccr_folder, layer, stringsAsFactors = FALSE,
                         quiet = TRUE)

    # transform CRS of station_locations and roads to JTSK
    # 5514 is a magic number for JTSK
    station_locations <- sf::st_transform(station_locations, 5514)
    roads <- sf::st_transform(roads, 5514)

    # find all roads (road segments) that are not further than max_dist meters from the
    # gas station, and then select the best fit based on some criterion, e.g. the shortest
    # distance
    # 1. add road id to all road segments, join roads to station_locations, and then
    #   select roads with the particulars ids (in their order)
    roads <- dplyr::mutate(roads, road_id = row_number())
    station_locations <- sf::st_join(station_locations, roads,
                                     join = sf::st_is_within_distance,
                                     dist = max_dist)
    sil_sort <- roads[station_locations$road_id, ]
    # 2. add a distance between the stations and roads to the tibble
    station_locations$road_dist <- purrr::map2_dbl(sf::st_geometry(station_locations),
                                                   sf::st_geometry(sil_sort),
                                                   st_distance)
    station_locations$road_dist[is.na(station_locations$road_id)] <- Inf
    # 3. remove all roads but the best one (the closest one now)
    station_locations <- station_locations %>%
        dplyr::group_by(id) %>%
        dplyr::arrange(road_dist) %>%
        dplyr::slice(1) %>%
        dplyr::ungroup()

    # return
    station_locations %>%
        sf::st_drop_geometry() %>%
        dplyr::select(tidyselect::all_of(ori_names),
                      arc_road_class = TRIDA,
                      arc_road_int_class = MEZINARODNI_OZNACENI,
                      arc_road_number = CISLO_SILNICE,
                      arc_road_lanes = PRUHY,
                      arc_road_distance = road_dist) %>%
        dplyr::mutate(across(c(arc_road_int_class, arc_road_number),
                             ~dplyr::if_else(. == "N_A", NA_character_, .)),
                      arc_road_lanes = as.integer(arc_road_lanes),
                      arc_highway = arc_road_class %in% highway_classes)
}


# osm_road_class_names() returns the fclass of OSM roads that are used within
# the analysis
#
# - motorway / motorway_link ... highways
# - trunk / "trunk_link ... important roads, very close to highways
# - primary, secondary, tertiary ... national, regional, local
# - residential, living_street ... residential, pedestrians have priority in
#   the latter
# - unclassified ... small local roads
# - service ... ervice roads for access to buildings, parking lots, etc.
#   (and especially to gas stations, e.g. on highways but on other roads too) --
#   for this very reason it is eliminated from the search for the nearest road
osm_road_class_names <- function() {
    c("motorway", "trunk", "primary", "secondary", "tertiary",
      "residential", "living_street", "unclassified",
      "motorway_link", "trunk_link", "primary_link", "secondary_link")
    #      "service")
}


# osm_highway_names() returns the fclass of OSM roads that are considered as
# highways
#
# - motorway / motorway_link ... highway no doubt
# - trunk / trunk_link ... more questionable -- they consist of three cases now:
#   - city loops in Prague, Brno, and Ostrava
#   - three big roads connected to motorways
#   - quite many very short roads, usually (but not always) connected to
#     highways if it is added to highways, should intercity loops be counted?
osm_highway_names <- function() {
    c("motorway", "motorway_link", "trunk", "trunk_link")
}


# add_osm_roads(station_locations, osm_folder) takes the station location table
# and adds three new columns: osm_road_class, osm_oneway, and osm_maxspeed; it
# grabs these data from OpenStreetMaps for the Czech Republic
#
# warning:
# - the function takes a lot of time (about 20 minuts on i5) because there are
#   very many roads in the OSM database; you can save some time when you
#   eliminate all foreign and non-existing gas stations from station_locations
#   tibble before you run this function
# - the function pre-filters the roads; if the OSM road classification (fclass)
#   changes, then osm_road_class_names() function must be modified accordingly
#
# inputs:
# - station_locations ... tibble including station locations, such as provided
#   by read_pumpdroid_station_locations()
# - osm_folder ... a path (character vector of length 1) to "OpenStreetMaps"
#   folder where OSM Maps are stored as shapefiles
# - layer ... (optional) the layer in OpenStreetmaps database that should be
#   used; until there is some ugly change in the database, don't change the
#   default value
# - max_dist ... (optional) the maximal distance in meters in which the roads
#   are searched
# - road_classes ... (optional) character vector of OSM fclasses that are used
#   as roads on which gass stations could be; default = osm_road_class_names()
# - highway_classes ... (optional) character vector of OSM fclasses that are
#   used as highway definition; default = osm_highway_names()
#
# output:
# - a tibble of stations_locations where four new columns are added;
#   - osm_road_class (character) ... OSM road classification, see their
#     documentation
#   - osm_oneway ... whether the road is oneway or not
#   - osm_maxspeed ... the roads maximum allowed speed
#   - osm_road_dist (double) ... a distance to the selected road (Inf if not
#     found)
#   - osm_highway (logical) ... highway or motorway (see variable highway_names
#     for the definition)
#
# usage:
#   station_locations <- read_pumpdroid_station_locations(station_data) %>%
#       add_osm_roads("map_shapes/OpenStreetMaps/")
add_osm_roads <- function(station_locations, osm_folder,
                          layer = "gis_osm_roads_free_1", max_dist = 150,
                          road_classes = osm_road_class_names(),
                          highway_classes = osm_highway_names()) {
    ori_names <- base::names(station_locations)

    # create spatial object; the CRS is WGS84 ~ GPS coordinates
    station_locations <- station_locations %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326, remove = FALSE)

    # load map of roads (OSM maps)
    roads <- sf::st_read(osm_folder, layer, stringsAsFactors = FALSE,
                         quiet = TRUE)

    # filter only car-roads
    roads <- roads %>%
        dplyr::filter(fclass %in% road_classes)

    # transform CRS of station_locations and roads to JTSK
    # 5514 is a magic number for JTSK
    station_locations <- sf::st_transform(station_locations, 5514)
    roads <- sf::st_transform(roads, 5514)

    # find all roads (road segments) that are not further than max_dist meters
    # from the gas station, and then select the best fit based on some
    # criterion, e.g. the shortest distance
    # 1. add road id to all road segments, join roads to station_locations, and
    # then select roads with the particulars ids (in their order)
    roads <- dplyr::mutate(roads, road_id = row_number())
    station_locations <- sf::st_join(station_locations, roads,
                                     join = st_is_within_distance,
                                     dist = max_dist)
    sil_sort <- roads[station_locations$road_id, ]
    # 2. add a distance between the stations and roads to the tibble
    station_locations$road_dist <- purrr::map2_dbl(st_geometry(station_locations),
                                                   st_geometry(sil_sort),
                                                   st_distance)
    station_locations$road_dist[is.na(station_locations$road_id)] <- Inf
    # 3. remove all roads but the best one (the closest one now)
    station_locations <- station_locations %>%
        dplyr::group_by(id) %>%
        dplyr::arrange(road_dist) %>%
        dplyr::slice(1) %>%
        dplyr::ungroup()

    # return
    station_locations %>%
        st_drop_geometry() %>%
        dplyr::select(tidyselect::all_of(ori_names),
                      osm_road_class = fclass,
                      osm_oneway = oneway,
                      osm_maxspeed = maxspeed,
                      osm_road_dist = road_dist) %>%
        dplyr::mutate(osm_highway = osm_road_class %in% highway_classes)
}





# removing gas stations --------------------------------------------------------

# remove_foreign_gas_stations(stations, locations) removes from tibble stations
# all rows for foreing gas stations; the locality information is taken from
# locations tibble
#
# inputs:
# - stations ... a tibble including a column id containing gas station ids
# - locations ... a tibble augmented by add_arccr_municipality(); it must
#   include a column country which must contain a character "Česká republika"
#   if the gas station is in the Czech Republic and something else otherwise
#
# output:
# - a tibble with the same structure as stations; all rows for foreign stations
#   are eliminated
#
# usage:
#   price_data <- remove_foreign_gas_stations(price_data, station_locations)
remove_foreign_gas_stations <- function(stations, locations) {
    czech_stations <- locations %>%
        dplyr::filter(country == "Česká republika")
    dplyr::semi_join(stations, czech_stations, by = "id")
}


# remove_gas_stations_with_no_price_data(stations, price_data) removest from
# tibble stations all rows for gas stations that have no observation in
# price_data whatsoever
#
# inputs:
# - stations ... a tibble including a column id containing gas station ids
# - price_data .. a tibble with price informations; it must include a column id
#
# output:
# - a tibble with the same structure as stations; all rows for stations with no
#   price observations are eliminated
#
# usage:
#   station_locations <- remove_foreign_gas_stations(station_locations,
#                                                    price_data)
remove_gas_stations_with_no_price_data <- function(stations, price_data) {
    ids <- base::unique(price_data$id)
    stations %>%
        dplyr::filter(id %in% ids)
}





# beeline and driving distances ------------------------------------------------

# function get_beeline_distances() takes station_location dataset and returns a
# dataset of beeline distance between stations
#
# it seems that beeline-distance calculation is much faster than navigation;
# hence, beeline distances are used to select the pairs of gas stations for
# which the driving distance is calculated
#
# inputs:
# - station_location ... tibble of station locations
#
# output:
# - tibble with the following columns:
#   - from_id ... (integer) id of the origin gas station
#   - to_id ... (integer) id of the destination gas station
#   - beeline_distance ... (double) the beeline destination between them
#   The dataset includes for each pair of stations only one record (the bee-line
#   distances are symmetrical).
#   Therefor a search in the table must look like this:
#       from_id <- 2
#       to <- 1
#       station_get_beeline_distances %>%
#       dplyr::filter((from_id == from & to == to) | (from_id == from & to == to))
#   or
#       f <- min(c(from_id, to))
#       t <- max(c(from_id, to))
#       station_get_beeline_distances %>%
#       dplyr::filter(from_id == f, to == t)
#
# warnings:
# - the run of the function takes a lot of time (about 20 minutes on i5 with
#   four cores)
#
# details:
# - The function works with vectors and not datsets because filling the datasets
#   in is more than thousand times slower!)
# - This is a paralellized version of the function. Its speed depends on the
#   number of cores used.
# - A better balancing is possible but it probably doesn't pay. Several
#   approaches (not taken here) are stupid:
#   - To split all tasks to N chunks (where N is the number of cores) and apply
#     st_distance() to each chunk. It would definetly be faster but would create
#     huge vectors (each several gigabytes).
#   - To create the indices and run st_distance() for each origin--destination
#     pair by future_map(). The overhead it too heavy.
#   While the current approach definitely is suboptimal, the better optimization
#   would create a clumsy code and the speed improvement wouldn't be important.
#   Moreover, speed--memory consumption balance would be quite difficult.
get_beeline_distances <- function(station_locations) {
    # check that every station is in the table just once
    if (length(station_locations$id) != length(unique(station_locations$id)))
        stop("Some stations are present several times.")
    # trasform to a spatial dataset and set CRS to WGS84 ~ GPS coordinates
    station_locations <- station_locations %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326) %>%
        dplyr::select(id)
    # transform CRS of station_locations and roads to JTSK
    # 5514 is a magic number for JTSK
    station_locations <- sf::st_transform(station_locations, 5514)
    # calculate the beeline distance lookup table
    N <- base::nrow(station_locations)
    # function to compute distnaces among selected stations
    many_dist <- function(p, station_locations, N) {
        purrr::map(p,
                   ~ tibble::tibble(
                       from_id = station_locations$id[1:(N - .)],
                       to_id = station_locations$id[(. + 1):N],
                       beeline_distance = units::drop_units(
                           sf::st_distance(station_locations$geometry[1:(N - .)],
                                           station_locations$geometry[(. + 1):N],
                                           by_element = TRUE))
                   )) %>%
            dplyr::bind_rows()
    }
    # roughly balancing the load
    parts <- base::min(used_future_workers(), N - 1)
    plan <- purrr::map(1:parts, ~ base::seq(from = ., to = (N - 1), by = parts))
    # the computation
    furrr::future_map(plan,
                      ~ many_dist(., station_locations, N)) %>%
        dplyr::bind_rows() %>%
        dplyr::arrange(from_id, to_id)
}


# get_driving_distances() calculates for selected pairs of gas stations driving
# distances between them. Since the driving distances are not symmetrical, the
# resulting table must include both routes from a to b and from b to a. Only
# routes with distances no longer than max_dist are typically kept in the table.
#
# inputs:
# - station_locations ... tibble of station locations
# - beeline_distances ... (optional) a tibble created by get_beeline_distances();
#   if not given, it is calcluated within the function;
#   beware: it takes a lot of time
# - max_dist ... (numeric) a maximal distance [in km] for which the routes are
#   calculated
# - shortest_route ... (logical) if TRUE, the distance of the shortest route
#   is calculated; if FALSE (default), the distance of the fastest route is
#   computed (see the details)
# - map_path ... (character scalar) a path to the OpenStreeMap .osm.pbf file
#   provided by Geofabric, i.e. czech-republic-latest.osm.pbf
# - roadrunner_folder ... (character scalar) a path to the folder where
#   RoadRunner.jar is located
# - remove_longer ... (optional, logical scalar) if TRUE (default), the
#   distances longer than max_dist are removed from the outcome table; if you
#   want to keep them for some reason (e.g. you want to use the ideal driving
#   time instead of driving distance) set it to FALSE; however, then you are
#   responsible for what's going on
# - temp_map_folder ... (optional, character scalar) a path to folder where
#   GraphHopper creates its graphs; the transitory CSV files and possibly
#   an error file are stored as well; if not given, the temporary folder is
#   created in in tempdir() and is removed when R session ends; since the graph
#   creation takes some time, the folder can be specified to save time when the
#   function is run several times; however the temporary folder must be deleted
#   when .osm.pbf map file changes---the change would be neglected otherwise
# - err_file_path ... (optional, character scalar) a path where RoadRunner
#   creates a CSV list of failed routes; if not given, it is created in
#   tempdir() and is erased as soon as R session closes
#
# output:
# - tibble with the following columns
#   - from_id ... (integer) id of the origin station
#   - to_id ... (integer) id of the destination station
#   - driving_distance ... (double) ... driving distance in meters
#   - driving_time ... (doubler) ... driving time in minutes calculated based on
#     speed limits; this does not consider any actual road jams, accidents, or
#     traffic conditions
#   Only routes that are no longer than max_dist are kept in the table.
#
# details:
# - The function runs a Java software RoadRunner (build upon GraphHopper). Hence
#   Java must be installed on the system. The minimal supported Java version is
#   14. The code is tested with Java 14.0.2 on Linux and 15.0.1 on Windows.
# - Usually, only stations up to a small distance (such as 20 km) are
#   considered. If max_dist is small, than it may be faster first compute
#   beeline distances among all stations and then compute the driving distances
#   only for stations with beeline distance no longer than max_dist. If,
#   however, max_dist is large, then it may be faster to set max_dist = Inf.
#   Then the computation of beeline distances is skipped and driving distances
#   among all stations are calculated.
#
# warnings:
# - The run of the function can take a lot of time if max_dist is longer than
#   few dozens kilometers or when beeline distances must be computed within it.
# - The shortest routes may not be always calculated due to errors in
#   OpenStreetMaps. It seems that the errors have been largely corrected by the
#   end of 2020 but they can be reintroduced at any time. As the shorter
#   distances make better sense anyway, the user should rather stick to it.
# - On Windows, the progress bar not adjust smoothly. This is a problem with
#   Windows shell() function and there may not be any workaround.
#
# tips:
# - If you work on Windows and the path to your home folder includes accented
#   letters, spaces and so on, the path to tempdir() may not work properly, and
#   hence this functions collapses. The remedy is to set parameter
#   temp_map_folder to a reasonable existing path.
#
# usage:
#  driving_distances <-
#      get_driving_distances(station_locations, station_beeline_distances,
#          max_dist = 5,
#          map_path = "~/map_shapes/czech-republic-latest.osm.pbf",
#          roadrunner_folder = "~/roadrunner")
#  driving_distances <-
#      get_driving_distances(station_locations,
#          max_dist = Inf,
#          map_path = "~/map_shapes/czech-republic-latest.osm.pbf",
#          roadrunner_folder = "~/roadrunner")
get_driving_distances <- function(station_locations, beeline_distances = NULL,
                                  max_dist = 20, shortest_route = FALSE,
                                  map_path, roadrunner_folder,
                                  remove_longer = TRUE,
                                  temp_map_folder = base::tempdir(),
                                  err_file_path = base::file.path(temp_map_folder,
                                                                  "roadrunner_errors.csv")) {
    # check that every station is in the table just once
    if (base::length(station_locations$id) !=
        base::length(base::unique(station_locations$id)))
        stop("Some stations are present several times.")
    # if beeline_distances tibble is not given, compute it
    if (is.null(beeline_distances) && is.finite(max_dist))
        beeline_distances <- get_beeline_distances(station_locations)
    # if max_dist is finite, prefilter all possible connection so that only
    # short routes are computed by GraphHopper; otherwise compute distances for
    # all possible combinations of origin--destination
    if (is.infinite(max_dist)) {
        # if all routes are computed, construct all possible routes (not from A
        # to A)
        driving_distances <- tidyr::crossing(from_id = station_locations$id,
                                             to_id = station_locations$id) %>%
            dplyr::filter(from_id != to_id)
    } else {
        # if only selected routes, create the tibble with routes that are no
        # longer than max_dist [km], i.e. their beeline distance is no longer
        # than max_dist
        driving_distances <- beeline_distances %>%
            dplyr::filter(beeline_distance <= max_dist * 1e3) %>%
            dplyr::select(-beeline_distance)
        driving_distances <- dplyr::bind_rows(
            driving_distances,
            tibble::tibble(from_id = driving_distances$to_id,
                           to_id = driving_distances$from_id)
            )
    }
    # add locations
    driving_distances <- driving_distances %>%
        dplyr::left_join(station_locations %>%
                             dplyr::select(id, lat, lon),
                         by = c("from_id" = "id")) %>%
        dplyr::rename(from_lat = lat, from_lon = lon) %>%
        dplyr::left_join(station_locations %>%
                             dplyr::select(id, lat, lon),
                         by = c("to_id" = "id")) %>%
        dplyr::rename(to_lat = lat, to_lon = lon)
    # create csv file that would be processed by GraphHopper
    routes_file <- base::tempfile(tmpdir = temp_map_folder, fileext = ".csv")
    readr::write_csv(driving_distances, file = routes_file)
    # run GraphHopper
    navigation_type <- if (shortest_route) "shortest" else "fastest"
    distances_file <- base::tempfile(tmpdir = temp_map_folder, fileext = ".csv")
    command <- stringr::str_c("java -jar",
                              base::file.path(roadrunner_folder,
                                              "RoadRunner.jar"),
                              navigation_type, routes_file, distances_file,
                              map_path, temp_map_folder,
                              base::nrow(driving_distances),
                              err_file_path,
                              sep = " ")
    # on Windows it is better to run Java via shell(), on Linux use system()
    if  (exists("shell", where = "package:base")) {
        run <- base::shell
    } else {
        run <- base::system
    }
    run(command)
    # read the result of GraphHopper in
    driving_distances <- readr::read_csv(distances_file, col_types = "iidddddd")
    if (base::nrow(driving_distances) >
        base::nrow(tidyr::drop_na(driving_distances)))
        warning("RoadRunner was not able to find some routes. See list of errors in ",
                err_file_path)
    # eliminate routes longer than max_dist and transform the table; max_dist
    # must be converted to meters
    if (remove_longer)
        driving_distances <- driving_distances %>%
        filter(driving_distance <= max_dist * 1e3)
    # delete temporary files
    file.remove(routes_file, distances_file)
    # return
    driving_distances %>%
        dplyr::select(from_id, to_id, driving_distance, driving_time)
}





# prepare data for one fuel ----------------------------------------------------

# is_true(x) takes a logical vector x and returns a logical vector that has TRUE
# for every TRUE in x, and FALSE otherwise
is_true <- function(x) {
    dplyr::case_when(x ~ TRUE, TRUE ~ FALSE)
}


# test station existance, see prepare_average_fuel_data()
station_exists <- function(x, max_lag) {
    len <- base::length(x)
    result <- base::vector(mode = "logical", length = len)
    for (k in base::seq_along(x))
        result[k] <- base::any(x[base::max(c(1, (k - max_lag))):base::min(c(len, (k + max_lag)))])
    result
}


# prepare_daily_fuel_data() selects one fuel and creates a dataset of daily
# (interpolated) observations; all misssing values are interpolated from the
# last known observation
#
# inputs:
# - selected_fuel ... (character vector of lenght 1) name of the selected fuel
# - price_data ... tibble of price data
# - station_locations ... tibble of location data
# - station_names ... tibble of names data
# - features ... (optional) list of tibbles of feature data; every tibble must
#   contain date, id, and only one feature
#
# output:
# - a tibble with the union of columns of all provided tables (plus some
#   created) that have only rows on the selected fuel; every station has one
#   observation for every date in the period covered in any of the supplied
#   tables; missing values are substituted with the last known value
#
# details:
# - filling prices from the last known values does make sense for Pumpdroid data
#   where users usually submit only price changes; it is up to the user to asses
#   whether this procedure does make sense for other kinds of data
#
# usage:
#   natural_daily <- prepare_daily_fuel_data("NATURAL95", price_data,
#   station_locations, station_names, station_features)
prepare_daily_fuel_data <- function(selected_fuel,
                                    price_data,
                                    station_locations,
                                    station_names,
                                    features) {
    # features
    min_feature_date <- NA
    max_feature_date <- NA
    if (base::length(features) > 0) {
        min_feature_date <- purrr::map(features,
                                       ~base::min(.$date, na.rm = TRUE)) %>%
            base::unlist() %>%
            base::min(na.rm = TRUE) %>%
            lubridate::as_date()
        max_feature_date <- purrr::map(features,
                                       ~base::max(.$date, na.rm = TRUE)) %>%
            base::unlist() %>%
            base::max(na.rm = TRUE) %>%
            lubridate::as_date()
    }
    feature_names <- purrr::map_chr(features,
                                    ~names(dplyr::select(., - date, -id)))

    # find the beginning and the end
    from <- base::min(base::min(price_data$date, na.rm = TRUE),
                      base::min(station_names$date, na.rm = TRUE),
                      min_feature_date,
                      na.rm = TRUE)
    to <- base::max(base::max(price_data$date, na.rm = TRUE),
                    base::max(station_names$date, na.rm = TRUE),
                    max_feature_date,
                    na.rm = TRUE)

    # create the empty data structure -- all combinations of dates and ids
    station_data <- tidyr::crossing(date = base::seq(from, to, by = "days"),
                                    id = price_data %>%
                                        dplyr::filter(fuel %in% selected_fuel) %>%
                                        distinct(id) %>%
                                        dplyr::arrange(id) %>%
                                        pull(id),
                                    fuel = selected_fuel)

    # join in the price data
    station_data <- dplyr::left_join(station_data,
                                     price_data %>%
                                         dplyr::filter(fuel %in% selected_fuel) %>%
                                         dplyr::mutate(true_price_observation = TRUE),
                                     by = c("date", "id", "fuel")) %>%
        dplyr::mutate(true_price_observation = is_true(true_price_observation))

    # join in the brand names data
    station_data <- dplyr::left_join(station_data,
                                     station_names %>%
                                         dplyr::mutate(true_name_observation = TRUE),
                                     by = c("date", "id")) %>%
        dplyr::mutate(true_name_observation = is_true(true_name_observation))

    # join in the location data
    station_data <- dplyr::left_join(station_data,
                                     station_locations,
                                     by = "id")

    # join in the data on features
    for (k in base::seq_along(features)) {
        fn <- base::names(dplyr::select(features[[k]], -id, -date))
        tfn <- stringr::str_c("true_", fn, "_observation")
        station_data <- dplyr::left_join(station_data,
                                         features[[k]] %>%
                                             dplyr::mutate(!!tfn := TRUE),
                                         by = c("date", "id"))
        station_data[[tfn]] <- is_true(station_data[[tfn]])
    }

    # fill in vacancies in prices, brand names, and features
    var_names <- station_data %>% base::names() %>% base::setdiff("id")
    station_data <- station_data %>%
        dplyr::group_by(id) %>%
        dplyr::arrange(date, .by_group = TRUE) %>%
        tidyr::fill(tidyselect::all_of(var_names)) %>%
        dplyr::ungroup()

    # return
    station_data
}


# prepare_average_fuel_data() prepares data for one fuel in a given time
# granularity
#
# inputs:
# - selected_fuel ... (character vector of lenght 1) the name of the selected
#   fuel
# - by ... (character vector of lenght 1) the lenght of the periods as are used
#   in seq.Date(); default is "1 month", i.e. the function produces monthly data
#   unless told otherwise
# - exist_lag ... (integer vector) if provided, new column(s) are provided that
#   indicate whether the station probably existed and sold a given fuel in the
#   period
# - price_data ... tibble with price data
# - station_locations ... tibble with location data
# - station_names ... tibble with names data
# - ... ... (optional tibbles) tibble with features data; only three columns:
#   id, date, and one feature column such as sgs, nonstop etc.
# - remove_source ... (optional logical scalar, default is FALSE) if TRUE, it
#   removes columns source from the supplied data (the column is present when
#   the input data were prepared from Pumpdroid source data)
#
# output:
# - a tibble with columns that are unions of columns of the tibbles provided
#   with few exceptions:
#   - date is substituted with from_date and to_date (the beginning and the end
#     of the period)
#   - no_of_true_price_obs ... number of true observations on price for a given
#     station in a given period
#   - station_exists_# (where # is a number given in exist_lag) ... (logical)
#     TRUE if the station probably existed in the given period
#
# details:
# 1. the daily data are calculated with prepare_daily_fuel_data(); it filters
#    out all fuels except the given one and interpolated all
# 2. the data are grouped into a given periods; if the period is monthly or
#    quarterly, the beginning of the periods is January 1; if weekly, than
#    Monday before January 1; date is eliminated and from_date and to_date added
# 3. some values are "averaged":
#    - price ... mean price is calculated from the interpolated prices
#    - no_of_true_price_obs ... number of true price observations within the
#      period is added
#    - name ... the most common value is used (the first one if there is a tie)
#    - features ... the most common value is used (the first one if there is a
#      tie)
# 4. logical values on probable existence are added; if there is any price
#    observation within the period and a given number of days before and after
#    it, the station probably exists and sells the given fuel even within the
#    period; the days around are used 1) for very short periods and 2) because
#    of marginal stations with very few observations
# 5. all other values are supposed to be constant (e.g. address or gps); the
#    first value within the period is taken
#
# notes:
# - the function works well for by in weeks, months, quarters, and years;
#   if one wants to use it for its mutliples (e.g. 2 weeks), it must be
#   updated
# - some of the stations may not have sufficiently reliable data (it may e.g.
#   exists only for some time or may have only several true observations; these
#   stations should be removed from the data
#
# usage:
#   natural_monthly <- prepare_average_fuel_data(p"NATURAL95",
#       by = "1 month", exist_lag = 30L, price_data, station_locations,
#       station_names)
# or
#   natural_monthly <- prepare_average_fuel_data(p"NATURAL95",
#       by = "1 month", exist_lag = 30L, price_data, station_locations,
#       station_names, station_sgs, station_nonostop)
prepare_average_fuel_data <- function(selected_fuel,
                                      by = "1 month",
                                      exist_lag = c(0L, 14L, 30L),
                                      price_data,
                                      station_locations,
                                      station_names,
                                      ...,
                                      remove_source = FALSE) {
    # features
    features <- list(...)

    # remove source variable from tables
    if (remove_source) {
        price_data <- price_data %>% dplyr::select(-source)
        station_locations <- station_locations %>% dplyr::select(-source)
        station_names <- station_names %>% dplyr::select(-source)
        features <- map(features, ~dplyr::select(., -source))
    }

    # feature names
    feature_names <- purrr::map_chr(features,
                                    ~base::names(dplyr::select(., - date, -id)))

    # prepare daily data for one fuel
    daily_data <- prepare_daily_fuel_data(selected_fuel,
                                          price_data,
                                          station_locations,
                                          station_names,
                                          features)

    # dates
    # note: this algorithm works well for weeks, months, quarters, and years;
    # if one wants to use it for its mutliples (e.g. 2 weeks), it must be
    # updated
    min_date <- base::min(daily_data$date)
    max_date <- base::max(daily_data$date)
    from <- lubridate::ymd(stringr::str_c(lubridate::year(min_date), "-01-01"))
    to <- lubridate::ymd(stringr::str_c(lubridate::year(max_date), "-12-31"))
    czech_wday <- function(x) {
        x <- lubridate::wday(x)
        dplyr::if_else(x == 1, 7, x - 1)
    }
    if (stringr::str_detect(by, "week")) {
        from <- from - (czech_wday(from) - 1)
        to <- to + (7 - czech_wday(to))
    }

    # add start-date, end-date, and middle-date for each period
    from_dates <- base::seq.Date(from = from, to = to, by = by)
    to_dates <- c(from_dates[-1] - 1, base::as.Date(to))
    min_idx <- base::max(c(base::which(from_dates >= min_date)[1] - 1, 1))
    max_idx <- base::min(c(base::which(!(to_dates <= max_date))[1],
                           base::length(to_dates)))
    from_dates <- from_dates[min_idx:max_idx]
    to_dates <- to_dates[min_idx:max_idx]
    from_dates_added <- c(from_dates, lubridate::ymd("6000-12-31"))
    daily_data <- daily_data %>%
        dplyr::mutate(from_date = base::as.Date(NA),
                      to_date = base::as.Date(NA))
    for (k in base::seq_along(from_dates)) {
        idx <- daily_data$date >= from_dates_added[k] &
            daily_data$date < from_dates_added[k + 1]
        daily_data$from_date[idx] <- from_dates[k]
        daily_data$to_date[idx] <- to_dates[k]
    }

    # most_often(x) returns the most ususual value of x; it two values are
    # observed the same number of times, then the first observed value is used
    most_often <- function(x) {
        tab <- base::table(x)
        if (base::length(tab) == 0)
            return(as(NA, base::typeof(x)))  # necessary to be type-stable
        max_obs <- base::max(tab)
        alts <- base::unique(base::names(tab)[base::which(tab == max_obs)])
        x[base::which(x %in% alts)[1]]
    }

    # group by id and date and compute the representing value for the period
    group_names <- c("id", "from_date", "to_date")
    aggragates_names <- c("price", "no_of_true_price_obs", "name",
                          feature_names)
    station_data <- daily_data %>%
        dplyr::group_by(id, from_date, to_date) %>%
        summarise(price = base::mean(price, na.rm = TRUE),
                  no_of_true_price_obs = base::sum(true_price_observation,
                                                   na.rm = TRUE),
                  name = most_often(name),
                  .groups = "drop")
    ftab <- daily_data %>%
        dplyr::group_by(id, from_date, to_date) %>%
        dplyr::summarise(dplyr::across(dplyr::all_of(feature_names),
                                       most_often),
                         .groups = "drop")
    station_data <- dplyr::left_join(station_data, ftab,
                                     by = c("id", "from_date", "to_date"))

    # determine whether the station existed or not
    exists_prefix <- "station_exists_"
    exists_names <- stringr::str_c(exists_prefix, exist_lag)
    daily_data <- daily_data %>%
        dplyr::group_by(id) %>%
        dplyr::arrange(date, .by_group = TRUE)
    for (l in exist_lag) {
        var_name <- stringr::str_c(exists_prefix, l)
        daily_data <- daily_data %>%
            dplyr::mutate(!!var_name := station_exists(true_price_observation,
                                                       l))
    }
    daily_data <- dplyr::ungroup(daily_data)
    station_existence <- daily_data %>%
        dplyr::group_by(id, from_date, to_date) %>%
        dplyr::summarise(dplyr::across(dplyr::all_of(exists_names), base::any),
                         .groups = "drop")

    # station other
    other_names <- base::setdiff(
        base::names(daily_data),
        base::unique(c(group_names,
                       aggragates_names,
                       exists_names,
                       "date",
                       "true_price_observation",
                       "true_name_observation",
                       str_c("true_", feature_names, "_observation")))
    )
    other_data <- daily_data %>%
        dplyr::group_by(id, from_date, to_date) %>%
        dplyr::summarise(dplyr::across(dplyr::all_of(other_names), ~.[1]),
                         .groups = "drop")

    # joining
    station_data <- dplyr::left_join(station_data, station_existence,
                              by = c("id", "from_date", "to_date"))
    station_data <- dplyr::left_join(station_data, other_data,
                              by = c("id", "from_date", "to_date"))

    # return
    station_data %>%
        dplyr::select(
            tidyselect::all_of(c("from_date", "to_date", "id", "fuel",
                                 "name", "price", "no_of_true_price_obs",
                                 feature_names)),
            everything())
}





# voronoi polygony, spatial clusters atd. --------------------------------------

# see: https://github.com/r-spatial/sf/issues/824


# get_voronoi_polygons() splits the country into voronoi polygons, identifies
# station in neighboring polygons, and optionally computes the polygons' area
#
# inputs:
# - station_data ... tibble created by prepare_average_fuel_data() and filtered
#   for just one period of time
# - arccr_folder ... (character scalar) path to folder where ARC CR
#   administrative data are stored as shapefiles, i.e. path to
#   "AdministrativniCleneni_v13.gdb"
# - layer ... (character scalar) layer of the shape of the Czech Republic,
#   default is "StatPolygon"
# - existence ... (optional character scalar) name of the logical column in
#   station_data tibble which indicates which stations existed and sold the
#   selected fuel in the period; if not given, it is assumed that all stations
#   existed
# - keep_geometry ... (logical scalar, default is FALSE) if TRUE, then an object
#   of class sf is returned; if FALSE (default), a tibble without geometry is
#   returned
#
# output:
# - tibble (or sf) with four or five columns:
#   - id ... stations' ids
#   - voronoi_neighbors_id ... list of ids of the stations whose voronoi
#     polygons are neighbors of the station's polygon
#   - station_exists ... (logical) whether the staton exists or not in the
#     period
#   - voronoi_area ... area of the voronoi polygons clipped to the shape of the
#     Czech Republic in km^2
#   - border_station ... (logical) TRUE if the polygon touches the country
#     border
#   - geometry ... (returned only when keep_geometry is TRUE) voronoi polygons
#     (useful for drawing and testing)
#
# details:
# - if any station does not exist or sell the selected fuel in the period (its
#   existence value is FALSE), its values are NA and the station is ignored when
#   the polygons are constructed
#
# usage:
#   natural_monthly <- prepare_average_fuel_data(price_data, station_locations,
#       station_names, station_features, "NATURAL95")
#   natural_sel <- natural_monthly %>% dplyr::filter(from_date == "2017-05-01")
#   vor2 <- get_voronoi_polygons(natural_sel,
#       "~/map_shapes/AdministrativniCleneni_v13.gdb",
#       existence = "station_exists_90")
#
# TODO: funkce get_voronoi_polygons() vypisuje varování (nevím proč) -- opravit
get_voronoi_polygons <- function(station_data,
                                 arccr_folder, layer = "StatPolygon",
                                 existence = NULL, keep_geometry = FALSE) {
    # existence
    if (is.null(existence)) {
        station_data$station_exists <- TRUE
    } else {
        station_data$station_exists <- station_data[[existence]]
    }

    # convert to spatial object and set CRS to Krovak
    station_data <- station_data %>%
        dplyr::select(id, lon, lat, station_exists) %>%
        sf::st_as_sf(coords = c("lon", "lat"), crs = 4326) %>%
        sf::st_transform(5514)

    # get the country shape
    country <- sf::st_read(arccr_folder, layer,
                           stringsAsFactors = FALSE, quiet = TRUE) %>%
        sf::st_transform(5514) %>%
        sf::st_geometry() %>%
        sf::st_cast("POLYGON")

    # filter the existing stations and convert to MULTIPOLYGON
    stations <- station_data %>%
        dplyr::filter(station_exists) %>%
        sf::st_geometry() %>%
        base::do.call(c, .)

    # construct the voronoi polygons and add stations' ids
    voronoi <- sf::st_voronoi(stations, envelope = country) %>%
        sf::st_collection_extract("POLYGON") %>%
        sf::st_sf(crs = 5514)
    voronoi <-  sf::st_join(voronoi,
                            station_data %>% dplyr::filter(station_exists)) %>%
        dplyr::arrange(id)

    # crop the polygons to the shape of the country -- it is assumed that
    # polygons are not neighbors if they touch abroad only; it is necessary to
    # prefilter the polygons because cropping is a costly operation and
    # preselecting makes it much faster
    bordering <- sf::st_intersects(voronoi,
                                   country %>% sf::st_cast("LINESTRING"),
                                   sparse = FALSE)[, 1]
    # the following line is necessary to remove warning "attribute variables
    # are assumed to be spatially constant throughout all geometries"
    voronoi$geometry[bordering] <- sf::st_intersection(
        sf::st_geometry(voronoi[bordering, ]),
        country
    )
    voronoi$border_stations <- bordering

    # calculate polygons' area
    voronoi$voronoi_area <- as.double(sf::st_area(voronoi)) / 1e6  # in km^2

    # add neigbors
    neighbors <- sf::st_intersects(voronoi)
    voronoi$neighbors <- purrr::map(base::seq_along(neighbors),
                                    ~base::setdiff(neighbors[[.]], .))
    voronoi$voronoi_neighbors_id <- purrr::map(voronoi$neighbors,
                                               ~voronoi$id[.])

    # return selected columns only and drop the geometry (unless the user says
    # otherwise)
    result <- dplyr::left_join(station_data %>% sf::st_drop_geometry(),
                               voronoi %>% sf::st_drop_geometry(),
                               by = c("id", "station_exists")) %>%
        dplyr::select(id, station_exists, border_stations,
                      voronoi_neighbors_id, voronoi_area) %>%
        dplyr::mutate(voronoi_neighbors_id = purrr::map_if(voronoi_neighbors_id,
                                                           is.null,
                                                           ~integer(0)))
    if (keep_geometry)
        result <- dplyr::left_join(result, voronoi %>%
                                       dplyr::select(id), by = "id") %>%
        sf::st_sf(sf_column_name = "geometry")
    result
}


# vn(voronoi, i) takes the table returned by get_voronoi_polygons() and i = id
# of one station and returns a list of all its voronoi neighbors
#
# usage:
#   vor <- get_voronoi_polygons(natural_sel, arccr_folder,
#      existence = "station_exists_90")
#   vn(vor, 1000)
vn <- function(voronoi, i) {
    voronoi$voronoi_neighbors_id[voronoi$id == i][[1]]
}


# only_same_brand(station_data, i, x) takes station's id i, id of stations x
# (a numeric vector) and station_data (tibble with columns id and name at least)
# and returns selected stations from x that have the same brand as station i;
# when there is no such station, it returns numeric(0)
#
# usage:
#   only_same_brand(natural_selection, 20, c(1337, 1351, 1, 2, 3, 1360, 2841, 3019))
only_same_brand <- function(station_data, i, x) {
    i_brand_name <- station_data$name[station_data$id == i]
    ids <- station_data$id[(station_data$id %in% x) & (station_data$name == i_brand_name)]
    ids[!is.na(ids)]  # every NA name is different
}


# one_station_spatial_clustering() computes spatial clustering statistics for
# one station; station_data should be aggregated values for just one period;
# voronoi should be voronoi data created by get_voronoi_polygons() for
# station_data
#
# warnings:
# - this function should not be used by a user -- use add_spatial_clustering()
#   instead
#- row X <- unique(c(X, x)) is fast but not safe;
#   it has to append to X the members of x that are not in X at the same time,
#   and it must preserve the order of the items; it works now because of the
#   present implementation of unique(); however, it may change in the future
#   (even though it is exteremely unlikely)
#   if the implementation of unique() changes, the idiom muste be replaced with
#   something like
#       f <- function(X, x) {x <- base::setdiff(x, X); c(X, x)}; f(X, x)
#   it is not used now becuase it is much slower (12 instead of
#   4.6 microseconds) (remember: it is used many many times so speed matters)
#   see discussion here:
#   https://stackoverflow.com/questions/36413440/does-r-unique-always-return-values-in-same-order
one_station_spatial_clustering <- function(station_data, voronoi, i) {
    # vector of ids of the cluster under construction
    X <- base::integer(0)
    # list of all clusters
    s <- base::list()
    # vector of ids of station i and its neighbors
    v <- c(i, vn(voronoi, i))
    # number N_i in the sc formula; the number of i's direct neighbors plus 1
    Ni <- base::length(v)

    # repeat while there are any neighbors that are not allocated into clusters;
    # this creates a new cluster
    while (base::length(v) > 0) {
        # basis of a new cluster is a direct neighbor of station i (station i in
        # the first step)
        X <- v[1]
        # remove the station that is handled now from vector v
        v <- v[-1]
        # number of already checked elements of X
        xs <- 0
        # repeat while there are any elements of X that has not been handled yet
        # = any member for which their neighbors were not added to their cluster
        # (if they have the same brand)
        while (xs < base::length(X)) {
            # which members of X should be handled (for cycle xstart:lenght(X))
            xstart <- xs + 1
            # present members of X will be handled by for cycle; they will be
            # solved in the next while cycle
            xs <- base::length(X)
            # repeat for all present members of X;
            # for each member of X (that has not been handled yet) its neighbors
            # are found; only those with the same brand are retained;
            # they are removed from v and added to X;
            # the new members of X will be handled in a new iteration of the
            # inner while cycle
            for (k in xstart:base::length(X)) {
                # the presently handled member of X
                z <- X[k]
                # ids of z's neighbors that have the same brand name
                x <- only_same_brand(station_data, z, vn(voronoi, z))
                # remove items in x from v
                v <- base::setdiff(v, x)
                # add items x into X
                X <- base::unique(c(X, x))
            }
        }
        # add a new cluster into s; s[[j]] is indices of the cluster members
        s[[base::length(s) + 1]] <- X
    }

    # calculate spatial clustering number
    # number of clusters (M_i)
    Mi <- base::length(s)
    # number of items in individual clusters (k_{{m_i},i})
    kmii <- purrr::map_int(s, length)
    # spatial clustering measure (SC_i) for station i
    SCi <- base::sum(kmii)/Mi/Ni

    # area of station i's onw cluster (in km^2)
    cluster_area <- base::sum(voronoi$voronoi_area[voronoi$id %in% s[[1]]])

    # return
    list(sc = SCi, cluster_members_ids = s[[1]], cluster_area = cluster_area)
}


# add_spatial_clustering_for_one_period() adds Pennerstorfer's SC (spatial
# clustering measure) and some other statistics
#
# warnings:
# - this function is not to be used by the user directly -- the user should use
#   add_spatial_clustering() which has the same inputs and output but checks its
#   inputs
# - it is expected that station_data are aggregate values for just one period
# - it does not correct for brands owned by the same entity
add_spatial_clustering_for_one_period <- function(station_data,
                                                  arccr_folder,
                                                  layer = "StatPolygon",
                                                  existence = NULL) {
    # get voronoi polygons and related statistics
    voronoi <- get_voronoi_polygons(station_data, arccr_folder, layer,
                                    existence = existence)
    # for each station, calculate its spatial clustering and related statistics
    res <- purrr::map(station_data$id,
                      ~one_station_spatial_clustering(station_data, voronoi, .))
    # add the results to statation_data
    station_data <- station_data %>%
        dplyr::left_join(voronoi, by = "id") %>%
        dplyr::mutate(sc = purrr::map_dbl(res, "sc"),
               cluster_area = purrr::map_dbl(res, "cluster_area"),
               cluster_members_ids = purrr::map(res, "cluster_members_ids"))
    # set the statistics for non-existing stations to NA
    station_data$sc[!station_data$station_exists] <- NA
    station_data$cluster_area[!station_data$station_exists] <- NA
    station_data$cluster_members_ids[!station_data$station_exists] <- NA
    station_data$voronoi_neighbors_id[!station_data$station_exists] <- NA
    # return
    station_data %>%
        dplyr::select(-station_exists)
}


# add_spatial_clustering() adds Pennerstorfer's spatial clustering measure (SC)
# and some related statistics to table station_data (tibble of aggregated data)
#
# inputs:
# - station_data ... (tibble) table created by prepare_average_fuel_data();
#   it includes aggregated values for selected periods, e.g. months; it must
#   include columns from_date and to_date instead of date
# - arccr_folder ... (character scalar) path to folder with ARC CR data
#   "AdministrativniCleneni_v13.gdb"
# - layer ... (character scalar) layer with the country's border shapefile;
#   default is "StatPolygon"
# - existence ... (optional, character scalar) name of logical column in table
#   station_data which says which stations existed and sold the selected fuel in
#   the given period
# - same_group ... (tibble) correction table which says which brands are owned
#   by a same entity, and hence do no compete with each other, see
#   unite_same_owner_station_names()
#
# output:
# - the same tibble as station_data but the following columns are added:
#   - border_station ... (logical) TRUE if the station's voronoi poygon touches
#     border of the country
#   - voronoi_neighbors_id ... (integer vector) ids of all stations whose
#     voronoi polygons are neighbors of the polygon of this station
#   - voronoi_area ... (double) area of stations' voronoi polygons in km^2
#   - sc ... (double) Pennerstorfer's spatial clustering measure (SC)
#   - cluster_area ... (double) area of the station clusters' polygon in km^2
#   - cluster_members_ids ... (integer vector) ... ids of stations that are
#     parts of the station's cluster including its own id
#   if the station does not exist in the period, then its statistics are NA
#
# details:
# - the statistics are computed for each period separately; it must be so
#   because some stations cease to exist, start their business, or change their
#   ownership
# - if a station does not exist or does not sell the selected kind of fuel (as
#   indicated in column existence), it is not used in voronoi polygons
#   computation and computation of statistics that are based on the polygons
# - when the station's name is NA, then it is never a part of any cluster, or
#   more precisely, it is always its own cluster than never includes any other
#   station
# - the function is a wrapper for add_spatial_clustering_for_one_period() -- it
#   runs it individually for each period defined by column from_date and then
#   joins the result into one table
#
# warnings:
# - see warning to one_station_spatial_clustering()
# usage:
#   natural_monthly <- add_spatial_clustering(natural_monthly,
#       "~/map_shapes/AdministrativniCleneni_v13.gdb",
#       existence = "station_exists_90")
add_spatial_clustering <- function(station_data,
                                   arccr_folder,
                                   layer = "StatPolygon",
                                   existence = NULL,
                                   same_group = NULL) {
  station_data %>%
    unite_same_owner_station_names(correction_table = same_group) %>%
    dplyr::group_by(from_date) %>%
    dplyr::group_split() %>%
    furrr::future_map(~ add_spatial_clustering_for_one_period(., arccr_folder,
                                                              layer, existence),
               .progress = TRUE) %>%
    dplyr::bind_rows() %>%
    restore_same_owner_names(station_data)
}


# plot_voronoi() plot the voronoi tessalation for one station; it can be used to
# demonstrate or test the computation of Pennerstorfer's spatial clustering
# measure
#
# inputs:
#   - stations ... station data created by add_spatial_clustering() but filtered for one
#       period
#   - voronoi ... (optional) voroni tessalation created by get_voroi_polygons() for
#       stations; it must keep the geometry
#   - id ... (integer) the id of the station
#   - buffer ... (one integer or four integers) how much should the bounding box of the
#       station's bounding box be enhanced to the left, bottom, right, and top [in km]
#   - arccr_folder, layer ... if voronoi is not given, these two must be given, see
#       get_voronoi_polygons()
#
# output:
#   a graph
#
# usage:
#   # prepare data:
#   natural_selection <- natural_monthly %>% dplyr::filter(from_date == "2015-07-01")
#   voronoi <- get_voronoi_polygons(natural_selection,
#       "map_shapes/AdministrativniCleneni_v13.gdb",keep_geometry = TRUE)
#   # -- symmetric region
#   plot_voronoi(natural_selection, voronoi, 975, 8)
#   # -- asymmetric region
#   plot_voronoi(natural_selection, voronoi, 975, c(8, 8, 2, 2))
#   # when voronoi is not pre-calculated (it takes some time, so it is usually better to
#   # cache it)
#   plot_voronoi(natural_selection, id = 975, buffer = c(8,8, 2, 2),
#       arccr_folder = "map_shapes/AdministrativniCleneni_v13.gdb")
#
# BEWARE:
#   1) the function ignores stations' existence
#   2) the function does not test that the data provided are correct
plot_voronoi <- function(stations, voronoi = NULL, id, buffer = 5,
                         arccr_folder, layer = "StatPolygon", scale = FALSE) {
    require(tmap)
    # construct the buffer
    if (length(buffer) == 1)
        buffer <- rep(buffer, 4)
    buffer <- buffer * c(-1, -1, 1, 1) * 1000
    # create voronoi if not given
    if (is.null(voronoi))
        voronoi <- get_voronoi_polygons(stations, arccr_folder, layer,
                                        keep_geometry = TRUE)
    # add brand names to voronois
    df <- dplyr::left_join(voronoi, stations %>% dplyr::select(id, name), by = "id")
    # compute the clip
    bb <- df[df$id == id, ] %>% st_bbox() + buffer
    clip <- bb %>% st_as_sfc()
    # plot it
    p <- df %>%
        st_intersection(clip) %>%
        tm_shape() +
        tm_polygons("name") +
        tm_text("name") +
        tm_legend(legend.outside = TRUE) +
        tm_shape(df[df$id == id, ]) +
        tm_borders(col = "red", lwd = 3)
    if (scale)
        p <- p + tm_scale_bar()
    p
}





# number of stations within some distances -------------------------------------

# add_competitors_in_distnaces_for_one_period() adds many columns with number of
# competitors and own stations within given driving distances (i.e. within
# concentric "rings", or more precisely "donuts")
#
# inputs:
# - station_data ... (tibble) table created by prepare_average_fuel_data() and
#   filtered so that they include only one period (defined by from_date)
# - driving_distances ... (tibble) table of driving distances created by
#   get_driving_distances()
# - distances ... (integer vector) distances, for which the number of
#   competitors and own stations should be counted (deafault is 0:10);
#   the user should use the default just to learn -- something like c(0, 3, 6)
#   would be better for a real analysis;
#   the number of stations is calculated in distances >= distance[1] and
#   < distance[2] and so on
# - existence ... (optional, character scalar) name of logical column in
#   station_data which indicates whether individual stations existed and sold
#   the selected fuel in the particular period; if not given, then it is assumed
#   that all stations exist
#
# output:
# - tibble with the same columns as station_data -- the following columns are
#   added
#   - other_stations_in_driving_distance_#1_#2 ... the number of competitors
#     (stations of other brands including all stations with brand NA, as NA is
#     always seen as a competitor) within a distance >= #1 and < #2
#   - own_stations_in_driving_distance_#1_#2 ... the number of own stations
#     (i.e. stations of the same brand) within a distance >= #1 and < #2;
#     two stations are not of the same brand if the brand differs or at least
#     one station has brand NA
#
# details:
# - if a station does not exist in the period, then both its number of
#   competitors and own stations is NA in each distance; it is the only wya how
#   NA can arise here: if a station has no neighbors within a distance, their
#   number is 0
#
# warnings:
# - this function is not to be used by the user directly -- the user should use
#   add_competitors_in_distnaces()
# - this function does not check its inputs
# - it does not correct for brands owned by the same entity
# - it is expected that station_data are aggregate values for just one period
add_competitors_in_distnaces_for_one_period <- function(station_data,
                                                        driving_distances,
                                                        distances = 0:10,
                                                        existence = NULL) {
    # existence
    if (is.null(existence)) {
        existence <- base::rep_len(TRUE, length.out = base::nrow(station_data))
    } else {
        existence <- station_data[[existence]]
    }

    # compute the number of competitors and same stations for every distance
    for (j in 2:base::length(distances)) {
        d1 <- distances[j - 1]
        d2 <- distances[j]
        m1 <- d1 * 1e3  # convert to meters
        m2 <- d2 * 1e3  #      ""
        vec_other <- base::rep(0, base::nrow(station_data))
        vec_same <- base::rep(0, base::nrow(station_data))
        # for each station
        for (k in 1:base::nrow(station_data)) {
            id <- station_data$id[k]
            # find all neighbors within the given distances (which are in meters)
            neighbors <- driving_distances$to_id[
                driving_distances$from_id == id &
                    driving_distances$driving_distance >= m1 &
                    driving_distances$driving_distance < m2
            ]
            # remove all neighbors that do not exist for one of two reasons:
            # 1) their existence is FALSE or
            # 2) they are not present in station_data (because they do not sell
            #    the selected fuel
            neighbors <- station_data$id[existence &
                                             (station_data$id %in% neighbors)]
            neighbors_same <- only_same_brand(station_data, id, neighbors)
            neighbors_other <- base::setdiff(neighbors, neighbors_same)
            # count the number of each kind of neighbors within the distances
            vec_other[k] <- base::length(neighbors_other)
            vec_same[k] <- base::length(neighbors_same)
        }
        vec_other[!existence] <- NA
        vec_same[!existence] <- NA
        station_data[ , stringr::str_c("other_stations_in_driving_distance_", d1, "_", d2)] <- vec_other
        station_data[ , stringr::str_c("own_stations_in_driving_distance_", d1, "_", d2)] <- vec_same
    }

    # return
    station_data
}


# add_competitors_in_distnaces() adds many columns with number of competitors
# and own stations within given driving distances (i.e. within concentric
# "rings", or more precisely "donuts")
#
# inputs:
# - station_data ... (tibble) table created by prepare_average_fuel_data() and
#   filtered so that they include only one period (defined by from_date)
# - driving_distances ... (tibble) table of driving distances created by
#   get_driving_distances()
# - distances ... (integer vector) distances, for which the number of
#   competitors and own stations should be counted (deafault is 0:10);
#   the user should use the default just to learn -- something like c(0, 3, 6)
#   would be better for a real analysis;
#   the number of stations is calculated in distances >= distance[1] and
#   < distance[2] and so on
# - existence ... (optional, character scalar) name of logical column in
#   station_data which indicates whether individual stations existed and sold
#   the selected fuel in the particular period; if not given, then it is assumed
#   that all stations exist
# - same_group ... (tibble) correction table which says which brands are owned
#   by a same entity, and hence do not compete with each other, see
#   unite_same_owner_station_names()
# - quiet ... (logical scalar) if zero is missing from distances and
#   quiet = TRUE, a warning is thrown; default is FALSE
#
# output:
# - tibble with the same columns as station_data -- the following columns are
#   added
#   - other_stations_in_driving_distance_#1_#2 ... the number of competitors
#     (stations of other brands including all stations with brand NA, as NA is
#     always seen as a competitor) within a distance >= #1 and < #2
#   - own_stations_in_driving_distance_#1_#2 ... the number of own stations
#     (i.e. stations of the same brand) within a distance >= #1 and < #2;
#     two stations are not of the same brand if the brand differs or at least
#     one station has brand NA
#
# details:
# - if a station does not exist in the period, then both its number of
#   competitors and own stations is NA in each distance; it is the only wya how
#   NA can arise here: if a station has no neighbors within a distance, their
#   number is 0
# - this function iterates add_competitors_in_distnaces_for_one_period()
#   for each period separately, and then joins the data together
add_competitors_in_distances <- function(station_data,
                                         driving_distances,
                                         distances = 0:10,
                                         existence = NULL,
                                         same_group = NULL,
                                         quiet = FALSE) {
  # correct thedistances
  distances <- base::sort(base::unique(distances))
  if (!(0 %in% distances) && !quiet)
    warning("The first 'donut' (distance 0) is missing from the setting.")
  # the computation
  station_data %>%
      # correct for the same owner
      unite_same_owner_station_names(correction_table = same_group) %>%
      dplyr::group_by(from_date) %>%
      dplyr::group_split() %>%
      furrr::future_map(
          ~ add_competitors_in_distnaces_for_one_period(., driving_distances,
                                                        distances,
                                                        existence),
          .progress = TRUE
      ) %>%
    dplyr::bind_rows() %>%
    restore_same_owner_names(station_data)  # restore the true brand names
}





# distance to the closest competitor (driving distance) ------------------------

# add_distance_to_closest_competitor_for_one_period() takes tibble station_data
# created by prepare_average_fuel_data() and adds a distance to the closest
# competitor and the closest station of the same brand; both distances are in
# kilometers
#
# inputs:
# - station_data ... (tibble) table created by prepare_average_fuel_data() and
#   filtered for one period (defined by from_date)
# - driving_distances ... (tibble) table created by get_driving_distances()
# - existence ... (optional, character scalar) name of a column within
#   station_data which says whether individual stations exist in this period
#   (TRUE) or not (FALSE)
#
# output:
# - (tibble) a same station as station_data plus two new columns:
#   - distance_to_closest_competitor ... (double) driving distance to the
#     closest competitor (i.e. station of any other brand; if any station has
#     brand NA, the pair of stations is always considered as competitors); the
#     value may be Inf when there is no competitor in driving_distances table --
#     if this happens, the table driving_distances must be recreated with
#     a higher parameter max_dist
#   - distance_to_closest_same_brand ... (double) driving distance to the
#     closest station of the same brand (if at least one of the stations has
#     brand NA, the pair of stations is considered as competitors); the value
#     may be Inf either when there is no other station of the same brand or when
#     it is not covered in driving_distances; thus value Inf need not be problem
#     here
#
# details:
# - the distances are driving distances in meters
#
# warnings:
# - this function does not check its inputs because it is not intended for end
#   user use; an end user should use function
#   add_distance_to_closest_competitor()
add_distance_to_closest_competitor_for_one_period <- function(station_data,
                                                              driving_distances,
                                                              existence = NULL) {
    # existence
    if (is.null(existence)) {
        station_data$.tmp_existence <- TRUE
    } else {
        station_data$.tmp_existence <- station_data[[existence]]
    }

    # add station names to driving distances -- must be done here because the
    # names and existence differ among periods
    driving_distances <- driving_distances %>%
        dplyr::left_join(station_data %>% dplyr::select(id, from_name = name),
                         by = c("from_id" = "id")) %>%
        dplyr::left_join(station_data %>%
                             dplyr::select(id, to_name = name,
                                           to_existence = .tmp_existence),
                         by = c("to_id" = "id"))

    # cycle over all stations in station_data
    distance_to_closest_competitor <- base::rep(NA, base::nrow(station_data))
    distance_to_closest_same_brand <- base::rep(NA, base::nrow(station_data))
    for (k in 1:base::nrow(station_data)) {
        id <- station_data$id[k]
        ids_name <- station_data$name[k]
        sel <- driving_distances %>%
            dplyr::filter(from_id == id, to_existence)
        # distance to the closest competitor (name NA is always considered as
        # another brand); may be Inf if there is no competitor in
        # driving_distances because it was created with to short distance; if
        # this happens, driving_distances must be re-generated
        if (is.na(ids_name)) {
            ssel <- sel
        } else {
            ssel <- sel %>% dplyr::filter(to_name != ids_name | is.na(to_name))
        }
        distance_to_closest_competitor[k] <-
            if (base::nrow(ssel) > 0) base::min(ssel$driving_distance) else Inf
        # distance to the closest station of the same brand;
        # may be Inf if there is no station of the same brand even if the
        # driving_distances table is calculated for a sufficient distance
        ssel <- sel %>% dplyr::filter(to_name == ids_name)
        distance_to_closest_same_brand[k] <-
            if (base::nrow(ssel) > 0) base::min(ssel$driving_distance) else Inf
    }

    # for non-existing stations, set distances to NA
    distance_to_closest_same_brand[!station_data$.tmp_existence] <- NA
    distance_to_closest_competitor[!station_data$.tmp_existence] <- NA

    # add, convert from meters to kilometers, and return
    station_data$distance_to_closest_competitor <-
        distance_to_closest_competitor / 1e3
    station_data$distance_to_closest_same_brand <-
        distance_to_closest_same_brand / 1e3
    station_data %>%
      dplyr::select(-.tmp_existence)
}


# function add_distance_to_closest_competitor() takes table station_data created
# by function prepare_average_fuel_data() and adds to it driving distance to the
# closest competitor and to the closest station of the same brand, both in
# kilometers
#
# inputs:
# - station_data ... (tibble) table created by function
#   prepare_average_fuel_data()
# - driving_distances ... (tibble) table created by function
#   get_driving_distances()
# - existence ... (optional, character scalar) name of the column in table
#   station_data which says whether the particular stations exist in given
#   periods (TRUE) or not (FALSE)
# - same_group ... (tibble) correction table which says which brands have the
#   same owner, and hence do not compete with each other, see
#   unite_same_owner_station_names()
#
# output:
# - (tibble) same table as station_data plus two new columns:
#   - distance_to_closest_competitor ... (double) driving distance to the
#     closest competitor (i.e. station of other brand; if any station has brand
#     NA, the pair of stations is always considered as competitors) in
#     kilometers; the value may be Inf when there is no competitor in
#     driving_distances table -- if this happens, the table driving_distances
#     must be recreated with a higher parameter max_dist
#   - distance_to_closest_same_brand ... (double) driving distance to the
#     closest station of the same brand in kilometers (if at least one of the
#     stations has brand NA, the pair of stations is considered as competitors);
#     the value may be Inf either when there is no other station of the same
#     brand or when it is not covered in driving_distances; thus value Inf need
#     not be problem here
add_distance_to_closest_competitor <- function(station_data,
                                               driving_distances,
                                               existence = NULL,
                                               same_group = NULL) {
    station_data %>%
        unite_same_owner_station_names(correction_table = same_group) %>%
        dplyr::group_by(from_date) %>%
        dplyr::group_split() %>%
        furrr::future_map(
            ~ add_distance_to_closest_competitor_for_one_period(.,
                                                                driving_distances,
                                                                existence),
            .progress = TRUE
        ) %>%
        dplyr::bind_rows() %>%
        restore_same_owner_names(station_data)
}





# correction for the same owner of several brands ------------------------------

# unite_same_owner_station_names(station_data, correction_table) corrects in
# table station_data created by prepare_average_fuel_data() names of stations
# owned by the same entity; the corrections are taken from correction_table
#
# inputs:
# - station_data ... (tibble) table created by prepare_average_fuel_data()
# - correction_table ... (tibble) table of periods when several stations have
#   a same owner; the table must have four columns:
#   - from_date, to_date ... (charackter) beginning and end of the period when
#     the staitions have a same owner; date format is YYYY-MM-DD
#   - group_name ... (character) name of the group -- the individual brand names
#     will be substituted with this group_name
#   - name ... (character) brand name that should be (temporarily) substituted;
#   example of the table:
#     station_correction_table <- tribble(
#       ~from_date, ~to_date, ~group_name, ~name,
#       "2015-01-01", "2017-05-01", "mol", "papoil")
#   as the group name "mol" is a brand name in the data, it is not necessary to
#   substitute it explicitly
#
# output:
# - tibble with the same structure and size as station_data, but where some
#   brand names are replaced with the group names
#
# details:
# - the reason forthis function is that some brands have a same owner, and hence
#   they do not compete with each other; in the same time, it is not possible
#   to unify the brand names in the data because they may have different price
#   levels, and hence they must have different intercepts in regression
# - the correction is done only in the periods that are fully covered in the
#   periods in correction_table; if any part of the true period lies outside
#   (even by just one day), then the correction is not done
#
# warnings:
# - this function is not to be used by an end user; it is to be used only in
#   functions that compute various competition measures
# - after the usage, the original brand names must be recovered by
#   restore_same_owner_names()
#
# TODO: aplikovat ve všech funkcích, které stojí na kontrole jmen
# TODO: rozhodnout, jak správně nastavit podmínku, tj. má tam patřit celá perioda, část?
unite_same_owner_station_names <- function(station_data, correction_table = NULL) {
    # if there is no correction table, return the original table
    if (is.null(correction_table))
        return(station_data)

    # convert dates to Dates
    correction_table <- correction_table %>%
        dplyr::mutate(from_date = base::as.Date(from_date),
                      to_date = base::as.Date(to_date))

    # do the correction
    for (k in base::seq_len(base::nrow(correction_table)))
        station_data$name[
            station_data$name == correction_table$name[k] &
                station_data$from_date >= correction_table$from_date[k] &
                station_data$to_date <= correction_table$to_date[k]
        ] <-
        correction_table$group_name[k]

    # return
    station_data
}


# restore_same_owner_names(current_station_data, original_station_data) restores
# the original brand names in station_data created by
# prepare_average_fuel_data() and temporarily corrected by
# unite_same_owner_station_names()
#
# warnings:
# - this function is not to be used by an end user; it is to be used only in
#   functions that compute various competition measures
restore_same_owner_names <- function(current_station_data, original_station_data) {
    dplyr::left_join(current_station_data,
                     original_station_data %>%
                         dplyr::select(from_date, to_date, id,
                                       .tmp_restore_same_owner_names = name),
                     by = c("from_date", "to_date", "id")) %>%
        dplyr::mutate(name = .tmp_restore_same_owner_names) %>%
        dplyr::select(-.tmp_restore_same_owner_names)
}



# fixed effects for cross-sectional data ---------------------------------------

# add_city_effects(data) add dummies for city sizes
#
# inputs:
# - data ... (tibble) table created by prepare_average_fuel_data(); it must
#   include column "obec"
#
# output:
# - the same tibble as data but several logical columns are added:
#   - praha ... TRUE if the station is in Prague and FALSE otherwise
#   - brno ... TRUE if the station is in Brno and FALSE otherwise
#   - ostrava ... TRUE if the station is in Ostrava and FALSE otherwise
#   - mesto300 ... TRUE if the station is in cities with more than 300k
#       inhabitants and less than 1M inhabitants (currenty Brno or Ostrava) and
#       FALSE otherwise
#   - mesto100 ... TRUE if the station is in cities with more than 100k
#       and less than 300k inhabitants and FALSE otherwise
#   - mesto50 ... TRUE if the station is in cities with more than 50k
#       inhabitants and less than 100k inhabitants and FALSE otherwise
#   - mesto20 ... TRUE if the station is in cities with more than 20k
#       inhabitants and less than 50k inhabitants and FALSE otherwise
#
# warnings:
# - the function uses a fixed list of cities; hence it must be updated from time
#   to time
#
# usage:
#   natural_monthly <- add_city_effects(natural_monthly)
add_city_effects <- function(data){
    # zdroj: http://cs.wikipedia.org/wiki/Seznam_m%C4%9Bst_v_%C4%8Cesku_podle_po%C4%8Dtu_obyvatel
    mesta300 <- c("brno", "ostrava")
    mesta100 <- c("plzeň", "liberec", "olomouc")
    mesta50 <- c("ústí nad labem", "hradec králové", "české budějovice",
                 "pardubice", "havířov", "zlín", "kladno", "most", "karviná",
                 "opava", "frýdek-místek", "karlovy vary", "jihlava", "teplice",
                 "děčín")
    mesta20 <- c("chomutov", "přerov", "jablonec nad nisou", "mladá boleslav",
                 "prostějov", "třebíč", "česká lípa", "třinec", "tábor",
                 "znojmo", "příbram", "cheb", "orlová", "kolín", "trutnov",
                 "písek", "kroměříž", "šumperk", "vsetín", "valašské meziříčí",
                 "litvínov", "uherské hradiště", "hodonín", "český těšín",
                 "břeclav", "krnov", "litoměřice", "sokolov", "nový jičín",
                 "havlíčkův brod", "chrudim", "strakonice", "kopřivnice",
                 "klatovy", "žďár nad sázavou", "bohumín", "jindřichův hradec",
                 "vyškov", "blansko", "kutná hora", "náchod", "jirkov")
    .obec_code_name = stringr::str_to_lower(data$obec)
    result <- data %>% dplyr::mutate(praha = (.obec_code_name == "praha"),
                                     brno = (.obec_code_name == "brno"),
                                     ostrava = (.obec_code_name == "ostrava"),
                                     mesto300 = (.obec_code_name %in% mesta300),
                                     mesto100 = (.obec_code_name %in% mesta100),
                                     mesto50 = (.obec_code_name %in% mesta50),
                                     mesto20 = (.obec_code_name %in% mesta20))
    return(result)
}


# find_big_brands_in_prepared_data() find brand names of the biggest brands
#
# inputs:
# - data ... (tibble) table created by prepare_average_fuel_data filtered for
#   just one period (only one unique from_date)
# - min_n ... (positive numeric scalar) only brand names with min_n or more
#   stations are reported
#
# output:
# - tibble with two columns:
#   - name ... brand names
#   - n ... number of the stations within the data
find_big_brands_in_prepared_data <- function(data, min_n = 10) {
    stopifnot("from_date" %in% base::names(data),
              base::length(base::unique(data$from_date)) == 1)
    stopifnot("name" %in% base::names(data))
    data %>%
        dplyr::group_by(name) %>%
        dplyr::summarise(n = n(), .groups = "drop") %>%
        dplyr::arrange(desc(n)) %>%
        dplyr::filter(!is.na(name), n >= min_n)
}


# add_big_brand_name_effects() adds factor brand_name to data -- used for
# cross-sectional econometrics only
#
# inputs:
# - data ... (tibble) table created by prepare_average_fuel_data
# - big_brand_names ... (optional character vector or tibble created by
#   find_big_brands_in_prepared_data()) brand names that should have a level in
#   the factor (i.e. a uniquie intercept shift in the regression); if not given,
#   it is created from the data -- brands with min_n or more stations are
#   listed; all other stations have level "other", which is the reference level
# - min_n ... (optional positive numeric scalar) if big_brand_names are selected
#   from the data, only brand names with min_n or more stations are covered
#   (default is 10)
add_big_brand_name_effects <- function(data,
                                       big_brand_names = NULL,
                                       min_n = 10) {
    if (is.null(big_brand_names))
        big_brand_names <- find_big_brands_in_prepared_data(data, min_n = min_n)
    if (tibble::is_tibble(big_brand_names) &&
        "name" %in% base::names(big_brand_names))
        big_brand_names <- big_brand_names$name
    big_brand_names <- unique(big_brand_names)
    data %>%
        dplyr::mutate(brand_name = dplyr::if_else(name %in% big_brand_names,
                                                  name,
                                                  "other"),
                      brand_name = base::factor(brand_name,
                                                levels = unique(c("other",
                                                                  big_brand_names))))
}
