# -------------------------------------
# Script: preparatory_functions_pumpdroid.R
# Author: Michal Kvasnička
# Purpose: This script contains functions used to prepare data provided by
#   PumpDroid.
# Notes: For usage see script prepare_pumpdroid_data.R.
#
# Copyright(c) Masaryk University
# -------------------------------------



# load libraries
library(tibble)
library(dplyr)
library(purrr)
library(tidyr)
library(stringr)
library(lubridate)
library(readr)
library(sf)





# reading PumpDroid data -------------------------------------------------------

# Functions in this section are used to read the data provided by PumpDroid and
# convert them into a standard format used later.
#
# It is supposed that Pumpdroid provided two types of files:
#
# 1. files including price data -- for their structure see help on
# read_pumpdroid_price_data() function
#
# 2. files including data on stations (their brand, location, etc) -- for their
# structure see help on read_pumpdroid_station_data() function



# read_pumpdroid_price_data(path_to_folder) reads the quasi-CSV files (possibly
# gzipped) of price data provided by PumpDroid and returns the standard price
# data set.
#
# inputs:
# - path_to_folder ... path to a directory where all .csv and .csv.gz files
#   contain the price data provided by Pumpdroid
#
# output:
# - a tibble of all stations (even abroad or probably non-existing, all fuels,
#   only days when there is a record in Pumpdroid database); it contains the
#   following columns:
#   - date ... the data of the observation (Date)
#   - id ... the id of the gas station (integer)
#   - fuel ... the name of the fuel (character)
#   - price ... the price fo the fuel at the station on the given day (double)
#
# usage:
#   price_data <- read_pumpdroid_price_data("rawdata/pumproid_price_data_folder")
#
# This function must solve the following problems:
#

# 1) The CSV files provided by PumpDroid are poorly formated. Each line contains
# observations for one station and one day. However, it can include any number
# of fuels. Thus each line consists of any number of quaternions (station id,
# date, price, fuel type). These items are separated by semicolons. All
# quaternions are separated by semicolons too. Moreover, there are some
# redundancies (duplicities) in the data.
#
# 2) There are two or more prices for some combinations of day, gas station id,
# and fuel type. I try to keep the newest information. I expect that all data
# files have the same name except the year number and they are in the same
# folder. I also expect that the lower row is the newer one. If it is not, then
# I just keep one radnom value.
read_pumpdroid_price_data <- function(path_to_folder) {
    # read all CSV files in path folder
    data_files <- base::list.files(path = path_to_folder,
                                   pattern = ".*\\.csv(\\.gz)+",
                                   full.names = TRUE)

    # this function processes one data file
    process_one_file <- function(file) {
        # read the data file
        data_lines <- readr::read_lines(file)

        # split every line into a matrix on semicolons
        data_parts <- stringr::str_split(data_lines, ";", simplify = TRUE)

        # create a new tibble of every four colums in the matrix and then glue
        # the tibbles together
        max_fuels <- base::ncol(data_parts) / 4
        price_data <- purrr::map(seq_len(max_fuels),
                                 function(f) {
                                     dat <- data_parts[, (4 * (f - 1) + 1):(4 * f),
                                                       drop = FALSE]
                                     base::colnames(dat) <- c("id", "date",
                                                              "price", "fuel")
                                     tibble::as_tibble(dat) %>%
                                         dplyr::mutate(row_no = base::seq_along(id))
                                 }) %>%
            dplyr::bind_rows()

        # convert columns to the selected data types
        explicit_character_na <- function(x) {
            x <- stringr::str_trim(x)
            x[stringr::str_detect(x, "^$")] <- NA
            x
        }
        price_data %>%
            dplyr::mutate(id = base::as.integer(id),
                          date = base::as.Date(date),
                          price = base::as.numeric(price),
                          fuel = explicit_character_na(fuel)) %>%
            dplyr::filter(!is.na(id)) %>%
            dplyr::mutate(file_name = file)
    }

    # process all data and remove duplicities, even near duplicities -- see
    # above
    purrr::map(data_files, process_one_file) %>%
        dplyr::bind_rows() %>%
        dplyr::arrange(id, fuel, date, dplyr::desc(file_name),
                       dplyr::desc(row_no)) %>%
        dplyr::distinct(id, date, fuel, .keep_all = TRUE) %>%
        dplyr::select(date, id, fuel, price) %>%
        dplyr::arrange(date, id, fuel, price) %>%
        dplyr::mutate(source = id)
}


# read_pumpdroid_station_data(path_to_folder) reads all CSV files (possibly
# gzipped) and joins them into one tibble. The dates to which the information
# corresponds are not included in the CSV files -- it is read from the file
# name. Hence all the file names must be of the form
# export-pumps.YYYY-MM-DD.csv.gz
#
# inputs:
# - path_to_folder ... path to a directory where all .csv and .csv.gz files
#   contain the station data provided by Pumpdroid
#
# output:
# - a tibble of all stations (even abroad or probably non-existing, all fuels,
#   only days when there is a record in Pumpdroid database); it contains the
#   following columns:
#   - date ... the first day when the combination is observed
#   - id ... station id (integer)
#   - name ... station brand name (character)
#   - address ... station address (character)
#   - sgs ... SGS certificate of quality (logical)
#   - gps ... GPS location (character: decimal latitude,decimal longitude)
#   - lat ... latitude ... (character: decimal latitude)
#   - lon ... longitude ... (character: decimal longitude)
#   - hours ... opening hours (character; NS = non-stop)
#   - nonstop ... true if the station is open non-stop (logical)
read_pumpdroid_station_data <- function(path_to_folder) {
    # function to read one data file
    read_one_file <- function(f) {
        date <- base::as.Date(stringr::str_match(f, "\\d{4}-\\d{2}-\\d{2}")[1, 1])
        dat <- readr::read_delim(f,
                                 delim = ";",
                                 col_names = c("id", "name", "address", "sgs",
                                               "gps", "hours"),
                                 col_types = "icclcc")
        dat <- dplyr::mutate(dat, date = date)
        dat
    }

    # the list of all CSV files in the data folder
    data_files <- base::list.files(path = path_to_folder,
                                   pattern = ".*\\.csv(\\.gz)+",
                                   full.names = TRUE)

    # read and process all files in the data folder
    station_data <- purrr::map(data_files, read_one_file) %>%
        dplyr::bind_rows() %>%
        dplyr::select(date, dplyr::everything()) %>%
        dplyr::arrange(date, id)
}


# read_pumpdroid_station_locations(station_data) reads pumpdroid station data
# and returns the standard location format
#
# input:
# - station_data ... either a path to the pumpdroid station data (character)
#   or a tibble produced by read_pumpdroid_station_data()
#
# output:
# - a tibble describing the station location -- the most basic version appended
#   later with other data; it contains the following columns:
#   - id ... gas station id (integer)
#   - lat ... latitude (double: decimal form)
#   - lon ... longitude (double: decimal form)
#   - gps ... GPS location (character: deccimal latitude, decimal longitude)
#   - address ... the address provided by PumpDroid users (the last one of them)
#
# notes:
# - all unique locations of a station are kept now; one of them must be selected
#   later on or something terrible might happen (e.g. some stations might be
#   duplicated due to left joint)
#
# usage:
# - either
#   sl <- read_pumpdroid_station_locations("rawdata/pumpdroid_station_data/")
# - or
#   st <- read_pumpdroid_station_data("rawdata/pumpdroid_station_data/")
#   sl <- read_pumpdroid_station_locations(st)
read_pumpdroid_station_locations <- function(station_data) {
    # if station_data is a path to the raw data folder, read the data
    if (is.character(station_data) && length(station_data) == 1)
        station_data <- read_pumpdroid_station_data(station_data)
    #
    station_data %>%
        remove_pumpdroid_duplicities("gps") %>%
        tidyr::separate(col = gps, into = c("lat", "lon"), sep = ",",
                        remove = FALSE) %>%
        dplyr::mutate(lat = as.numeric(lat),
                      lon = as.numeric(lon),
                      source = id) %>%
        dplyr::select(date, id, lat, lon, gps, address, source)
}


# function pumpdroid_select_station_locations() selects from table of locations
# for each station just one location by a given criterium
#
# inputs:
# - station_locations ... (tibble) table of locations created by function
#   read_pumpdroid_station_locations() or clean_pumpdroid_station_locations()
# - what ... (character scalar) the criterium how the location should be
#   selected when there are several candidates:
#   - "first" ... the oldest value
#   - "last" ... the newest value (from any of the united stations; this is the
#     implicite value)
#   - "lastoriginal" ... the newest value of the original station,
#     i.e. the station to which several stations were united
#
# notes:
# - there is no difference between "last" and "lastoriginal" for stations that
#   were not united in clean_pumpdroid_station_locations()
#
# output:
# - a tibble with the same structure as stations_location but fewer rows;
#   each station has only one location; column date is dropped too
pumpdroid_select_station_locations <- function(station_locations, what = "last") {
    # check inputs
    stopifnot(what %in% c("first", "last", "lastoriginal"))
    # group by id
    station_locations <- station_locations %>%
        dplyr::group_by(id)
    # sort in such a way that the selected location is the first one
    if (what == "first")
        station_locations <- station_locations %>%
            dplyr::arrange(date, .by_group = TRUE)
    if (what == "last")
        station_locations <- station_locations %>%
            dplyr::arrange(desc(date), .by_group = TRUE)
    if (what == "lastoriginal")
        station_locations <- station_locations %>%
            dplyr::arrange(desc(date), id == source, .by_group = TRUE)
    # select the first value, ungroup, remove date, and return
    station_locations %>%
        dplyr::slice(1) %>%
        dplyr::ungroup() %>%
        dplyr::select(id, lat, lon, gps, address, source) %>%
        dplyr::arrange(id)
}

# read_pumpdroid_station_sgs(station_data) and
# read_pumpdroid_station_nonstop(station_data) reads pumpdroid station data and
# returns a table with the given feature (either sgs, or nonstop)
#
# input:
# - station_data ... either a path to the pumpdroid station data (character) or
#   a tibble produced by read_pumpdroid_station_data()
#
# output:
# - a tibble describing the given stations' feature; it contains the following
#   columns:
#   - date ... first date (Date) when the feature combination has been observed
#   - id ... gas station id (integer)
#   - the feature, one of
#       - sgs ... SGS certificate of quality (logical)
#       - nonstop ... true if open non-stop (logical)
#
# usage:
# - either
#   sf <- read_pumpdroid_station_sgs("rawdata/pumpdroid_station_data/")
# - or
#   st <- read_pumpdroid_station_data("rawdata/pumpdroid_station_data/")
#   sf <- read_pumpdroid_station_sgs(st)
# - the same for read_pumpdroid_station_nonstop()
read_pumpdroid_station_sgs <- function(station_data) {
    # if station_data is a path to the raw data folder, read the data
    if (base::is.character(station_data) && base::length(station_data) == 1)
        station_data <- read_pumpdroid_station_data(station_data)
    station_data %>%
        dplyr::select(date, id, sgs) %>%
        dplyr::mutate(source = id) %>%
        remove_pumpdroid_duplicities("sgs")
}
read_pumpdroid_station_nonstop <- function(station_data) {
    # if station_data is a path to the raw data folder, read the data
    if (base::is.character(station_data) && base::length(station_data) == 1)
        station_data <- read_pumpdroid_station_data(station_data)
    station_data %>%
        dplyr::mutate(nonstop = (hours == "NS")) %>%
        dplyr::select(date, id, nonstop) %>%
        dplyr::mutate(source = id) %>%
        remove_pumpdroid_duplicities("nonstop")
}


# normalize_pumpdroid_names(stations) converts station names into a standard
# format; it converts names into lower letters, removes strange symbols, such as
# s.r.o., a.s., .cz, čerpací stanice, and names of the roads.
#
# inputs:
# - stations ... a tibble with column names (character); it should be a tibble
#   produced by read_pumpdroid_station_data()
#
# output:
# - a tibble with the same structure as the input
#
# usage:
#   station_names <- read_pumpdroid_station_data(STATION_DATA_FOLDER) %>%
#       select(date, id, name) %>%
#       normalize_pumpdroid_names()
normalize_pumpdroid_names <- function(stations) {
    stations$name <- stations$name %>%
        base::tolower() %>%
        stringr::str_remove_all("&quot;") %>%
        stringr::str_remove_all("&apos;") %>%
        stringr::str_trim() %>%
        stringr::str_remove_all(",") %>%
        stringr::str_replace_all("\\s+", " ") %>%
        stringr::str_remove_all("s\\.\\s*r\\.\\s*o\\.*") %>%
        stringr::str_remove_all("a\\s*\\.\\s*s\\.*") %>%
        stringr::str_remove_all("\\.cz") %>%
        stringr::str_remove("^(č|c)s ") %>%
        stringr::str_remove("^[čc]erpac[ií] stanice\\s+(phm)*") %>%
        stringr::str_remove("^ii/\\d{3} ") %>%
        stringr::str_remove("^i/\\d{1,2} ") %>%
        stringr::str_trim() %>%
        stringr::str_replace_all("\\s+", " ")
    stations$name[stations$name == ""] <- NA
    stations
}


# unify_pumpdroid_station_names(stations, cor_tab) unifies various variants of
# gas stations' brand names which were filled in incorrectly by PumpDroidu
# users; the correction table created in script pumpdroid_handmade_corrections.R
# is used for this
#
# inputs:
# - stations ... (tibble) a table of stations' names as created by
#   read_pumpdroid_station_names()
# - cor_tab ... (list) a correction table created in script
#   pumpdroid_handmade_corrections.R; if not given, the function leaves the
#   brand names alone
#
# output:
# - a tibble with the same structure as stations, but the brand names are
#   unified
#
# usage:
# - this functions is not supposed to be used by the user; it is only called by
#   read_pumpdroid_station_names()
unify_pumpdroid_station_names <- function(stations, cor_tab) {
    if (is.null(cor_tab))
        return(stations)
    for (k in base::seq_along(cor_tab))
        stations$name <- unify_station_names(stations$name, cor_tab[[k]])
    stations
}


# remove_pumpdroid_generic_names() replaces generic names, such as
# "čerpací stanice", with NAs
#
# inputs:
# - a tibble with a character column name
#
# output:
# - a tibble with the same structure
#
# usage:
#   station_names %>% remove_pumpdroid_generic_names()
remove_pumpdroid_generic_names <- function(stations) {
    generic_names <- c("", "??", "pumpa", "čs", "cs", "čerpací stanice",
                       "neznama", "neznáma", "neznámá", "unknown",
                       "soukromá čerpací stanice", "soukromý",
                       "noname", "no-name", "zemědělské družstvo",
                       "nově otevřeno", "ostatni", "ostatní",
                       "n/a", "na")
    stations$name[stations$name %in% generic_names] <- NA
    stations
}


# read_pumpdroid_station_names(station_data, unify_names) reads pumpdroid
# station data and returns information on the station names in a standard format
#
# input:
# - station_data ... either a path to the pumpdroid station data (character) or
#   a tibble produced by read_pumpdroid_station_data()
# - unify_names ... an optional list of character vectors that say which brand
#   names are miss-typed; if not supplied, the station names are left alone; if
#   supplied, all brand names in the list are replaced with the brand name that
#   is the first one in character vector that contains it
#
# output:
# - a tibble describing the station names; it contains the following columns:
#   - date ... first date (Date) when the name has been observed
#   - id ... gas station id (integer)
#   - name ... the station name (brand)
#
# usage:
# - either
#   sn <- read_pumpdroid_station_names("rawdata/pumpdroid_station_data/")
# - or
#   st <- read_pumpdroid_station_data("rawdata/pumpdroid_station_data/")
#   sn <- read_pumpdroid_station_names(st)
# - or
#   sn <- read_pumpdroid_station_names(st, unify_names = unify_pumpdroid_names)
read_pumpdroid_station_names <- function(station_data, names_correction = NULL) {
    # if station_data is a path to the raw data folder, read the data
    if (is.character(station_data) && base::length(station_data) == 1)
        station_data <- read_pumpdroid_station_data(station_data)
    station_data %>%
        dplyr::select(date, id, name) %>%
        normalize_pumpdroid_names() %>%
        unify_pumpdroid_station_names(names_correction) %>%
        remove_pumpdroid_generic_names() %>%
        remove_pumpdroid_duplicities("name") %>%
        dplyr::mutate(source = id)
}


# function rename_missing_names(data, rowwise) renames gas stations having
# unknown name (coded as NA) to a unique key "unknown_#" where # is the
# station's id; if rowwise is TRUE, "_#" is added where # is a unique key. In
# this case, each unknown name is different, not only between stations but also
# in time.
#
# inputs:
# - data ... (tibble) a data table that includes at least two columns: stations'
#   id (id) and stations' names (name)
# - rowwise ... (logical) if TRUE, each NA is unique both among stations and in
#   time, see above; the implicit value is FALSE
#
# output:
# - a tibble with the same structure and data as the input data table but all NA
#   names are replaced with unique values
rename_missing_names <- function(data, rowwise = FALSE) {
    data %>%
        dplyr::group_by(id) %>%
        dplyr::mutate(.row = dplyr::row_number(),
                      name = dplyr::if_else(is.na(name),
                                            stringr::str_c("unknown", id,
                                                           if (rowwise) .row,
                                                           sep = "_"),
                                            name)) %>%
        dplyr::ungroup() %>%
        dplyr::select(-.row)
}





# corrections of station names -------------------------------------------------

# unify_station_names(station_name, ...) changes all names listed in ... to the
# first item in ...
#
# inputs:
# - station_name ... character vector of station names
# - ... various characters (at least one) -- every occurance of any string in
#   ... is replaced with the first item in ...
#
# output:
# - a character vector of unified names
#
# usage:
# - this function is not to be used by the user but called internally
#   c("ab oil", "aboil") %>% unify_station_names("ab oil", "aboil")
unify_station_names <- function(station_name, ...) {
    names <- base::unlist(base::list(...))
    new <- names[1]
    oris <- base::unique(base::setdiff(names, new))
    station_name[station_name %in% oris] <- new
    station_name
}


# function is_lag_same(var) returns a logical vector that is TRUE if the value
# is the same as the previous one, and FALSE otherwise; it is supposed to be
# used in mutate on data grouped and arranged in ascending order
is_lag_same <- function(var) {
    same <- (var == dplyr::lag(var)) | (is.na(var) & is.na(dplyr::lag(var)))
    same[is.na(same)] <- FALSE
    same[1] <- FALSE
    same
}


# function remove_pumpdroid_duplicities(df, feature) takes a dataset df that
# includes at least three columns: id (integer), date (Date), and one other
# column called feature; it removes true duplicities, which are the rows that
# have the same value of feature as their preceding rows when data are sorted by
# date and grouped by id
#
# inputs:
# - df ... tibble as described above
# - feature ... (character scalar) name of the feature column
#
# output:
# - (tibble) the same table as df but duplicities in feature are removed
remove_pumpdroid_duplicities <- function(df, feature) {
    df %>%
        dplyr::group_by(id) %>%
        dplyr::arrange(date, .by_group = TRUE) %>%
        dplyr::mutate(.same := is_lag_same(!!rlang::sym(feature))) %>%
        dplyr::ungroup() %>%
        dplyr::filter(!.same) %>%
        dplyr::select(-.same)
}

# function correct_pumpdroid_station_names(station_names, correction_table)
# corrects ownership table in lines with directions given a correction table
#
# inputs:
# - station_names ... (tibble) table of ownership created by function
#   read_pumpdroid_station_names() or clean_pumpdroid_station_names()
# - correction_table ... (tibble) table describing how the ownership should be
#   corrected; if it is NULL or has no rows then station_names are returned
#   unchanged; the tibble must have the following columns:
#   - id ... (integer) station id for which the brand name should be replaced
#   - from, to ...(date) the interval where the brand name should be replace
#   - name ... (character) the name that should replace the old one
#
# output:
# - (tibble) a table with the same structure as station_names where some brand
#   names are corrected
#
# details:
# - the changes are applied sequentially, i.e. the order of rows in the
#   correction table matters and must be right---the changes described in the
#   first row are applied first, then changes described in the second row, etc.;
#   it means that it is possible to change the brand name from 2016-01-01 to the
#   end of the universe to A first, and then from 2017-05-05 to 2018-12-31 to B
#   later, and the corrections would be applied correctly; but beware: the
#   opposite order of the rows in this example would be incorrect: the change to
#   B would be replaced by the change to A; this additivity (though dangerous)
#   is useful: the correction table may apply changes one by one, and when data
#   are updated, new rows may be added at the end of the old one
correct_pumpdroid_station_names <- function(station_names,
                                            correction_table = NULL) {
    # if the correction table is NULL, return the station_names as is
    if (is.null(correction_table))
        return(station_names)
    # remove the replaced rows (i.e. names in the corrected time span)
    keep <- base::rep(TRUE, length.out = nrow(station_names))
    for (k in base::seq_len(nrow(correction_table))) {
        keep[station_names$id == correction_table$id[k] &
                 station_names$date >= correction_table$from[k] &
                 station_names$date <= correction_table$to[k]] <- FALSE
    }
    station_names <- station_names[keep, ]
    # add rows with new (hopefully correct) names
    corrs <- correction_table %>%
        dplyr::rename(date = from) %>%
        dplyr::mutate(source = -1L) %>%
        dplyr::select(date, id, name, source)
    station_names <- dplyr::bind_rows(station_names, corrs)
    # sort the resulting table
    station_names <- dplyr::arrange(station_names, id, date)
    # remove duplicities
    station_names <- remove_pumpdroid_duplicities(station_names, "name")
    # return
    station_names
}





# elimination of non-existing stations -----------------------------------------

# remove_stations(dt) removes all observations of given stations from table dt
#
# inputs:
# - ds ... (tibble) table from which stations are to be eliminated; it must
#   includecolumn id which identifies the stations
# - ids ... (integer vector) vector of stations' ids which must be dropped from
#   dt; if it is NULL (the implicit value), then dt is returned as it was
#
# output:
# - a tibble with the same structure as dt but fewer rows---all observations
#   with id %in% ids are dropped
remove_stations <- function(ds, ids = NULL) {
    if (is.null(ids))
        return(ds)
    dplyr::filter(ds, !(id %in% ids))
}


# unite_stations(ds, unite, feature_target) unites given stations in table ds
#
# inputs:
# - ds ... (tible) table in which stations should be united
# - unite ... (tibble) the correction table; if NULL (the implicite value),
#   the function returns ds as it was
# - feature_target ... (character: "price", "name", "location", "sgs",
#   "nonstop") name of the table (feature) that is being cleaned / united
#
# output:
# - a tibble with the same structure as ds to which the unification is applied
#   as described by the correction table unite
#
# details:
# - function iterates over all rows of table unite which either have their
#   feature "all" or have the feature included in feature_target
# - the resulting dataset must be cleaned after this function is applied because
#   there might appear duplicities; every kind of dataset must be cleaned by a
#   specialized function
# - when the duplicities are eliminated, the resulting dataset must be re-sorted
unite_stations <- function(ds, unite = NULL, feature_target) {
    # if correction table unite is not given, return ds unchanged
    if (is.null(unite))
        return(ds)
    # check inputs
    stopifnot(base::is.character(feature_target))
    stopifnot(feature_target %in% c("price", "name", "location", "sgs",
                                    "nonstop"))
    # filter those rows of the corrction table unite that correct the given
    # feature
    unite <- dplyr::filter(unite, feature == "all" |
                        stringr::str_detect(feature, feature_target))
    # apply the changes described by correction table unite row by row
    for (k in base::seq_len(base::nrow(unite))) {
        # if the original values should not be kept within a time window (keep =
        # FALSE) then remove them from ds; if keep = TRUE, then values from
        # target and souce are mixed; they are replace by the source otherwise
        if (!unite$keep[k])
            ds <- dplyr::filter(ds, !(id == unite$target_id[k] &
                                          date >= unite$from[k] &
                                          date <= unite$to[k]))
        # change id of the source withing the time window to target's id
        ds$id[ds$id == unite$source_id[k] &
                  ds$date >= unite$from[k] &
                  ds$date <= unite$to[k]] <- unite$target_id[k]
    }
    # remove source stations from ds
    ds <- remove_stations(ds, unite$source_id)
    # return
    ds
}


# function clean_price_data(prices, unite, remove) removes non-existing stations
# from the price table
#
# input:
# - prices ... (tibble) table of prices created by function
#   read_pumpdroid_price_data(); it must include columns id, date, and fuel
# - unite ... (tibble) table that describes which stations must be united;
#   if NULL (which is the implicite value), no unification is done
# - remove ... (integer vector) atomic vector of ids of the stations that would
#   be dropped from the data; if NULL (the implicit value), no removal is done
#
# output:
# - (tibble) the same table as prices but with some rows eliminated;
#   see functions unite_stations() and remove_vector()
#
# notes:
# - if, after the unification, there are several prices for one station, date,
#   and fuel, one is chosen randomly
#
# TODO: odstranění duplicit je zatím náhodné, protože zatím nevím, jak to udělat správně
clean_pumpdroid_price_data <- function(prices,
                                       unite = NULL,
                                       remove = NULL) {
    prices %>%
        remove_stations(remove) %>%
        unite_stations(unite, "price") %>%
        dplyr::group_by(id, fuel, date) %>%
        dplyr::slice(1) %>%
        dplyr::ungroup() %>%
        dplyr::arrange(date, id, fuel)
}


# clean_station_locations(locations, unite, remove) removes from the table of
# station locations stations that do not exist
#
# inputs:
# - locations ... (tibble) table of station locations created by function
#   read_pumpdroid_station_locations(); it must include column id
# - unite ... (tibble) table that describes which stations must be united;
#   if NULL (which is the implicite value), no unification is done
# - remove ... (integer vector) atomic vector of ids of the stations that would
#   be dropped from the data; if NULL (the implicit value), no removal is done
#
# output:
# - (tibble) the same table as locations but with some rows eliminated;
#   see functions unite_stations() and remove_vector()
#
# notes:
# - after this step, there may still be several locations for some stations;
#   select which ones to keep with pumpdroid_select_station_locations()
clean_pumpdroid_station_locations <- function(locations,
                                              unite = NULL,
                                              remove = NULL) {
    locations %>%
        remove_stations(remove) %>%
        unite_stations(unite, "location") %>%
        dplyr::group_by(id, date) %>%
        dplyr::slice(1) %>%
        dplyr::ungroup() %>%
        dplyr::arrange(date, id)
}


# function clean_station_names(names, unite, remove) removes stations that do
# not exist from table of ownership
#
# inputs:
# - names ... (tibble) table of ownership created by function
#   read_pumpdroid_station_names(); it must include column id
# - unite ... (tibble) table that describes which stations must be united;
#   if NULL (which is the implicite value), no unification is done
# - remove ... (integer vector) atomic vector of ids of the stations that would
#   be dropped from the data; if NULL (the implicit value), no removal is done
#
# output:
# - the same tibble as names but with some rows eliminated; see functions
#       unite_stations() and remove_vector()
#
# TODO: odstranění duplicit je zatím náhodné, protože zatím nevím, jak to udělat správně
clean_pumpdroid_station_names <- function(names,
                                          unite = NULL,
                                          remove = NULL) {
    names %>%
        remove_stations(remove) %>%
        unite_stations(unite, "name") %>%
        remove_pumpdroid_duplicities("name") %>%
        dplyr::group_by(id, date) %>%
        dplyr::slice(1) %>%
        dplyr::ungroup() %>%
        dplyr::arrange(date, id)
}




# function clean_station_features(locations, unite, remove) removes from a table
# of station features non-existing stations
#
# inputs:
# - features ... (tibble) table of features created by function
#   read_pumpdroid_station_sgs() or read_pumpdroid_station_nonstop();
#   it must contain column id
# - unite ... (tibble) table that describes which stations must be united;
#   if NULL (which is the implicite value), no unification is done
# - remove ... (integer vector) atomic vector of ids of the stations that would
#   be dropped from the data; if NULL (the implicit value), no removal is done
#
# output:
# - the same tibble as features but with some rows eliminated; see function
#       unite_stations() a remove_vector()
#
# TODO: odstranění duplicit je zatím náhodné, protože zatím nevím, jak to udělat správně
clean_pumpdroid_station_features <- function(features,
                                             unite = NULL,
                                             remove = NULL) {
    fn <- features %>% dplyr::select(-date, -id, -source) %>% names()
    features %>%
        remove_stations(remove) %>%
        unite_stations(unite, fn) %>%
        remove_pumpdroid_duplicities(fn) %>%
        dplyr::group_by(id, date) %>%
        dplyr::slice(1) %>%
        dplyr::ungroup() %>%
        dplyr::arrange(date, id)
}





# brute force brand name correction after big takeovers ------------------------

# This function can be used to correct brand names after big takeovers. It
# assumes that there is some terminal date (possibly more than one terminal
# date) when all names for a brand are already correct after a takeover. It
# also assumes that we know the precise date of the takeovers. If so, we can
# detect all stations that were taken over and impose the correct date.
#
# Why the takeover may be seen in Pumpdroid data too late: It is because it
# takes some time before the station is physically repainted, and citizens can
# change the brand name usually only when they can see the station colors
# changed.

# function correct_brandnames_by_brute_force(station_names, cor_tab) corrects
# the takeover dates in station_names table by brute force
#
# inputs:
# - station_names ... (tibble) the table of station names (brand owernship)
#   evolution over time
# - cor_tab ... (tibble) the correction table as defined above in
#   takeovers_brute_force_correction_table; it must include the following
#   colums:
#   - takeover_date ... (date) the precise the of the takeover
#   - terminal_date ... (date) the date when all taken over stations have the
#       correct brand name in the Pumpdroid data
#   - source ... (character) the brand name being taken over
#   - target ... (character) the new brand name of the taken over stations
#
# value:
# - tibble of the same strucure as station_names but takeover dates are imposed
#   by brute force
correct_brandnames_by_brute_force <- function(station_names, cor_tab = NULL) {
    # if no cor_tab is given, return the station_names unchanged
    if (is.null(cor_tab) || base::nrow(cor_tab) == 0)
        return(station_names)

    # function min_date_of_name_while_takeover(dates, takeover_date) gets a Date
    # vector of dates for one station and return the number of the date of the
    # last known name before or at the date of the takeover
    min_date_of_name_while_takeover <- function(dates, takeover_date) {
        n <- base::seq_along(dates)[dates >= takeover_date]
        if (base::length(n) == 0)
            n <- 1
        if ((base::length(n) == 1 && n == 1) || dates[n[1]] != takeover_date)
            n <- c(n[1] - 1, n)
        base::min(n)
    }

    # function all_names_source_or_target(names, source, target) takes a
    # charactor vector of station names and returns a logical: TRUE if all the
    # names are either source or target, or FALSE otherwise
    all_names_source_or_target <- function(names, source, target) {
        base::all(names %in% c(source, target))
    }

    # if there is a cor_tab, apply its corrections row-by_row
    for (k in base::seq_len(nrow(cor_tab))) {
        # get ids of stations that belong to target at the end of the targeted
        # period
        targeted_ids <- station_names %>%
            dplyr::group_by(id) %>%
            dplyr::filter(date <= cor_tab$terminal_date[k]) %>%
            dplyr::arrange(dplyr::desc(date), .by_group = TRUE) %>%
            dplyr::slice(1) %>%
            dplyr::ungroup() %>%
            dplyr::filter(name == cor_tab$target[k]) %>%
            dplyr::pull(id)

        # get ids of stations that had only two names between the date of the
        # takeover and the targeted period: source and target from a row in
        # cor_tab (in any order, which is necessary because some people change
        # agip -> mol -> agip -> mol etc.)
        filtered_stations <- station_names %>%
            dplyr::filter(id %in% targeted_ids) %>%
            dplyr::group_by(id) %>%
            dplyr::summarise(
                min_n = min_date_of_name_while_takeover(date,
                                                        cor_tab$takeover_date[k]),
                max_n = n(),
                .groups = "drop"
            ) %>%
            dplyr::left_join(station_names, by = "id") %>%
            dplyr::group_by(id) %>%
            dplyr::mutate(row = dplyr::row_number()) %>%
            dplyr::filter(row >= min_n, row <= max_n) %>%
            dplyr::summarise(
                keep = all_names_source_or_target(name,
                                                  cor_tab$source[k],
                                                  cor_tab$target[k]),
                brandname_original = str_c(name, collapse = " -> "),
                .groups = "drop"
            ) %>%
            dplyr::filter(keep)
        # store the original brandnames
        original_brandnames <- filtered_stations %>%
            dplyr::select(id, brandname_original)
        # ids of the station that that follow the condition of this section
        corrected_ids <- filtered_stations %>%
            dplyr::pull(id)

        # for the corrected stations, create new rows that will be added to
        # station_names; it consits of the corrected_ids, the date of the
        # takeover, and the name of the target company that took source over
        additional_names <- tibble::tibble(date = cor_tab$takeover_date[k],
                                           id = corrected_ids,
                                           name = cor_tab$target[k],
                                           source = -1) %>%
            dplyr::left_join(original_brandnames, by = "id")

        # for the corrected stations, remove all names from takeover date to the
        # targeted period and add new names at the takeover date and resort
        station_names <- station_names %>%
            dplyr::filter(!(id %in% corrected_ids &
                                date >= cor_tab$takeover_date[k] &
                                date <= cor_tab$terminal_date[k])) %>%
            dplyr::bind_rows(additional_names) %>%
            dplyr::arrange(id, date)
    }

    # return
    station_names
}
