# -------------------------------------
# Script: prepare_pumpdroid_data.R
# Author: Michal Kvasnička
# Purpose: This script loads data provided by PumpDroid, cleans them, adds
#          various statistics and variables, and saves the data
#          in a standardized format that could be used for econometrics analysis
#          and simulation.
# Inputs:  # TODO: přidat vstupy
# Outputs:  # TODO: přidat výstupy
# Notes:
#
# Copyright(c) Masaryk University
# -------------------------------------


# clean workspace
rm(list = ls())


# paths to files and folders
#   - R code
CODEDIR <- "R"
GENERAL_FUNCTIONS <- file.path(CODEDIR, "preparatory_functions_general.R")
PUMPDROID_FUNCTIONS <- file.path(CODEDIR, "preparatory_functions_pumpdroid.R")
#   - raw data provided by PumpDroid
RAWDATA_FOLDER <- "rawdata"
PRICE_DATA_FOLDER <- file.path(RAWDATA_FOLDER, "pumpdroid_price_data")
STATION_DATA_FOLDER <- file.path(RAWDATA_FOLDER, "pumpdroid_station_data")
#   - maps
MAP_FOLDER <- "map_shapes"
ARCCR_ADMIN_FOLDER <- file.path(MAP_FOLDER, "AdministrativniCleneni_v13.gdb")
ARCCR_GEO_FOLDER <- file.path(MAP_FOLDER, "ArcCR500_v33.gdb")
OSM_FOLDER <- file.path(MAP_FOLDER, "OpenStreetMaps")
NAVIGATION_MAP <- file.path(MAP_FOLDER, "czech-republic-latest.osm.pbf")
#   - folder where roadrunner resides
ROADRUNNER_FOLDER <- "roadrunner"
#   - data generated by this script
DATA_FOLDER <- "data"
INPUT_DATA_FILE <- "pumpdroid_input_data.RData"  # TEMP just for saving inputs
ANALYTICAL_DATA_FILE <- "pumpdroid_data.RData"
#   - correction tables
GENERAL_CORRECTION_TABLES_FILE <-
    file.path(DATA_FOLDER, "handmade_corrections_general.RData")
PUMPDROID_CORRECTION_TABLES_FILE <-
    file.path(DATA_FOLDER, "handmade_corrections_pumpdroid.RData")


# load necessary scripts and packages
source(GENERAL_FUNCTIONS)
source(PUMPDROID_FUNCTIONS)


# set paralel processing
# - some processes would run on all cores; comment the following line if it's
#   not welcome
# - by default, it uses all available cores; you can change the number of used
#   cores by parameter workers; you can learn the number of available core by
#   availableCores(); thus if you want to use all cores but one, set
#   plan(multisession, workers = availableCores() - 1)
plan(multisession)


# PUMPDROID
# load corrections tables for correting Pumpdroid data if they exist; othewise
# they would be NULL
pumpdroid_station_names_to_be_unified <-
    load_from(PUMPDROID_CORRECTION_TABLES_FILE,
              "pumpdroid_station_names_to_be_unified")
pumpdroid_stations_to_be_united <-
    load_from(PUMPDROID_CORRECTION_TABLES_FILE,
              "pumpdroid_stations_to_be_united")
pumpdroid_stations_to_be_removed <-
    load_from(PUMPDROID_CORRECTION_TABLES_FILE,
              "pumpdroid_stations_to_be_removed")
pumpdroid_station_names_correction <-
    load_from(PUMPDROID_CORRECTION_TABLES_FILE,
              "pumpdroid_station_names_correction")
pumpdroid_takeovers_correction <-
    load_from(PUMPDROID_CORRECTION_TABLES_FILE,
              "takeovers_brute_force_correction_table")


# GENERAL
# load general correction tables if they exist; otherwise they are NULL
same_owners_correction_table <- load_from(GENERAL_CORRECTION_TABLES_FILE,
                                          "same_owners_correction_table")


# PUMPDROID
# read Pumpdroid data on prices
price_data <- read_pumpdroid_price_data(PRICE_DATA_FOLDER)


# PUMPDROID
# read and repair Pumpdroid data on gas station location, features, and
# ownership
# - read Pumpdroid data on stations and join them to one table
station_data <- read_pumpdroid_station_data(STATION_DATA_FOLDER)
# - get data on gas stations' locations (the last one for every station)
station_locations <- read_pumpdroid_station_locations(station_data)
# - get onwership and starndardize brand names
station_names <-
    read_pumpdroid_station_names(station_data,
                                 names_correction = pumpdroid_station_names_to_be_unified)
# - get data on gas station features
station_sgs <- read_pumpdroid_station_sgs(station_data)
station_nonostop <- read_pumpdroid_station_nonstop(station_data)


# PUMPDROID clean the data using hand-made correction tables---remove gas
# stations that do not exist: if they never existed, drop them; if several
# stations are in fact one, unite them
price_data <-
    clean_pumpdroid_price_data(price_data,
                               unite = pumpdroid_stations_to_be_united,
                               remove = pumpdroid_stations_to_be_removed)
station_locations <-
    clean_pumpdroid_station_locations(station_locations,
                                      unite = pumpdroid_stations_to_be_united,
                                      remove = pumpdroid_stations_to_be_removed)
station_names <-
    clean_pumpdroid_station_names(station_names,
                                  unite = pumpdroid_stations_to_be_united,
                                  remove = pumpdroid_stations_to_be_removed)
station_sgs <-
    clean_pumpdroid_station_features(station_sgs,
                                     unite = pumpdroid_stations_to_be_united,
                                     remove = pumpdroid_stations_to_be_removed)
station_nonostop <-
    clean_pumpdroid_station_features(station_nonostop,
                                     unite = pumpdroid_stations_to_be_united,
                                     remove = pumpdroid_stations_to_be_removed)


# PUMPDROID
# for each station, select just one location (the last one in this case)
station_locations <- pumpdroid_select_station_locations(station_locations,
                                                        "last")


# PUMPDROID
# correct station ownership using handmade correction table
# TODO: přidat všechny ruční opravy
station_names <-
    correct_pumpdroid_station_names(station_names,
                                    pumpdroid_station_names_correction)


# PUMPDROID
# correct dates of big takeovers -- a bit brave code
station_names <-
    correct_brandnames_by_brute_force(station_names,
                                      pumpdroid_takeovers_correction)


# GENERAL
# append the station location data
station_locations <- station_locations %>%
    add_arccr_municipalities(ARCCR_ADMIN_FOLDER)  # obec, okres, kraj, country


# GENERAL
# odstraň čerpací stanice v zahraničí
price_data <- remove_foreign_gas_stations(price_data, station_locations)
station_locations <- remove_foreign_gas_stations(station_locations, station_locations)
station_names <- remove_foreign_gas_stations(station_names, station_locations)
station_sgs <- remove_foreign_gas_stations(station_sgs, station_locations)
station_nonostop <- remove_foreign_gas_stations(station_nonostop, station_locations)


# GENERAL
# odstraň čerpací stanice, které nemají žádná pozorování o cenách
station_locations <- remove_gas_stations_with_no_price_data(station_locations, price_data)
station_names <- remove_gas_stations_with_no_price_data(station_names, price_data)
station_sgs <- remove_gas_stations_with_no_price_data(station_sgs, price_data)
station_nonostop <- remove_gas_stations_with_no_price_data(station_nonostop, price_data)


# GENERAL
# připoj informace o silnicích
station_locations <- station_locations %>%
    add_arccr_roads(ARCCR_GEO_FOLDER) %>%  # z ARC ČR
    add_osm_roads(OSM_FOLDER)  # z OpenStreetMaps


# GENERAL
# spočítej vzdálenosti mezi čerpacími stanicemi
#   - vzdušnou čarou mezi všemi (je symetrické)
station_beeline_distances <- get_beeline_distances(station_locations)
#   - vzdálenost po silnici je asymetrická a počítá se jen do zadané vzdálenosti
#     na Windows může být potřeba nastavit temp_map_folder na rozumný
#     existující adresář, např. "W:/work/benzin2018/goo" -- pokud cesta
#     k domovskému adresáři obsahuje diakritiku, mezery apod.
station_driving_distances <- get_driving_distances(station_locations,
                                                   station_beeline_distances,
                                                   max_dist = 40,
                                                   map_path = NAVIGATION_MAP,
                                                   roadrunner_folder = ROADRUNNER_FOLDER)


# PUMPDROID?
# replace all missing brand names coded as NA as "unknown_#" where # is the stations' id
station_names <- station_names %>% rename_missing_names()


# ulož dojezdové vzdálenosti
save(station_beeline_distances,
     file = file.path(DATA_FOLDER, "station_beeline_distances.RData"))
save(station_driving_distances,
     file = file.path(DATA_FOLDER, "station_driving_distances_fastest.RData"))


# TEMP save the all input data
save(price_data, station_locations, station_names, station_sgs, station_nonostop,
     station_beeline_distances, station_driving_distances,
     file = file.path(DATA_FOLDER, INPUT_DATA_FILE))


# GENERAL
# vytvoř data pro analýzu -- zvlášť pro každý zkoumaný druh paliva a zadanou
# frekvenci; obsahuje všechna data pro ekonometrii včetně měřítek konkurence;
# počítáno pro zadané kritérium existence čerpacích stanic; některá data jsou
# interpolovaná
existence_criterium <- "station_exists_30"
natural_monthly <- prepare_average_fuel_data("NATURAL95",
                                             by = "1 month",
                                             exist_lag = 30L,
                                             price_data,
                                             station_locations,
                                             station_names,
                                             station_sgs,
                                             station_nonostop,
                                             remove_source = TRUE) %>%
    filter(from_date >= "2013-01-01", to_date <= "2018-12-31") %>%
    add_spatial_clustering(ARCCR_ADMIN_FOLDER,
                           existence = existence_criterium,
                           same_group = same_owners_correction_table) %>%
    add_competitors_in_distances(station_driving_distances,
                                 distances = seq(from = 0, to = 12, by = 3),
                                 existence = existence_criterium,
                                 same_group = same_owners_correction_table) %>%
    add_distance_to_closest_competitor(station_driving_distances,
                                       existence = existence_criterium,
                                       same_group = same_owners_correction_table) %>%
    add_city_effects()


# GENERAL
# save the data prepared for analysis
save(natural_monthly,
     file = file.path(DATA_FOLDER, ANALYTICAL_DATA_FILE))
