# -------------------------------------
# Script: utilities.R
# Author: Michal Kvasnička
# Purpose: This script contains utilities functions.
#
# Copyright(c) Masaryk University
# -------------------------------------


# fun with future ---------------------------------------------------------

# used_future_workers() returns how many workers are actually used in the
# current plan
#
# inputs: none
#
# output:
# - a single number
#
# details:
# - You may wonder why not using future::availableCores(). Here's why: Typically
#   all cores would be used. However, a user may decrease the number of used
#   cores with plan(multisession, workers = 2).
used_future_workers <- function() {
    w <- base::formals(future::plan())$workers
    if (inherits(w, "call"))
        w <- eval(w)
    as.numeric(w)
}




# reading RData ---------------------------------------------------------------------

# list_rdata(file) takes a path to a RData and returns names of all variables
# stored in the RData
#
# inputs:
# - file ... (character) a path to a RData
#
# output:
# - character vector with  names of all variables stored in the RData
#
# notes:
# - the function has to load the whole RData to memory and then drop it
#
# usage:
#   var_names <- list_rdata("file.RData")
list_rdata <- function(file) {
    env <- new.env()
    base::load(file = file, envir = env)
    base::ls(env)
}


# load_from(file, variables) reads all variables from a RData file and returns
# them instead of loading them into the environment
#
# inputs:
# - file ... (character) a path to a RData file
# - variables ... (character vector) names of the variables in the RData file
#   that should be loaded and returned
#
# output:
# - if only one variable is loaded, it is returned
# - if more than one variable is loaded, the named list of these variables is
#   returned
# - if some variables in not stored in the RData file, NULL is returned and
#   a warning is thrown
#
# notes:
# - the function has to load the whole RData to memory and then drop it
#
# usage:
# - to read one variable (returns the variable):
#   x <- load_from("file.RData", "var")
# - to read many variables (returns the list of the variables):
#   x <- load_from("file.RData", c("var1", "var2"))
# - to read all variables (returns the list of the variables):
#   var_names <- list_rdata("file.RData")
#   x <- load_from("file.RData", var_names)
load_from <- function(file, variables) {
    if (!(base::is.character(file) && base::length(file) == 1))
        stop("Filename must be a character vector of lenght one.")
    if (!base::file.exists(file))
        stop("File ", file, "does not exist.")
    env <- new.env()
    ok <- FALSE
    try(
        {
            base::load(file = file, envir = env)
            ok <- TRUE
        },
        silent = TRUE)
    if (!ok)
        stop("File ", file, "is not readable, corrupted, or not RData.")
    result <- purrr::map(variables, ~env[[.]])
    waridx <- purrr::map_lgl(result, is.null)
    if (any(waridx))
        purrr::walk(variables[waridx],
                    ~warning("load_from() : Variable ", ., " is missing from the file!"),
                    call = FALSE)
    if (base::length(variables) == 1)
        return(result[[1]])
    stats::setNames(result, variables)
}
